/* aenc.h */

#ifndef AENC_H
#define AENC_H

#include <inttypes.h>
#include <pthread.h>

#include "aframe.h"
#include "aframe_mgr.h"
#include "aread.h"
#include "error_mgr.h"


#define AENC_STATE_NOT_READY       (1UL<<0)
#define AENC_STATE_STOP            (1UL<<1)
#define AENC_STATE_START           (1UL<<2)
#define AENC_STATE_PAUSE           (1UL<<3)
#define AENC_STATE_QUIT            (1UL<<4)
#define AENC_STATE_ERROR           (1UL<<5)
#define AENC_STATE_INTERFACE_WAIT  (1UL<<6)
#define AENC_STATE_THREAD_WAIT     (1UL<<7)
#define AENC_STATE_THREAD_FINISH   (1UL<<8)


typedef struct {
  pthread_t aenc_th;
  pthread_mutex_t mutex;
  pthread_cond_t interface_wait_cond;
  pthread_cond_t thread_wait_cond;

  uint32_t state;

  aframe_t *src_aframe;
  aframe_t *dst_aframe;
  aframe_chain_t aframe_chain;

  aread_t *aread;
  aframe_mgr_t *aframe_mgr;
  aframe_ready_cb_t aframe_ready_cb;
  void *aframe_ready_cb_data;

  int rate;
  int out_rate;
  int channels;
  int bits_per_sample;
  uint32_t acodec;
  int frame_sample_size;

  error_cb_t error_cb;
  void *error_cb_data;
} aenc_t;

#define AENC_OK    0
#define AENC_FAIL -1

int aenc_set_aframe_ready_cb(aenc_t *aenc,
              aframe_ready_cb_t aframe_ready_cb, void *aframe_ready_cb_data);
int aenc_set_aframe_mgr(aenc_t *aenc, aframe_mgr_t *afrm_mgr);
int aenc_set_error_cb(aenc_t *aenc, error_cb_t error_cb, void *error_cb_data);
aframe_t *aenc_get_aframe(aenc_t *aenc);
int aenc_get_frame_sample_size(aenc_t *aenc);
int aenc_get_out_rate(aenc_t *aenc);
int aenc_state_error(aenc_t *aenc);
int aenc_state_start(aenc_t *aenc);
int aenc_state_stop(aenc_t *aenc);
int aenc_state_quit(aenc_t *aenc);
int aenc_state_pause(aenc_t *aenc);
int aenc_state_toggle_pause(aenc_t *aenc);
int aenc_error_quit (aenc_t *aenc);
int aenc_quit (aenc_t *aenc);
int aenc_error_quit (aenc_t *aenc);
aenc_t* aenc_init (int rate, int channels, int bits_per_sample, uint32_t acodec, aread_t *aread);


#endif /* AENC_H */

