/* aframe.h */

#ifndef AFRAME_H
#define AFRAME_H

#include <inttypes.h>

#define AFRAME_FLAG_NORMAL (1<<0)
#define AFRAME_FLAG_EXT    (1<<1)
#define AFRAME_FLAG_START  (1<<2)
#define AFRAME_FLAG_END    (1<<3)
#define AFRAME_FLAG_QUIT   (1<<4)
#define AFRAME_FLAG_ERROR  (1<<5)
#define AFRAME_FLAG_DROP   (1<<6)

#define AFRAME_MAX_CHANNELS  2

typedef struct AFRAME_T {
  uint32_t flag;

  int rate;
  int channels;
  int bits_per_sample;

  int sample_buf_size;
  int samples;
  uint8_t *sample[AFRAME_MAX_CHANNELS];

  int stream_length;
  int stream_buf_size;
  uint8_t *stream;

  struct AFRAME_T *prev;
  struct AFRAME_T *next;

  uint8_t *bufp;
} aframe_t;


typedef int (*aframe_ready_cb_t)(void *);


typedef struct {
  int num_frames;
  aframe_t *aframes;
} aframe_chain_t;

static inline void
aframe_chain_init(aframe_chain_t *afrm_chain)
{
  afrm_chain->num_frames = 0;
  afrm_chain->aframes = NULL;
}

static inline int
aframe_chain_add(aframe_chain_t *afrm_chain, aframe_t *afrm)
{
  if (afrm_chain->aframes) {
    afrm->prev = afrm_chain->aframes->prev;
    afrm->next = afrm_chain->aframes;
    afrm_chain->aframes->prev->next = afrm;
    afrm_chain->aframes->prev = afrm;
  } else {
    afrm->prev = afrm->next = afrm;
    afrm_chain->aframes = afrm;
  }
  afrm_chain->num_frames++;
  return afrm_chain->num_frames;
}

static inline aframe_t*
aframe_chain_remove(aframe_chain_t *afrm_chain)
{
  aframe_t *afrm;

  if (!afrm_chain->aframes) {
    return NULL;
  }
  afrm = afrm_chain->aframes;
  if (afrm->next == afrm) {
    afrm_chain->aframes = NULL;
  } else {
    afrm_chain->aframes = afrm->next;
    afrm->prev->next = afrm->next;
    afrm->next->prev = afrm->prev;
  }
  afrm->prev = afrm->next = NULL;
  afrm_chain->num_frames--;
  return afrm;
}


#endif /* AFRAME_H */

