/* aframe_mgr.h */

#ifndef AFRAME_MGR_H
#define AFRAME_MGR_H

#include <inttypes.h>
#include <pthread.h>

#include "aframe.h"

typedef struct {
  pthread_mutex_t mutex;
  int rate;
  int channels;
  int bits_per_sample;
  int bytes_per_sample;         /* (bits_per_sample/8) * channels */

  int aframe_buf_sample_size;
  int aframe_buf_byte_size;

  unsigned int last_frame_alloc_count;
  unsigned int max_last_frame_alloc_count;
  int max_frame_num;
  int max_ext_frame_num;

  int alloc_frame_num;
  int alloc_ext_frame_num;
  int used_frame_num;
  int free_frame_num;

  aframe_chain_t free_chain;
} aframe_mgr_t;

#define AFRAME_MGR_OK     0
#define AFRAME_MGR_FAIL  -1

aframe_t* aframe_mgr_get_aframe(aframe_mgr_t *afrm_mgr);
aframe_t* aframe_mgr_get_ext_aframe(aframe_mgr_t *afrm_mgr);
int aframe_mgr_add_free_frame(aframe_mgr_t *afrm_mgr, aframe_t *afrm);

aframe_mgr_t* aframe_mgr_init(int rate, int channels, int bits_per_sample,
    int frame_sample_size, double buffer_time);
int aframe_mgr_quit(aframe_mgr_t *afrm_mgr);


#endif /* AFRAME_MGR_H */

