/* aread.h */

#ifndef AREAD_H
#define AREAD_H

#include <inttypes.h>
#include <pthread.h>

#include "get_cur_u_time.h"
#include "oss.h"
#include "aframe_mgr.h"
#include "error_mgr.h"

#define AREAD_STATE_NOT_READY       (1UL<<0)
#define AREAD_STATE_STOP            (1UL<<1)
#define AREAD_STATE_START           (1UL<<2)
#define AREAD_STATE_PAUSE           (1UL<<3)
#define AREAD_STATE_QUIT            (1UL<<4)
#define AREAD_STATE_ERROR           (1UL<<5)
#define AREAD_STATE_INTERFACE_WAIT  (1UL<<6)
#define AREAD_STATE_THREAD_WAIT     (1UL<<7)
#define AREAD_STATE_THREAD_FINISH   (1UL<<8)

typedef struct {
  pthread_t aread_th;
  pthread_mutex_t mutex;
  pthread_cond_t interface_wait_cond;
  pthread_cond_t thread_wait_cond;

  uint32_t state;
  u_time_t state_time;

  int rate;
  int channels;
  int bits_per_sample;
  int bytes_per_sample;
//  int fragsize;
//  int fragstotal;

  int sync_system_time_flag;
  int record_flag;

  oss_t *oss;
  int read_frame_byte_size;
  int read_frame_sample_size;
  int read_frame_num;
  uint8_t **read_frame;
  u_time_t *read_frame_time;
  int32_t read_frame_count;
  int32_t read_frame_cur;
  int32_t read_frame_cur_offset;

  long long record_samples;
  u_time_t record_time;
  u_time_t record_start_time;

  aframe_t *aframe;
  aframe_mgr_t *aframe_mgr;
  aframe_chain_t aframe_chain;
  aframe_ready_cb_t aframe_ready_cb;
  void *aframe_ready_cb_data;

  u_time_t read_start_time;
  int32_t read_start_count;
  int32_t check_count_interval;
  int32_t next_check_count;

  int gap;
  double read_gap;

  int interp_frame_interval;
  int next_interp_frame_count;

  error_cb_t error_cb;
  error_cb_t error_cb_data;
} aread_t;

#define AREAD_OK    0
#define AREAD_FAIL -1


int aread_get_rate(aread_t *aread);
int aread_get_channels(aread_t *aread);
int aread_get_bits_per_sample(aread_t *aread);
aframe_t* aread_get_aframe(aread_t *aread);
int aread_set_aframe_ready_cb(aread_t *aread, aframe_ready_cb_t aframe_ready_cb, void *aframe_ready_cb_data);
int aread_set_error_cb(aread_t *aread, error_cb_t error_cb,void *error_cb_data);
int aread_set_aframe_mgr(aread_t *aread, aframe_mgr_t *afrm_mgr);
int aread_state_start(aread_t *aread, u_time_t *state_time);
int aread_state_stop(aread_t *aread, u_time_t *state_time);
int aread_state_toggle_pause(aread_t *aread, u_time_t *state_time);
int aread_state_quit(aread_t *aread, u_time_t *state_time);
int aread_state_error(aread_t *aread, u_time_t *state_time);
int aread_quit (aread_t *aread, u_time_t *quit_time);
int aread_error_quit (aread_t *aread, u_time_t *quit_time);
aread_t* aread_init (
    int rate,
    int channels,
    int bits_per_sample,
    int sync_system_time,
    const char *device_name, const char *mixer_name, const char *source_name);

#endif /* AREAD_H */

