/* csp.c */

#include <stdio.h>
#include <inttypes.h>

#include "csp.h"


static CSP_INFO csp_info[] = {
  { CSP_UNKNOWN, "UNKNOWN",  0, 0, {0,0,0}, {0,0,0}, {0,0,0}, NULL },
//  { CSP_RGB32,   "RGB32",   32, 1, {8,8,8}, {1,0,0}, {1,0,0}, "RGB" },
  { CSP_ARGB32,  "ARGB32",  32, 1, {8,8,8}, {1,0,0}, {1,0,0}, "ARGB" },
  { CSP_RGB24,   "RGB24",   24, 1, {8,8,8}, {1,0,0}, {1,0,0}, "RGB" },
  { CSP_RGB565,  "RGB565",  16, 1, {5,6,5}, {1,0,0}, {1,0,0}, "RGB" },
  { CSP_RGB555,  "RGB555",  16, 1, {5,5,5}, {1,0,0}, {1,0,0}, "RGB" },

  { CSP_BGRA32,  "BGRA32",  32, 1, {8,8,8}, {1,0,0}, {1,0,0}, "BGRA" },
  { CSP_BGR24,   "BGR24",   24, 1, {8,8,8}, {1,0,0}, {1,0,0}, "BGR" },
//  { CSP_BGR565,  "BGR565",  16, 1, {5,6,5}, {1,0,0}, {1,0,0}, "BGR" },
//  { CSP_BGR555,  "BGR555",  16, 1, {5,5,5}, {1,0,0}, {1,0,0}, "BGR" },

  { CSP_YUV444, "YUV444", 24, 3, {8,8,8}, {1,1,1}, {1,1,1}, "YUV" },
  { CSP_YUV422, "YUV422", 16, 3, {8,8,8}, {1,2,2}, {1,1,1}, "YUV" },
  { CSP_YUV420, "YUV420", 12, 3, {8,8,8}, {1,2,2}, {1,2,2}, "YUV" },
  { CSP_YUV422PACK, "YUV422PACK", 16, 1, {8,8,8}, {1,0,0}, {1,0,0}, "YUYV" },
  { CSP_YUYV,   "YUYV",   16, 1, {8,8,8}, {1,0,0}, {1,0,0}, "YUYV" },
  { CSP_YUY2,   "YUY2",   16, 1, {8,8,8}, {1,0,0}, {1,0,0}, "YUYV" },
  { CSP_UYVY,   "UYVY",   16, 1, {8,8,8}, {1,0,0}, {1,0,0}, "UYVY" },
  { CSP_I420,   "I420",   12, 3, {8,8,8}, {1,2,2}, {1,2,2}, "YUV" },
  { CSP_YV12,   "YV12",   12, 3, {8,8,8}, {1,2,2}, {1,2,2}, "YVU" },

  { CSP_COMPRESSED, "COMPRESSED", 0, 1, {0,0,0}, {0,0,0}, {0,0,0}, NULL },
  { CSP_XVID,       "XVID",       0, 1, {0,0,0}, {0,0,0}, {0,0,0}, NULL },
  { CSP_RTJPEG,     "RTJPEG",     0, 1, {0,0,0}, {0,0,0}, {0,0,0}, NULL },

#if 0
#ifdef HAVE_LIBXVIDCORE
  { CSP_XVID,               "CSP_XVID",               -1, CSP_UNKNOWN  },
#endif /* HAVE_LIBXVIDCORE */
#ifdef HAVE_RTJPEG
  { CSP_RTJPEG,             "CSP_RTJPEG",             -1, CSP_UNKNOWN  },
#endif /* HAVE_RTJPEG */
#endif
};

static int csp_info_num = sizeof(csp_info) / sizeof(CSP_INFO);

int
csp_calc_buf_size(int csp, int width, int height, int horz_border, int vert_border, int size_ret[], int offset_ret[], int stride_ret[])
{
  CSP_INFO *info, ti;
  int i;
  int W, H;
  int offset, stride, plane_size;
  int buf_size;

  if ((width & 0x7) || (height & 0x7)) {
    fprintf(stderr, "csp_calc_buf_size:ERROR: width and height must be multiple of 8, width = %d, height = %d.\n", width, height);
    return CSP_FAIL;
  }

  if ((horz_border & 0x7) || (vert_border & 0x7)) {
    fprintf(stderr, "csp_calc_buf_size:ERROR: horz_border and vert_border must be multiple of 8, horz_border = %d, vert_border = %d.\n", horz_border, vert_border);
    return CSP_FAIL;
  }

  info = &ti;
  if (csp_get_info(csp, info) != CSP_OK) {
    return CSP_FAIL;
  }

  W = width + horz_border*2;
  H = height + vert_border*2;

  if (info->num_planes == 0 || info->bits_per_pixel == 0) {
    fprintf(stderr, "csp_calc_buf_size: cannot calculate \"%s\".\n",info->name);
    return CSP_FAIL;
  }
  else if (info->num_planes == 1) {
    stride = W;
    if (stride_ret) {
      stride_ret[0] = stride;
      stride_ret[1] = stride_ret[2] = 0;
    }
    offset = stride * vert_border + horz_border;
    if (offset_ret) {
      offset_ret[0] = offset;
      offset_ret[1] = offset_ret[2] = 0;
    }
    buf_size = stride * H * info->bits_per_pixel / 8;
    if (size_ret) {
      size_ret[0] = buf_size;
      size_ret[1] = size_ret[2] = 0;
    }
  } else {
    buf_size = 0;
    for (i = 0; i < 3; i++) {
      stride = W * info->sample_bits[i] / (8*info->horz_period[i]);
      if (stride_ret)
	stride_ret[i] = stride;

      offset = stride * (vert_border/info->vert_period[i])
        	+ (horz_border/info->horz_period[i]);
      if (offset_ret)
	offset_ret[i] = offset;

      plane_size = stride * H / (info->vert_period[i]);
      if (size_ret)
        size_ret[i] = plane_size;

      buf_size += plane_size;
    }
  }
  return buf_size;
}

int
csp_calc_buf_size_align(int csp, int width, int height, int horz_border, int vert_border, int size_ret[], int offset_ret[], int stride_ret[])
{
  CSP_INFO *info, ti;
  int i;
  int W, H;
  int offset, stride, align, plane_size;
  int buf_size;

  if ((width & 0x7) || (height & 0x7)) {
    fprintf(stderr, "csp_calc_buf_size:ERROR: width and height must be multiple of 8, width = %d, height = %d.\n", width, height);
    return CSP_FAIL;
  }

  if ((horz_border & 0x7) || (vert_border & 0x7)) {
    fprintf(stderr, "csp_calc_buf_size:ERROR: horz_border and vert_border must be multiple of 8, horz_border = %d, vert_border = %d.\n", horz_border, vert_border);
    return CSP_FAIL;
  }

  info = &ti;
  if (csp_get_info(csp, info) != CSP_OK) {
    return CSP_FAIL;
  }

  W = width + horz_border*2;
  H = height + vert_border*2;

  if (info->num_planes == 0 || info->bits_per_pixel == 0) {
    fprintf(stderr, "csp_calc_buf_size: cannot calculate \"%s\".\n",info->name);
    return CSP_FAIL;
  }
  else if (info->num_planes == 1) {
    stride = W;
    if (stride_ret) {
      stride_ret[0] = stride;
      stride_ret[1] = stride_ret[2] = 0;
    }
    offset = stride * vert_border + horz_border;
    align = ((offset+0xf) & ~0xf) - offset;
    offset += align;
    if (offset_ret) {
      offset_ret[0] = offset;
      offset_ret[1] = offset_ret[2] = 0;
    }
    buf_size = stride * H * info->bits_per_pixel / 8;
    buf_size += align;
    align = ((buf_size+0xf) & ~0xf) - buf_size;
    buf_size += align;
    if (size_ret) {
      size_ret[0] = buf_size;
      size_ret[1] = size_ret[2] = 0;
    }
  } else {
    buf_size = 0;
    for (i = 0; i < 3; i++) {
      stride = W * info->sample_bits[i] / (8*info->horz_period[i]);
      if (stride_ret)
	stride_ret[i] = stride;

      offset = stride * (vert_border/info->vert_period[i])
        	+ (horz_border/info->horz_period[i]);
      align = ((offset+0xf) & ~0xf) - offset;
      offset += align;
      if (offset_ret)
	offset_ret[i] = offset;

      plane_size = stride * H / (info->vert_period[i]);
      plane_size += align;
      align = ((plane_size+0xf) & ~0xf) - plane_size;
      plane_size += align;
      if (size_ret)
        size_ret[i] = plane_size;

      buf_size += plane_size;
    }
  }
  return buf_size;
}

int
csp_to_bits_per_pixel(int csp)
{
  int i;
  for (i = 0; i < csp_info_num; i++) {
    if (csp_info[i].csp == csp)
      return csp_info[i].bits_per_pixel;
  }
  fprintf(stderr, "csp_to_bits_per_pixel: unknown csp, %d.\n", csp);
  return CSP_FAIL;
}

const char *
csp_to_name(int csp)
{
  int i;
  for (i = 0; i < csp_info_num; i++) {
    if (csp_info[i].csp == csp)
      return csp_info[i].name;
  }
  fprintf(stderr, "csp_to_name: unknown csp, %d.\n", csp);
  return NULL;
}

int
csp_get_info(int csp, CSP_INFO *info_ret)
{
  int i;
  *info_ret = csp_info[0];
  for (i = 0; i < csp_info_num; i++) {
    if (csp_info[i].csp == csp) {
      *info_ret = csp_info[i];
      return CSP_OK;
    }
  }
  fprintf(stderr, "csp_get_info: unknown csp, %d.\n", csp);
  return CSP_FAIL;
}

