/* csp.h */

#ifndef CSP_H
#define CSP_H

#include "csp_conv.h"


#define CSP_UNKNOWN     0       /* unknown csp */

#define CSP_RGB32       1       /* 32bit RGB */ 
#define CSP_RGB24       2       /* 24bit RGB */
#define CSP_RGB565      3       /* 565 16 bit RGB */
#define CSP_RGB555      4       /* 555 15bit RGB */

#define CSP_ARGB32      5       /* 32bit RGB */ 
#define CSP_BGRA32      6       /* 32bit RGB */ 
#define CSP_BGR24       7       /* 24bit BGR */
#define CSP_BGR565      8       /* 565 16 bit RGB */
#define CSP_BGR555      9       /* 555 15bit RGB */


#define CSP_YUV444      21      /* YUV 4:4:4 Planar, "YUV" */
#define CSP_YUV422      22      /* YUV 4:2:2 Planar, "YUV" */
#define CSP_YUV422PACK  23      /* YUV 4:2:2 packed */
#define CSP_YUYV        24      /* YUV 4:2:2 packed, "YUYV" */
#define CSP_YUY2        25      /* YUV 4:2:2 packed, "YUYV" */
#define CSP_UYVY        26      /* YUV 4:2:2 packed, "UYVY" */
#define CSP_YUV420      27      /* YUV 4:2:0 Planar, "YUV" */
#define CSP_YV12        28      /* YUV 4:2:0 Planar, "YVU" */
#define CSP_I420        29      /* YUV 4:2:0 Planar, "YUV" */

#define CSP_COMPRESSED  100
#define CSP_XVID        101
#define CSP_RTJPEG      102

#define Y_LO_LIMIT 16
#define Y_HI_LIMIT 235
#define C_LO_LIMIT 16
#define C_HI_LIMIT 235

#define CSP_OK   0
#define CSP_FAIL -1

typedef struct {
  int csp;
  const char *name;
  int bits_per_pixel;
  int num_planes;
  int sample_bits[3];
  int horz_period[3];
  int vert_period[3];
  const char *order;
} CSP_INFO;

int csp_to_bits_per_pixel(int csp);
const char *csp_to_name(int csp);
int csp_get_info(int csp, CSP_INFO *info_ret);
int csp_calc_buf_size(int csp,
    int width, int height, int horz_border, int vert_border,
    int size_ret[], int offset_ret[], int stride_ret[]);

#endif /* CSP_H */

