/*
 * rgb_to_rgb.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "csp_conv_rgb_to_rgb.h"


#ifdef ARCH_X86
#define BSWAP32(v) asm ("bswap  %0": "=r" (v) : "0" (v))
#define Bswap32(p1, p2) \
  asm ("movl %0, %%eax\n\t" \
     "shl  $8, %%eax\n\t" \
     "bswap %%eax\n\t" \
     "movl %%eax, %1\n\t" \
     : : "m" (*(p1)), "m" (*(p2)) \
     : "%eax")
#else
#define BSWAP32(v) \
      (v) =  (  (((v)&0xff)<<24) \
              | (((v)&0xff00)<<8) \
	      | (((v)&0xff000000)>>24) \
	      | (((v)&0xff0000)>>8) )
#endif

/* RGB24 */

void csp_rgb24_to_argb32 (uint8_t *rgb24[], uint8_t *argb32[],
	  int rgb24_stride[], int argb32_stride[], int width, int height)
{
  uint32_t *p32;
  uint8_t *p24;
  int w;
  int next24, next32;

  p32 = (uint32_t *) argb32[0];
  p24 = rgb24[0];

  next24 = (rgb24_stride[0] - width) * 3;
  next32 = argb32_stride[0] - width;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
#if 0
      ((uint8_t*)p32)[0] = 0;         /* A */
      ((uint8_t*)p32)[1] = p24[0];    /* R */
      ((uint8_t*)p32)[2] = p24[1];    /* G */
      ((uint8_t*)p32)[3] = p24[2];    /* B */
#endif
      p32[0] = (*((uint32_t*)p24) << 8);
      p32[1] = (*((uint32_t*)(p24+1*3-1)) & 0xFFFFFF00);
      p32[2] = (*((uint32_t*)(p24+2*3-1)) & 0xFFFFFF00);
      p32[3] = (*((uint32_t*)(p24+3*3-1)) & 0xFFFFFF00);
      p32[4] = (*((uint32_t*)(p24+4*3-1)) & 0xFFFFFF00);
      p32[5] = (*((uint32_t*)(p24+5*3-1)) & 0xFFFFFF00);
      p32[6] = (*((uint32_t*)(p24+6*3-1)) & 0xFFFFFF00);
      p32[7] = (*((uint32_t*)(p24+7*3-1)) & 0xFFFFFF00);

      p32+=8;
      p24 += 3*8;
    }
    p24 += next24;
    p32 += next32;
  }
}

void csp_rgb24_to_bgra32 (uint8_t *rgb24[], uint8_t *bgra32[],
	  int rgb24_stride[], int bgra32_stride[], int width, int height)
{
  uint32_t *p32;
  uint8_t *p24;
  int w;
  int next24, next32;

  p32 = (uint32_t *) bgra32[0];
  p24 = rgb24[0];

  next24 = (rgb24_stride[0] - width) * 3;
  next32 = bgra32_stride[0] - width;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
#if 0
      ((uint8_t*)p32)[0] = p24[2];   /* B */
      ((uint8_t*)p32)[1] = p24[1];   /* G */
      ((uint8_t*)p32)[2] = p24[0];   /* R */
      ((uint8_t*)p32)[3] = 0;        /* A */
#endif
      register int v;

#if 0
      v = *((uint32_t*)p24) << 8;
      BSWAP32(v);
      *p32 = v;

      v = *((uint32_t*)(p24+1*3)) << 8;
      BSWAP32(v);
      *(p32+1) = v;
      v = *((uint32_t*)(p24+2*3)) << 8;
      BSWAP32(v);
      *(p32+2) = v;
      v = *((uint32_t*)(p24+3*3)) << 8;
      BSWAP32(v);
      *(p32+3) = v;
      v = *((uint32_t*)(p24+4*3)) << 8;
      BSWAP32(v);
      *(p32+4) = v;
      v = *((uint32_t*)(p24+5*3)) << 8;
      BSWAP32(v);
      *(p32+5) = v;
      v = *((uint32_t*)(p24+6*3)) << 8;
      BSWAP32(v);
      *(p32+6) = v;
      v = *((uint32_t*)(p24+7*3)) << 8;
      BSWAP32(v);
      *(p32+7) = v;
#else
      v = *((uint32_t*)p24) << 8; BSWAP32(v); p32[0] = v;
      v = *((uint32_t*)(p24+1*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[1] = v;
      v = *((uint32_t*)(p24+2*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[2] = v;
      v = *((uint32_t*)(p24+3*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[3] = v;
      v = *((uint32_t*)(p24+4*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[4] = v;
      v = *((uint32_t*)(p24+5*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[5] = v;
      v = *((uint32_t*)(p24+6*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[6] = v;
      v = *((uint32_t*)(p24+7*3-1)) & 0xFFFFFF00; BSWAP32(v); p32[7] = v;
#endif

      p32+=8;
      p24+=3*8;
    }
    p24 += next24;
    p32 += next32;
  }
}

void csp_bgr24_to_argb32 (uint8_t *bgr24[], uint8_t *argb32[],
	  int bgr24_stride[], int argb32_stride[], int width, int height)
{
  uint32_t *p32;
  uint8_t *p24;
  int w;
  int next24, next32;

  p32 = (uint32_t *) argb32[0];
  p24 = bgr24[0];

  next24 = (bgr24_stride[0] - width) * 3;
  next32 = argb32_stride[0] - width;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
#if 0
      ((uint8_t*)p32)[0] = 0;         /* A */
      ((uint8_t*)p32)[1] = p24[2];    /* R */
      ((uint8_t*)p32)[2] = p24[1];    /* G */
      ((uint8_t*)p32)[3] = p24[0];    /* B */
#endif
      register int v;

#if 0
      v = *((uint32_t*)p24); BSWAP32(v); p32[0] = v;
      v = *((uint32_t*)(p24+1*3)); BSWAP32(v); p32[1] = v;
      v = *((uint32_t*)(p24+2*3)); BSWAP32(v); p32[2] = v;
      v = *((uint32_t*)(p24+3*3)); BSWAP32(v); p32[3] = v;
      v = *((uint32_t*)(p24+4*3)); BSWAP32(v); p32[4] = v;
      v = *((uint32_t*)(p24+5*3)); BSWAP32(v); p32[5] = v;
      v = *((uint32_t*)(p24+6*3)); BSWAP32(v); p32[6] = v;
      v = *((uint32_t*)(p24+7*3)); BSWAP32(v); p32[7] = v;
#else
      v = (*((uint32_t*)p24) & 0xFFFFFF); BSWAP32(v); p32[0] = v;
      v = (*((uint32_t*)(p24+1*3-1))>>8); BSWAP32(v); p32[1] = v;
      v = (*((uint32_t*)(p24+2*3-1))>>8); BSWAP32(v); p32[2] = v;
      v = (*((uint32_t*)(p24+3*3-1))>>8); BSWAP32(v); p32[3] = v;
      v = (*((uint32_t*)(p24+4*3-1))>>8); BSWAP32(v); p32[4] = v;
      v = (*((uint32_t*)(p24+5*3-1))>>8); BSWAP32(v); p32[5] = v;
      v = (*((uint32_t*)(p24+6*3-1))>>8); BSWAP32(v); p32[6] = v;
      v = (*((uint32_t*)(p24+7*3-1))>>8); BSWAP32(v); p32[7] = v;
#endif

      p32+=8;
      p24+=3*8;
    }
    p24 += next24;
    p32 += next32;
  }
}

void csp_bgr24_to_bgra32 (uint8_t *bgr24[], uint8_t *bgra32[],
	  int bgr24_stride[], int bgra32_stride[], int width, int height)
{
  uint32_t *p32;
  uint8_t *p24;
  int w;
  int next24, next32;

  p32 = (uint32_t *) bgra32[0];
  p24 = bgr24[0];

  next24 = (bgr24_stride[0] - width) * 3;
  next32 = bgra32_stride[0] - width;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
#if 0
      ((uint8_t*)p32)[0] = p24[0];    /* B */
      ((uint8_t*)p32)[1] = p24[1];    /* G */
      ((uint8_t*)p32)[2] = p24[2];    /* R */
      ((uint8_t*)p32)[3] = 0;         /* A */
#endif
#if 0
      p32[0] = *((uint32_t*)p24);
      p32[1] = *((uint32_t*)(p24+1*3));
      p32[2] = *((uint32_t*)(p24+2*3));
      p32[3] = *((uint32_t*)(p24+3*3));
      p32[4] = *((uint32_t*)(p24+4*3));
      p32[5] = *((uint32_t*)(p24+5*3));
      p32[6] = *((uint32_t*)(p24+6*3));
      p32[7] = *((uint32_t*)(p24+7*3));
#else
#if 0
      p32[0] = (*((uint32_t*)p24) & 0xFFFFFF);
      p32[1] = (*((uint32_t*)(p24+(1*3-1))) >> 8);
      p32[2] = (*((uint32_t*)(p24+(2*3-1))) >> 8);
      p32[3] = (*((uint32_t*)(p24+(3*3-1))) >> 8);
      p32[4] = (*((uint32_t*)(p24+(4*3-1))) >> 8);
      p32[5] = (*((uint32_t*)(p24+(5*3-1))) >> 8);
      p32[6] = (*((uint32_t*)(p24+(6*3-1))) >> 8);
      p32[7] = (*((uint32_t*)(p24+(7*3-1))) >> 8);
#endif
      p32[0] = (*((uint32_t*)p24));
      p32[1] = (*((uint32_t*)(p24+(1*3))));
      p32[2] = (*((uint32_t*)(p24+(2*3))));
      p32[3] = (*((uint32_t*)(p24+(3*3))));
      p32[4] = (*((uint32_t*)(p24+(4*3))));
      p32[5] = (*((uint32_t*)(p24+(5*3))));
      p32[6] = (*((uint32_t*)(p24+(6*3))));
      p32[7] = (*((uint32_t*)(p24+(7*3-1))) >> 8);
#endif
      p32+=8;
      p24 += 3*8;
    }
    p24 += next24;
    p32 += next32;
  }
}

void csp_rgb24_to_rgb565 (uint8_t *rgb24[], uint8_t *rgb565[],
	  int rgb24_stride[], int rgb565_stride[], int width, int height)
{
  uint16_t *p16;
  uint8_t *p24;
  int w;
  int next24, next16;

  p16 = (uint16_t *) rgb565[0];
  p24 = rgb24[0];

  next24 = (rgb24_stride[0] - width) * 3;
  next16 = (rgb565_stride[0] - width) * 2;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
      register int v;
#if 0
      v = *((uint32_t*)p24);
      p16[0] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3));
      p16[1] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*2));
      p16[2] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*3));
      p16[3] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*4));
      p16[4] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*5));
      p16[5] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*6));
      p16[6] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
      v = *((uint32_t*)(p24+3*7));
      p16[7] = ((v&0xf8)<<8) | ((v&0xfc)<<3) | (v>>3);
#endif
      v = *((uint32_t*)p24);
      p16[0] = ((v&0xf8)<<8) | ((v&0xfc00)>>5) | ((v&0xf80000)>>19);
      v = *((uint32_t*)(p24+3-1));
      p16[1] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*2-1));
      p16[2] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*3-1));
      p16[3] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*4-1));
      p16[4] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*5-1));
      p16[5] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*6-1));
      p16[6] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*7-1));
      p16[7] = (v&0xf800) | ((v&0xfc0000)>>13) | ((v&0xf8000000)>>27);

      p16+=8;
      p24+=3*8;
    }
    p24 += next24;
    p16 += next16;
  }
}

void csp_rgb24_to_rgb555 (uint8_t *rgb24[], uint8_t *rgb555[],
	  int rgb24_stride[], int rgb555_stride[], int width, int height)
{
  uint16_t *p16;
  uint8_t *p24;
  int w;
  int next24, next16;

  p16 = (uint16_t *) rgb555[0];
  p24 = rgb24[0];

  next24 = (rgb24_stride[0] - width) * 3;
  next16 = (rgb555_stride[0] - width) * 2;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
      register int v;

      v = *((uint32_t*)p24);
      p16[0] = ((v&0xf8)<<7) | ((v&0xf800)>>6) | ((v&0xf80000)>>19);
      v = *((uint32_t*)(p24+3-1));
      p16[1] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*2-1));
      p16[2] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*3-1));
      p16[3] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*4-1));
      p16[4] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*5-1));
      p16[5] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*6-1));
      p16[6] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);
      v = *((uint32_t*)(p24+3*7-1));
      p16[7] = ((v&0xf800)>>1) | ((v&0xf80000)>>14) | ((v&0xf8000000)>>27);

      p16+=8;
      p24+=3*8;
    }
    p24 += next24;
    p16 += next16;
  }
}

static void csp_rgb24_cpy (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *src, *dst;
  int next_src, next_dst;

  src = src_rgb[0];
  dst = dst_rgb[0];

  next_src = src_stride[0] * 3;
  next_dst = dst_stride[0] * 3;
  width *= 3;

  for (; height > 0; height--)
  {
    memcpy(dst, src, width);
    src += next_src;
    dst += next_dst;
  }
}

static void csp_rgb24_swap (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *src, *dst;
  int w;
  int next_src, next_dst;

  src = src_rgb[0];
  dst = dst_rgb[0];

  next_src = (src_stride[0] - width) * 3;
  next_dst = (dst_stride[0] - width) * 3;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
      dst[0] = src[2]; dst[1] = src[1]; dst[2] = src[0];
      dst[0+3*1] = src[2+3*1]; dst[1+3*1] = src[1+3*1]; dst[2+3*1] = src[0+3*1];
      dst[0+3*2] = src[2+3*2]; dst[1+3*2] = src[1+3*2]; dst[2+3*2] = src[0+3*2];
      dst[0+3*3] = src[2+3*3]; dst[1+3*3] = src[1+3*3]; dst[2+3*3] = src[0+3*3];
      dst[0+3*4] = src[2+3*4]; dst[1+3*4] = src[1+3*4]; dst[2+3*4] = src[0+3*4];
      dst[0+3*5] = src[2+3*5]; dst[1+3*5] = src[1+3*5]; dst[2+3*5] = src[0+3*5];
      dst[0+3*6] = src[2+3*6]; dst[1+3*6] = src[1+3*6]; dst[2+3*6] = src[0+3*6];
      dst[0+3*7] = src[2+3*7]; dst[1+3*7] = src[1+3*7]; dst[2+3*7] = src[0+3*7];
      dst+=3*8;
      src+=3*8;
    }
    src += next_src;
    dst += next_dst;
  }
}

void csp_rgb24_to_rgb24 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb24_cpy(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_rgb24_to_bgr24 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb24_swap(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_bgr24_to_bgr24 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb24_cpy(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_bgr24_to_rgb24 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb24_swap(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}


/* RGB32 */

static void csp_rgb32_swap (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint32_t *src, *dst;
  int w;
  int next_src, next_dst;

  src = (uint32_t*)src_rgb[0];
  dst = (uint32_t*)dst_rgb[0];

  next_src = src_stride[0] - width;
  next_dst = dst_stride[0] - width;

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8)
    {
#if 0
      ((uint8_t*)dst)[0] = ((uint8_t*)src)[3];
      ((uint8_t*)dst)[1] = ((uint8_t*)src)[2];
      ((uint8_t*)dst)[2] = ((uint8_t*)src)[1];
      ((uint8_t*)dst)[3] = ((uint8_t*)src)[0];
#endif
      register int v;

      v = src[0]; BSWAP32(v); dst[0] = v;

      v = src[1]; BSWAP32(v); dst[1] = v;
      v = src[2]; BSWAP32(v); dst[2] = v;
      v = src[3]; BSWAP32(v); dst[3] = v;
      v = src[4]; BSWAP32(v); dst[4] = v;
      v = src[5]; BSWAP32(v); dst[5] = v;
      v = src[6]; BSWAP32(v); dst[6] = v;
      v = src[7]; BSWAP32(v); dst[7] = v;

      dst+=8;
      src+=8;
    }
    src += next_src;
    dst += next_dst;
  }
}

#if 0
#include "attributes.h"
#include "mmx.h"
static void csp_rgb32_cpy (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint32_t *src, *dst;
  int next_src, next_dst, w;

  src = (uint32_t*)src_rgb[0];
  dst = (uint32_t*)dst_rgb[0];

  next_src = (src_stride[0] - width);
  next_dst = (dst_stride[0] - width);

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w+=8) {
#if 0
      movq_m2r(src[0], mm0);
      movq_r2m(mm0, dst[0]);
      movq_m2r(src[2], mm0);
      movq_r2m(mm0, dst[2]);
      movq_m2r(src[4], mm0);
      movq_r2m(mm0, dst[4]);
      movq_m2r(src[6], mm0);
      movq_r2m(mm0, dst[6]);
#endif
      movq_m2r(src[0], mm0);
      movntq_r2m(mm0, dst[0]);
      movq_m2r(src[2], mm0);
      movntq_r2m(mm0, dst[2]);
      movq_m2r(src[4], mm0);
      movntq_r2m(mm0, dst[4]);
      movq_m2r(src[6], mm0);
      movntq_r2m(mm0, dst[6]);
      src += 8;
      dst += 8;
    }
    src += next_src;
    dst += next_dst;
  }
}
#else
static void csp_rgb32_cpy (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint32_t *src, *dst;
  int next_src, next_dst;

  src = (uint32_t*)src_rgb[0];
  dst = (uint32_t*)dst_rgb[0];

  next_src = src_stride[0];
  next_dst = dst_stride[0];
  width *= 4;

  for (; height > 0; height--)
  {
    memcpy(dst, src, width);
    src += next_src;
    dst += next_dst;
  }
}
#endif

void csp_argb32_to_bgra32 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb32_swap(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_bgra32_to_argb32 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb32_swap(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_argb32_to_argb32 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb32_cpy(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

void csp_bgra32_to_bgra32 (uint8_t *src_rgb[], uint8_t *dst_rgb[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_rgb32_cpy(src_rgb, dst_rgb, src_stride, dst_stride, width, height);
}

