/*
 * rgb_to_yuv420.c
 *
 * code from mpeg2dec <http://libmpeg2.sourceforge.net>
 *
 */

/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "csp_conv.h"
#include "csp_conv_rgb_to_yuv.h"
#include "rgb_to_yuv.h"

void csp_rgb_to_yuv420_init(void)
{
  rgb_to_yuv_table_init();
}

void csp_yuv_to_rgb24_quit(void)
{
  return;
}

void csp_rgb24_to_yuv420(uint8_t *rgb24[], uint8_t *yuv_planes[],
	  int rgb_stride[], int yuv_stride[], int width, int height)
{
  int i, j;
  int r, g, b;
  int u4, v4;
  int y_stride, u_stride, v_stride;
  unsigned char *yp0, *yp1, *up, *vp;
  uint8_t *rgb0, *rgb1;

  rgb0 = rgb24[0];
  rgb1 = rgb0 + rgb_stride[0];

  rgb_std = (rgb_stride[0]*2) - width;
  rgb_std *= 3;
  y_stride = (yuv_stride[0]*2) - width;
  u_stride = yuv_stride[1];
  v_stride = yuv_stride[2];

  yp0 = yuv_planes[0];
  yp1 = yp0 + y_stride;
  up = yuv_planes[1];
  vp = yuv_planes[2];

  for (j=0; j<height; j+=2)
  {
    for (i=0; i<width; i+=2)
    {
      r = *rgb0; g = *(rgb0+1); b = *(rgb0+2);
      *yp0 = ( rgb_to_yuv_Y_R[r] + rgb_to_yuv_Y_G[g]
	     + rgb_to_yuv_Y_B[b]) >>FP_BITS;
      u4 = ( rgb_to_yuv_U_R[r] + rgb_to_yuv_U_G[g]
	   + rgb_to_yuv_U_B[b]) >>FP_BITS;
      v4 = ( rgb_to_yuv_V_R[r] + rgb_to_yuv_V_G[g]
	   + rgb_to_yuv_V_B[b]) >>FP_BITS;
      rgb0 += 3;

      r = *rgb0; g = *(rgb0+1); b = *(rgb0+2);
      *(yp0+1) = ( rgb_to_yuv_Y_R[r] + rgb_to_yuv_Y_G[g]
	         + rgb_to_yuv_Y_B[b]) >>FP_BITS;
      u4 += ( rgb_to_yuv_U_R[r] + rgb_to_yuv_U_G[g]
	   + rgb_to_yuv_U_B[b]) >>FP_BITS;
      v4 += ( rgb_to_yuv_V_R[r] + rgb_to_yuv_V_G[g]
	   + rgb_to_yuv_V_B[b]) >>FP_BITS;
      rgb0 += 3;
      yp0 += 2;

      r = *rgb1; g = *(rgb1+1); b = *(rgb1+2);
      *yp1 = ( rgb_to_yuv_Y_R[r] + rgb_to_yuv_Y_G[g]
	   + rgb_to_yuv_Y_B[b]) >>FP_BITS;
      u4 = ( rgb_to_yuv_U_R[r] + rgb_to_yuv_U_G[g]
	   + rgb_to_yuv_U_B[b]) >>FP_BITS;
      v4 = ( rgb_to_yuv_V_R[r] + rgb_to_yuv_V_G[g]
	   + rgb_to_yuv_V_B[b]) >>FP_BITS;
      rgb1 += 3;

      r = *rgb1; g = *(rgb1+1); b = *(rgb1+2);
      *(yp1+1) = ( rgb_to_yuv_Y_R[r] + rgb_to_yuv_Y_G[g]
	         + rgb_to_yuv_Y_B[b]) >>FP_BITS;
      u4 += ( rgb_to_yuv_U_R[r] + rgb_to_yuv_U_G[g]
	   + rgb_to_yuv_U_B[b]) >>FP_BITS;
      v4 += ( rgb_to_yuv_V_R[r] + rgb_to_yuv_V_G[g]
	   + rgb_to_yuv_V_B[b]) >>FP_BITS;
      rgb1 += 3;
      yp1 += 2;

      *up = (u4+2) >> 2;
      *vp = (v4+2) >> 2;
      up++; vp++;
    }
    yp0 += y_stride;
    yp1 += y_stride;
    up += u_stride;
    vp += v_stride;
    rgb0 += rgb_std;
    rgb1 += rgb_std;
  }
}

