/*
 * yuv2rgb_mmx.c
 * Copyright (C) 2000-2002 Silicon Integrated System Corp.
 * All Rights Reserved.
 *
 * Author: Olie Lho <ollie@sis.com.tw>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#ifdef ARCH_X86

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "csp_conv_yuv_to_rgb.h"

#include "attributes.h"
#include "mmx.h"

#include "yuv_to_rgb_mmx.h"

static inline void mmxext_unpack_32argb (uint8_t * image)
{
    /*
     * convert RGB plane to RGB packed format,
     * mm0 -> B, mm1 -> R, mm2 -> G, mm3 -> 0,
     * mm4 -> GB, mm5 -> AR pixel 4-7,
     * mm6 -> GB, mm7 -> AR pixel 0-3
     */

    pxor_r2r (mm3, mm3);               /* A7 A6 A5 A4 A3 A2 A1 A0 */
    movq_r2r (mm2, mm4);               /* G7 G6 G5 G4 G3 G2 G1 G0 */
    punpcklbw_r2r (mm0, mm4);          /* B3 G3 B2 G2 B1 G1 B0 G0 */
    punpcklbw_r2r (mm1, mm3);          /* R3 A3 R2 A2 R1 A1 R0 A0 */
    movq_r2r (mm3, mm5);               /* R3 A3 R2 A2 R1 A1 R0 A0 */
    punpcklwd_r2r (mm4, mm5);          /* B1 G1 R1 A1 B0 G0 R0 A0 */
    movntq_r2m (mm5, *image);
    punpckhwd_r2r (mm4, mm3);          /* B3 G3 R3 A3 B2 G2 R2 A2 */
    movntq_r2m (mm3, *(image+8));

    pxor_r2r (mm3, mm3);
    movq_r2r (mm2, mm4);               /* B7 B6 B5 B4 B3 B2 B1 B0 */
    punpckhbw_r2r (mm0, mm4);          /* B7 G7 B6 G6 B5 G5 B4 G4 */
    punpckhbw_r2r (mm1, mm3);          /* R7 A7 R6 A6 R5 A5 R4 A4 */
    movq_r2r (mm3, mm5);
    punpcklwd_r2r (mm4, mm5);          /* B5 G5 R5 A5 B4 G4 R4 A4 */
    movntq_r2m (mm5, *(image+16));
    punpckhwd_r2r (mm4, mm3);          /* B7 G7 R7 A7 B6 G6 R6 A6 */
    movntq_r2m (mm3, *(image+24));
}

static inline void mmx_unpack_32argb (uint8_t * image)
{
    /*
     * convert RGB plane to RGB packed format,
     * mm0 -> B, mm1 -> R, mm2 -> G, mm3 -> 0,
     * mm4 -> GB, mm5 -> AR pixel 4-7,
     * mm6 -> GB, mm7 -> AR pixel 0-3
     */

    pxor_r2r (mm3, mm3);               /* A7 A6 A5 A4 A3 A2 A1 A0 */
    movq_r2r (mm2, mm4);               /* G7 G6 G5 G4 G3 G2 G1 G0 */
    punpcklbw_r2r (mm0, mm4);          /* B3 G3 B2 G2 B1 G1 B0 G0 */
    punpcklbw_r2r (mm1, mm3);          /* R3 A3 R2 A2 R1 A1 R0 A0 */
    movq_r2r (mm3, mm5);               /* R3 A3 R2 A2 R1 A1 R0 A0 */
    punpcklwd_r2r (mm4, mm5);          /* B1 G1 R1 A1 B0 G0 R0 A0 */
    movq_r2m (mm5, *image);
    punpckhwd_r2r (mm4, mm3);          /* B3 G3 R3 A3 B2 G2 R2 A2 */
    movq_r2m (mm3, *(image+8));

    pxor_r2r (mm3, mm3);
    movq_r2r (mm2, mm4);               /* B7 B6 B5 B4 B3 B2 B1 B0 */
    punpckhbw_r2r (mm0, mm4);          /* B7 G7 B6 G6 B5 G5 B4 G4 */
    punpckhbw_r2r (mm1, mm3);          /* R7 A7 R6 A6 R5 A5 R4 A4 */
    movq_r2r (mm3, mm5);
    punpcklwd_r2r (mm4, mm5);          /* B5 G5 R5 A5 B4 G4 R4 A4 */
    movq_r2m (mm5, *(image+16));
    punpckhwd_r2r (mm4, mm3);          /* B7 G7 R7 A7 B6 G6 R6 A6 */
    movq_r2m (mm3, *(image+24));
}

void csp_yuv420_to_argb32_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

//    printf("csp_yuv420_to_argb32: rgb_stride %d, yuv_stride %d, %d, %d, width %d, height %d\n", rgb_stride[0], yuv_stride[0], yuv_stride[1], yuv_stride[2], width, height);
    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
//    rgb_std  = rgb_stride[0] - 4 * width;
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmxext_unpack_32argb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
	if (height & 1) {
	    pu += u_stride;
	    pv += v_stride;
	} else {
	    pu -= 4 * width;
	    pv -= 4 * width;
	}
    } while (--height);
    emms();
}


void csp_yuv422_to_argb32_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
//    rgb_std  = rgb_stride[0] - 4 * width;
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmxext_unpack_32argb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

void csp_yuv444_to_argb32_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - width;
    v_stride = yuv_stride[2] - width;
//    rgb_std  = rgb_stride[0] - 4 * width;
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv44_to_rgb (py, pu, pv);
	    mmxext_unpack_32argb (image);
	    py += 8;
	    pu += 8;
	    pv += 8;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}


void csp_yuv420_to_argb32_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
//    rgb_std  = rgb_stride[0] - 4 * width;
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmx_unpack_32argb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
	if (height & 1) {
	    pu += u_stride;
	    pv += v_stride;
	} else {
	    pu -= 4 * width;
	    pv -= 4 * width;
	}
    } while (--height);
    emms();
}


void csp_yuv422_to_argb32_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmx_unpack_32argb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

void csp_yuv444_to_argb32_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - width;
    v_stride = yuv_stride[2] - width;
    rgb_std  = (rgb_stride[0] - width) * 4;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv44_to_rgb (py, pu, pv);
	    mmx_unpack_32argb (image);
	    py += 8;
	    pu += 8;
	    pv += 8;
	    image += 32;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

#endif

