/*
 * yuv_to_rgb_mmx.h
 *
 */

/*
 * yuv2rgb_mmx.c
 * Copyright (C) 2000-2002 Silicon Integrated System Corp.
 * All Rights Reserved.
 *
 * Author: Olie Lho <ollie@sis.com.tw>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#ifdef ARCH_X86

#ifndef YUV_TO_RGB_MMX_H
#define YUV_TO_RGB_MMX_H


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "attributes.h"
#include "mmx.h"

extern mmx_t csp_conv_mmx_80w;
extern mmx_t csp_conv_mmx_U_green;
extern mmx_t csp_conv_mmx_U_blue;
extern mmx_t csp_conv_mmx_V_red;
extern mmx_t csp_conv_mmx_V_green;
extern mmx_t csp_conv_mmx_10w;
extern mmx_t csp_conv_mmx_00ffw;
extern mmx_t csp_conv_mmx_Y_coeff;

extern mmx_t csp_conv_mmx_bluemask;
extern mmx_t csp_conv_mmx_greenmask;
extern mmx_t csp_conv_mmx_greenmask5;
extern mmx_t csp_conv_mmx_redmask;


static inline void mmx_yuv42_to_rgb (uint8_t * py, uint8_t * pu, uint8_t * pv)
{
    movd_m2r (*pu, mm0);		/* mm0 = 00 00 00 00 u3 u2 u1 u0 */
    movd_m2r (*pv, mm1);		/* mm1 = 00 00 00 00 v3 v2 v1 v0 */
    movq_m2r (*py, mm6);		/* mm6 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
    pxor_r2r (mm4, mm4);		/* mm4 = 0 */
    /* XXX might do cache preload for image here */

    /*
     * Do the multiply part of the conversion for even and odd pixels
     * register usage:
     * mm0 -> Cblue, mm1 -> Cred, mm2 -> Cgreen even pixels
     * mm3 -> Cblue, mm4 -> Cred, mm5 -> Cgreen odd  pixels
     * mm6 -> Y even, mm7 -> Y odd
     */

    punpcklbw_r2r (mm4, mm0);		/* mm0 = u3 u2 u1 u0 */
    punpcklbw_r2r (mm4, mm1);		/* mm1 = v3 v2 v1 v0 */
    psubsw_m2r (csp_conv_mmx_80w, mm0);		/* u -= 128 */
    psubsw_m2r (csp_conv_mmx_80w, mm1);		/* v -= 128 */
    psllw_i2r (3, mm0);			/* promote precision */
    psllw_i2r (3, mm1);			/* promote precision */
    movq_r2r (mm0, mm2);		/* mm2 = u3 u2 u1 u0 */
    movq_r2r (mm1, mm3);		/* mm3 = v3 v2 v1 v0 */
    pmulhw_m2r (csp_conv_mmx_U_green, mm2);	/* mm2 = u * u_green */
    pmulhw_m2r (csp_conv_mmx_V_green, mm3);	/* mm3 = v * v_green */
    pmulhw_m2r (csp_conv_mmx_U_blue, mm0);	/* mm0 = chroma_b */
    pmulhw_m2r (csp_conv_mmx_V_red, mm1);	/* mm1 = chroma_r */
    paddsw_r2r (mm3, mm2);		/* mm2 = chroma_g */

    psubusb_m2r (csp_conv_mmx_10w, mm6);		/* Y -= 16 */
    movq_r2r (mm6, mm7);		/* mm7 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
    pand_m2r (csp_conv_mmx_00ffw, mm6);		/* mm6 =    Y6    Y4    Y2    Y0 */
    psrlw_i2r (8, mm7);			/* mm7 =    Y7    Y5    Y3    Y1 */
    psllw_i2r (3, mm6);			/* promote precision */
    psllw_i2r (3, mm7);			/* promote precision */
    pmulhw_m2r (csp_conv_mmx_Y_coeff, mm6);	/* mm6 = luma_rgb even */
    pmulhw_m2r (csp_conv_mmx_Y_coeff, mm7);	/* mm7 = luma_rgb odd */

    /*
     * Do the addition part of the conversion for even and odd pixels
     * register usage:
     * mm0 -> Cblue, mm1 -> Cred, mm2 -> Cgreen even pixels
     * mm3 -> Cblue, mm4 -> Cred, mm5 -> Cgreen odd  pixels
     * mm6 -> Y even, mm7 -> Y odd
     */

    movq_r2r (mm0, mm3);		/* mm3 = chroma_b */
    movq_r2r (mm1, mm4);		/* mm4 = chroma_r */
    movq_r2r (mm2, mm5);		/* mm5 = chroma_g */
    paddsw_r2r (mm6, mm0);		/* mm0 = B6 B4 B2 B0 */
    paddsw_r2r (mm7, mm3);		/* mm3 = B7 B5 B3 B1 */
    paddsw_r2r (mm6, mm1);		/* mm1 = R6 R4 R2 R0 */
    paddsw_r2r (mm7, mm4);		/* mm4 = R7 R5 R3 R1 */
    paddsw_r2r (mm6, mm2);		/* mm2 = G6 G4 G2 G0 */
    paddsw_r2r (mm7, mm5);		/* mm5 = G7 G5 G3 G1 */
    packuswb_r2r (mm0, mm0);		/* saturate to 0-255 */
    packuswb_r2r (mm1, mm1);		/* saturate to 0-255 */
    packuswb_r2r (mm2, mm2);		/* saturate to 0-255 */
    packuswb_r2r (mm3, mm3);		/* saturate to 0-255 */
    packuswb_r2r (mm4, mm4);		/* saturate to 0-255 */
    packuswb_r2r (mm5, mm5);		/* saturate to 0-255 */
    punpcklbw_r2r (mm3, mm0);		/* mm0 = B7 B6 B5 B4 B3 B2 B1 B0 */
    punpcklbw_r2r (mm4, mm1);		/* mm1 = R7 R6 R5 R4 R3 R2 R1 R0 */
    punpcklbw_r2r (mm5, mm2);		/* mm2 = G7 G6 G5 G4 G3 G2 G1 G0 */
}

static inline void mmx_yuv44_to_rgb (uint8_t * py, uint8_t * pu, uint8_t * pv)
{
    movd_m2r (*pu, mm0);		/* mm0 = 00 00 00 00 u3 u2 u1 u0 */
    movd_m2r (*pv, mm1);		/* mm1 = 00 00 00 00 v3 v2 v1 v0 */
    movd_m2r (*py, mm5);		/* mm6 = 00 00 00 00 Y3 Y2 Y1 Y0 */
    pxor_r2r (mm4, mm4);		/* mm4 = 0 */
    /* XXX might do cache preload for image here */

    /*
     * Do the multiply part of the conversion for even and odd pixels
     * register usage:
     * mm0 -> Cblue, mm1 -> Cred, mm2 -> Cgreen even pixels
     * mm3 -> Cblue, mm4 -> Cred, mm5 -> Cgreen odd  pixels
     * mm6 -> Y even, mm7 -> Y odd
     */

    punpcklbw_r2r (mm4, mm0);		/* mm0 = u3 u2 u1 u0 */
    punpcklbw_r2r (mm4, mm1);		/* mm1 = v3 v2 v1 v0 */
    psubsw_m2r (csp_conv_mmx_80w, mm0);		/* u -= 128 */
    psubsw_m2r (csp_conv_mmx_80w, mm1);		/* v -= 128 */
    psllw_i2r (3, mm0);			/* promote precision */
    psllw_i2r (3, mm1);			/* promote precision */
    movq_r2r (mm0, mm2);		/* mm2 = u3 u2 u1 u0 */
    movq_r2r (mm1, mm3);		/* mm3 = v3 v2 v1 v0 */
    pmulhw_m2r (csp_conv_mmx_U_green, mm2);	/* mm2 = u * u_green */
    pmulhw_m2r (csp_conv_mmx_V_green, mm3);	/* mm3 = v * v_green */
    pmulhw_m2r (csp_conv_mmx_U_blue, mm0);	/* mm0 = chroma_b */
    pmulhw_m2r (csp_conv_mmx_V_red, mm1);	/* mm1 = chroma_r */
    paddsw_r2r (mm3, mm2);		/* mm2 = chroma_g */

    psubusb_m2r (csp_conv_mmx_10w, mm6);		/* Y -= 16 */
    punpcklbw_r2r (mm4, mm6);		/* mm6 = Y3 Y2 Y1 Y0 */
    psllw_i2r (3, mm6);			/* promote precision */
    pmulhw_m2r (csp_conv_mmx_Y_coeff, mm6);	/* mm6 = luma_rgb even */

    /*
     * Do the addition part of the conversion for even and odd pixels
     * register usage:
     * mm0 -> Cblue, mm1 -> Cred, mm2 -> Cgreen even pixels
     * mm6 -> Y
     */

    paddsw_r2r (mm6, mm0);		/* mm0 = B3 B2 B1 B0 */
    paddsw_r2r (mm6, mm1);		/* mm1 = R3 R2 R1 R0 */
    paddsw_r2r (mm6, mm2);		/* mm2 = G3 G2 G1 G0 */
    packuswb_r2r (mm0, mm0);		/* saturate to 0-255 */
    packuswb_r2r (mm1, mm1);		/* saturate to 0-255 */
    packuswb_r2r (mm2, mm2);		/* saturate to 0-255 */


    movd_m2r (*(pu+4), mm3);		/* mm3 = 00 00 00 00 u7 u6 u5 u4 */
    movd_m2r (*(pv+4), mm4);		/* mm4 = 00 00 00 00 v7 v6 v5 v4 */
    pxor_r2r (mm7, mm7);		/* mm7 = 0 */
    /* XXX might do cache preload for image here */

    /*
     * Do the multiply part of the conversion for even and odd pixels
     * register usage:
     * mm3 -> Cblue, mm4 -> Cred, mm5 -> Cgreen
     * mm6 -> Y
     */

    punpcklbw_r2r (mm7, mm3);		/* mm3 = u7 u6 u5 u4 */
    punpcklbw_r2r (mm7, mm4);		/* mm4 = v7 v6 v5 v4 */
    psubsw_m2r (csp_conv_mmx_80w, mm3);		/* u -= 128 */
    psubsw_m2r (csp_conv_mmx_80w, mm4);		/* v -= 128 */
    psllw_i2r (3, mm3);			/* promote precision */
    psllw_i2r (3, mm4);			/* promote precision */
    movq_r2r (mm3, mm5);		/* mm2 = u7 u6 u5 u4 */
    movq_r2r (mm4, mm6);		/* mm3 = v7 v6 v5 v4 */
    pmulhw_m2r (csp_conv_mmx_U_green, mm5);	/* mm2 = u * u_green */
    pmulhw_m2r (csp_conv_mmx_V_green, mm6);	/* mm3 = v * v_green */
    pmulhw_m2r (csp_conv_mmx_U_blue, mm3);	/* mm0 = chroma_b */
    pmulhw_m2r (csp_conv_mmx_V_red, mm4);	/* mm1 = chroma_r */
    paddsw_r2r (mm6, mm5);		/* mm2 = chroma_g */

    movd_m2r (*(py+4), mm6);		/* mm6 = 00 00 00 00 Y7 Y6 Y5 Y4 */
    psubusb_m2r (csp_conv_mmx_10w, mm6);		/* Y -= 16 */
    punpcklbw_r2r (mm7, mm6);		/* mm6 = Y7 Y6 Y5 Y4 */
    psllw_i2r (3, mm6);			/* promote precision */
    pmulhw_m2r (csp_conv_mmx_Y_coeff, mm6);	/* mm6 = luma_rgb even */

    paddsw_r2r (mm6, mm3);		/* mm3 = B7 B6 B5 B4 */
    paddsw_r2r (mm6, mm4);		/* mm4 = R7 R6 R5 R4 */
    paddsw_r2r (mm6, mm5);		/* mm5 = G7 G6 G5 G4 */
    packuswb_r2r (mm3, mm3);		/* saturate to 0-255 */
    packuswb_r2r (mm4, mm4);		/* saturate to 0-255 */
    packuswb_r2r (mm5, mm5);		/* saturate to 0-255 */

    punpckldq_r2r(mm3, mm0);            /* mm0 = B7 B6 B5 B4 B3 B2 B1 B0 */
    punpckldq_r2r(mm4, mm1);            /* mm0 = R7 R6 R5 R4 R3 R2 R1 R0 */
    punpckldq_r2r(mm5, mm2);            /* mm0 = G7 G6 G5 G4 G3 G2 G1 G0 */
}


#endif /* YUV_TO_RGB_MMX_H */

#endif /* ARCH_X86 */

