/*
 * yuv_to_rgbp.c
 *
 * code from mpeg2dec <http://libmpeg2.sourceforge.net>
 *
 */

/*
 * yuv2rgb.c
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "csp_conv.h"
#include "csp_conv_yuv_to_rgb.h"
#include "yuv_to_rgb.h"

static uint8_t ** table_rV = NULL;
static uint8_t ** table_gU = NULL;
static int     *  table_gV = NULL;
static uint8_t ** table_bU = NULL;
static uint8_t *  table_8 = NULL;

void csp_yuv_to_rgbp_init (void)
{
  YUV_to_RGB_Tables tables;

  if (table_8)
    return;

  yuv_to_rgb_init_table(YUV_TO_RGB_MODE_RGB, 24, &tables);

  table_rV = (uint8_t**)tables.table_rV;
  table_gU = (uint8_t**)tables.table_gU;
  table_gV = (int     *)tables.table_gV;
  table_bU = (uint8_t**)tables.table_bU;
  table_8 = (uint8_t *)tables.table;
}

void csp_yuv_to_rgbp_quit(void)
{
  if (table_rV) free(table_rV);
  if (table_gU) free(table_gU);
  if (table_gV) free(table_gV);
  if (table_bU) free(table_bU);
  if (table_8) free(table_8);

  table_rV = table_gU = table_bU = NULL;
  table_gV = NULL;
  table_8 = NULL;
}

#define RGB(i)                                              \
        U = pu[i];                                          \
        V = pv[i];                                          \
        r = table_rV[V];                                    \
        g = table_gU[U] + table_gV[V];                      \
        b = table_bU[U];

#define DSTRGB2(py,dstr,dstg,dstb,i)                                    \
        Y = py[2*i];                                                    \
        dstr[i] = r[Y]; dstg[i] = g[Y]; dstb[i] = b[Y];          \
        Y = py[2*i+1];                                                  \
        dstr[i+1] = r[Y]; dstg[i+1] = g[Y]; dstb[i+1] = b[Y];          \

#define DSTRGB1(py,dstr,dstg,dstb,i)                                  \
        Y = py[i];                                                    \
        dstr[i] = r[Y]; dstg[i] = g[Y]; dstb[i] = b[Y];

void csp_yuv420_to_rgbp (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint8_t * r, * g, * b;
  uint8_t *dstr, *dstg, *dstb;
  uint8_t *py1, *py2, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_8)
//    csp_yuv_to_rgbp_init ();

  dstr = rgb[0];
  dstg = rgb[1];
  dstb = rgb[2];
  py1 = yuv_planes[0];
  py2 = py1 + yuv_stride[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0]*2 - width;
  next_rgb = (rgb_stride[0]*2 - width)*3;
  next_uv = yuv_stride[1];

  width >>= 1;
  height >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DSTRGB2(py1,dstr1,dstg1,dstb1,0);
      DSTRGB2(py2,dstr2,dstg2,dstb2,0);

      py1 +=2; py2 +=2;
      dstr1+=2; dstr2+=2;
      dstg1+=2; dstg2+=2;
      dstb1+=2; dstb2+=2;
    }
    pu += next_uv;
    pv += next_uv;
    py1 += next_y;
    py2 += next_y;
    dst1 += next_rgb;
    dst2 += next_rgb;
  }
}

void csp_yuv422_to_rgbp (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint8_t * r, * g, * b;
  uint8_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_8)
//    csp_yuv_to_rgbp_init ();

  dst = (uint8_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0] - width;
  next_rgb = (rgb_stride[0] - width)*3;
  next_uv = yuv_stride[1];

  width >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DSTRGB2(py,dst,0);

      py += 2;
      dst+= 6;
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}

void csp_yuv444_to_rgbp (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint8_t * r, * g, * b;
  uint8_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_8)
//    csp_yuv_to_rgbp_init ();

  dst = (uint8_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0];
  next_rgb = (rgb_stride[0] - width)*3;
  next_uv = yuv_stride[1];

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DSTRGB1(py,dst,0);
      dst += 3;
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}

