/*
 * yuv_to_yuv.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "csp_conv_yuv_to_yuv.h"


#ifdef ARCH_X86
#define YUUY(sp,dp) \
      { register unsigned int v;           \
	v = *((uint32_t*)(sp));              \
        asm ("bswap	%0\n\t"            \
	     "rorl	$16, %0\n\t"       \
	     : "=r" (v) : "0" (v));        \
	*((uint32_t*)(dp)) = v;}
#else
#define YUUY(sp,dp) \
      (dp)[0] = (sp)[1];       \
      (dp)[1] = (sp)[0];       \
      (dp)[2] = (sp)[3];       \
      (dp)[3] = (sp)[2];
#endif

void csp_yuv420_cpy (uint8_t *src420[], uint8_t *dst420[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, k;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src420[0];
      dp = dst420[0];
      next_src = src_stride[0];
      next_dst = dst_stride[0];
    } else if (k == 2) {
      sp = src420[1];
      dp = dst420[1];
      next_src = src_stride[1];
      next_dst = dst_stride[1];
      height >>= 1;
      width >>= 1;
    } else {
      sp = src420[2];
      dp = dst420[2];
      next_src = src_stride[2];
      next_dst = dst_stride[2];
    }
    for (h = height; h > 0; h--)
    {
      memcpy(dp, sp, width);
      sp += next_src;
      dp += next_dst;
    }
  }
}

void csp_yuv420_to_yuv420 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv420_to_i420 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv420_to_yv12 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_yuv420 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_i420 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_yv12 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy (src, dst, src_stride, dst_stride, width, height);
}



void csp_yuv422_cpy (uint8_t *src422[], uint8_t *dst422[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, k;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src422[0];
      dp = dst422[0];
      next_src = src_stride[0];
      next_dst = dst_stride[0];
    } else if (k == 2) {
      sp = src422[1];
      dp = dst422[1];
      next_src = src_stride[1];
      next_dst = dst_stride[1];
      width >>= 1;
    } else {
      sp = src422[2];
      dp = dst422[2];
      next_src = src_stride[2];
      next_dst = dst_stride[2];
    }
    for (h = height; h > 0; h--)
    {
      memcpy(dp, sp, width);
      sp += next_src;
      dp += next_dst;
    }
  }
}

void csp_yuv422_to_yuv422 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422_cpy (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv422_to_yuy2 (uint8_t *yuv422_planes[], uint8_t *yuy2_planes[],
	  int yuv422_stride[], int yuy2_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *yuy2;
  int next_y, next_u, next_v, next_yuy2;
  int h, w;

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  yuy2 = yuy2_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_yuy2 = (yuy2_stride[0] - width) * 2;

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w--)
    {
      yuy2[0] = yp[0];
      yuy2[1] = up[0];
      yuy2[2] = yp[1];
      yuy2[3] = vp[0];
      yp += 2;
      up++;
      vp++;
      yuy2 += 4;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    yuy2 += next_yuy2;
  }
}

void csp_yuv422_to_uyvy (uint8_t *yuv422_planes[], uint8_t *uyvy_planes[],
	  int yuv422_stride[], int uyvy_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *uyvy;
  int next_y, next_u, next_v, next_uyvy;
  int h, w;

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  uyvy = uyvy_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_uyvy = (uyvy_stride[0] - width) * 2;

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w--)
    {
      uyvy[0] = up[0];
      uyvy[1] = yp[0];
      uyvy[2] = vp[0];
      uyvy[3] = yp[1];
      yp += 2;
      up++;
      vp++;
      uyvy += 4;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    uyvy += next_uyvy;
  }
}

void csp_yuy2_to_yuv422 (uint8_t *yuy2_planes[], uint8_t *yuv422_planes[],
	  int yuy2_stride[], int yuv422_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *yuy2;
  int next_y, next_u, next_v, next_yuy2;
  int h, w;

  yuy2 = yuy2_planes[0];
  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];

  next_yuy2 = (yuy2_stride[0] - width) * 2;
  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w--)
    {
#if 0
      yp[0] = yuy2[0];
      up[0] = yuy2[1];
      yp[1] = yuy2[2];
      vp[0] = yuy2[3];
#else
      register int v;
      v = *((uint32_t*)yuy2);
      yp[0] = v;
      yp[1] = v>>16;
      up[0] = v>>8;
      vp[0] = v>>24;
#endif
      yp += 2;
      up++;
      vp++;
      yuy2 += 4;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    yuy2 += next_yuy2;
  }
}

void csp_yuv422pack_cpy (uint8_t *src_planes[], uint8_t *dst_planes[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *src, *dst;
  int next_src, next_dst;
  int h;

  src = src_planes[0];
  dst = dst_planes[0];

  next_src = src_stride[0] * 2;
  next_dst = dst_stride[0] * 2;

  width <<= 1;
  for (h = height; h > 0; h--)
  {
    memcpy(dst, src, width);
    src += next_src;
    dst += next_dst;
  }
}

void csp_yuy2_to_yuy2 (uint8_t *src_planes[], uint8_t *dst_planes[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422pack_cpy(src_planes, dst_planes,
	  src_stride, dst_stride, width, height);
}

void csp_yuy2_to_uyvy (uint8_t *yuy2_planes[], uint8_t *uyvy_planes[],
	  int yuy2_stride[], int uyvy_stride[], int width, int height)
{
  uint8_t *uyvy, *yuy2;
  int next_yuy2, next_uyvy;
  int h, w;

  yuy2 = yuy2_planes[0];
  uyvy = uyvy_planes[0];

  next_yuy2 = (yuy2_stride[0] - width) * 2;
  next_uyvy = (uyvy_stride[0] - width) * 2;

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w-=4)
    {
#if 0
      uyvy[0] = yuy2[1];
      uyvy[1] = yuy2[0];
      uyvy[2] = yuy2[3];
      uyvy[3] = yuy2[2];
#endif
      YUUY(yuy2, uyvy);
      YUUY(yuy2+4, uyvy+4);
      YUUY(yuy2+8, uyvy+8);
      YUUY(yuy2+12, uyvy+12);
      uyvy += 16;
      yuy2 += 16;
    }
    uyvy += next_uyvy;
    yuy2 += next_yuy2;
  }
}


void csp_uyvy_to_yuv422 (uint8_t *uyvy_planes[], uint8_t *yuv422_planes[],
	  int uyvy_stride[], int yuv422_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *uyvy;
  int next_y, next_u, next_v, next_uyvy;
  int h, w;

  uyvy = uyvy_planes[0];
  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];

  next_uyvy = (uyvy_stride[0] - width) * 2;
  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w--)
    {
      yp[0] = uyvy[0];
      up[0] = uyvy[1];
      yp[1] = uyvy[2];
      vp[0] = uyvy[3];
      yp += 2;
      up++;
      vp++;
      uyvy += 4;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    uyvy += next_uyvy;
  }
}

void csp_uyvy_to_uyvy (uint8_t *src_planes[], uint8_t *dst_planes[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422pack_cpy(src_planes, dst_planes,
	  src_stride, dst_stride, width, height);
}

void csp_uyvy_to_yuy2 (uint8_t *uyvy_planes[], uint8_t *yuy2_planes[],
	  int uyvy_stride[], int yuy2_stride[], int width, int height)
{
  uint8_t *uyvy, *yuy2;
  int next_yuy2, next_uyvy;
  int h, w;

  uyvy = uyvy_planes[0];
  yuy2 = yuy2_planes[0];

  next_uyvy = (uyvy_stride[0] - width) * 2;
  next_yuy2 = (yuy2_stride[0] - width) * 2;

  width >>= 1;
  for (h = height; h > 0; h--)
  {
    for (w = width; w > 0; w-=4)
    {
#if 0
      yuy2[0] = uyvy[1];
      yuy2[1] = uyvy[0];
      yuy2[2] = uyvy[3];
      yuy2[3] = uyvy[2];
#endif
      YUUY(uyvy, yuy2);
      YUUY(uyvy+4, yuy2+4);
      YUUY(uyvy+8, yuy2+8);
      YUUY(uyvy+12, yuy2+12);
      uyvy += 16;
      yuy2 += 16;
    }
    uyvy += next_uyvy;
    yuy2 += next_yuy2;
  }
}




void csp_yuv444_cpy (uint8_t *src444[], uint8_t *dst444[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, k;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src444[0];
      dp = dst444[0];
      next_src = src_stride[0];
      next_dst = dst_stride[0];
    } else if (k == 2) {
      sp = src444[1];
      dp = dst444[1];
      next_src = src_stride[1];
      next_dst = dst_stride[1];
    } else {
      sp = src444[2];
      dp = dst444[2];
      next_src = src_stride[2];
      next_dst = dst_stride[2];
    }
    for (h = height; h > 0; h--)
    {
      memcpy(dp, sp, width);
      sp += next_src;
      dp += next_dst;
    }
  }
}


void csp_yuv444_to_yuv444 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv444_cpy (src, dst, src_stride, dst_stride, width, height);
}

