/*
 * yuv_to_yuv_mmx.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#ifdef ARCH_X86


#include "csp_conv_yuv_to_yuv.h"


#include "attributes.h"
#include "mmx.h"


static void csp_yuv420_cpy_mmxext (uint8_t *src420[], uint8_t *dst420[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, w, k;
  int tw;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src420[0];
      dp = dst420[0];
      next_src = src_stride[0] - width;
      next_dst = dst_stride[0] - width;
    } else if (k == 2) {
      sp = src420[1];
      dp = dst420[1];
      height >>= 1;
      width >>= 1;
      next_src = src_stride[1] - width;
      next_dst = dst_stride[1] - width;
    } else {
      sp = src420[2];
      dp = dst420[2];
      next_src = src_stride[2] - width;
      next_dst = dst_stride[2] - width;
    }
    tw = width - 32;
    for (h = height; h > 0; h--)
    {
      for (w = 0; w <= tw; w+=32) {
//	prefetchnta(*sp);
	movq_m2r (*sp, mm0);
	movntq_r2m (mm0, *dp);
	movq_m2r (sp[8], mm0);
	movntq_r2m (mm0, dp[8]);
	movq_m2r (sp[16], mm0);
	movntq_r2m (mm0, dp[16]);
	movq_m2r (sp[24], mm0);
	movntq_r2m (mm0, dp[24]);
	sp += 32;
	dp += 32;
      }
//      w = 0;
      for (; w < width; w+=8) {
//	prefetchnta(*sp);
	movq_m2r (*sp, mm0);
	movntq_r2m (mm0, *dp);
	sp += 8;
	dp += 8;
      }
      sp += next_src;
      dp += next_dst;
    }
  }
  emms();
}

void csp_yuv420_to_yuv420_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy_mmxext (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv420_to_i420_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy_mmxext (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv420_to_yv12_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy_mmxext (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_yuv420_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy_mmxext (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_i420_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *tdst[3];
  int tstrd[3];
  tdst[0] = dst[0];
  tdst[1] = dst[2];
  tdst[2] = dst[1];
  tstrd[0] = dst_stride[0];
  tstrd[1] = dst_stride[2];
  tstrd[2] = dst_stride[1];
  csp_yuv420_cpy_mmxext (src, tdst, src_stride, tstrd, width, height);
}

void csp_yv12_to_yv12_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv420_cpy_mmxext (src, dst, src_stride, dst_stride, width, height);
}


/* YUV422 */

static void csp_yuv422_cpy_mmxext (uint8_t *src422[], uint8_t *dst422[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, w, k;
  int tw;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src422[0];
      dp = dst422[0];
      next_src = src_stride[0] - width;
      next_dst = dst_stride[0] - width;
    } else if (k == 2) {
      sp = src422[1];
      dp = dst422[1];
      width >>= 1;
      next_src = src_stride[1] - width;
      next_dst = dst_stride[1] - width;
    } else {
      sp = src422[2];
      dp = dst422[2];
      next_src = src_stride[2] - width;
      next_dst = dst_stride[2] - width;
    }
    tw = width - 32;
    for (h = height; h > 0; h--)
    {
      for (w = 0; w <= tw; w+=32) {
	movq_m2r (*sp, mm0);
	movntq_r2m (mm0, *dp);
	movq_m2r (*(sp+8), mm0);
	movntq_r2m (mm0, *(dp+8));
	movq_m2r (*(sp+16), mm0);
	movntq_r2m (mm0, *(dp+16));
	movq_m2r (*(sp+24), mm0);
	movntq_r2m (mm0, *(dp+24));
	sp += 32;
	dp += 32;
      }
      for (; w < width; w+=8) {
	movq_m2r (*sp, mm0);
	movntq_r2m (mm0, *dp);
	sp += 8;
	dp += 8;
      }
      sp += next_src;
      dp += next_dst;
    }
  }
  emms();
}

void csp_yuv422_to_yuv422_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422_cpy_mmxext (src, dst, src_stride, dst_stride, width, height);
}

void csp_yuv422_to_yuy2_mmxext (uint8_t *yuv422_planes[],uint8_t *yuy2_planes[],
          int yuv422_stride[], int yuy2_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *yuy2;
  int next_y, next_u, next_v, next_yuy2;
  int h, w, tw;

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  yuy2 = yuy2_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_yuy2 = (yuy2_stride[0] - width) * 2;

//  width >>= 1;
  tw = width - 16;
  pxor_r2r (mm7, mm7);       /* 00 00 00 00 00 00 00 00 */
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=16)
    {
#if 0
      yuy2[0] = yp[0];
      yuy2[1] = up[0];
      yuy2[2] = yp[1];
      yuy2[3] = vp[0];
#endif
      movq_m2r (*yp, mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_m2r (*up, mm1);       /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movq_m2r (*vp, mm2);       /* V7 V6 V5 V4 V3 V2 V1 V0 */

      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_r2r (mm1, mm4);       /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movq_r2r (mm2, mm5);       /* V7 V6 V5 V4 V3 V2 V1 V0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */

      punpcklbw_r2r (mm7, mm4);  /* 00 U3 00 U2 00 U1 00 U0 */
      movq_r2r (mm4, mm6);       /* 00 U3 00 U2 00 U1 00 U0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U1 00 00 00 U0 */
      psllq_i2r (8, mm6);        /* 00 00 U1 00 00 00 U0 00 */
      por_r2r (mm6, mm3);        /* 00 Y3 U1 Y2 00 Y1 U0 Y0 */

      punpcklbw_r2r (mm7, mm5);  /* 00 V3 00 V2 00 V1 00 V0 */
      movq_r2r (mm5, mm6);       /* 00 V3 00 V2 00 V1 00 V0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V1 00 00 00 V0 */
      psllq_i2r (24, mm6);       /* V1 00 00 00 V0 00 00 00 */
      por_r2r (mm6, mm3);        /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */

      movntq_r2m (mm3, *(yuy2));

      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */

      punpckhwd_r2r (mm7, mm4);  /* 00 00 00 U3 00 00 00 U2 */
      psllq_i2r (8, mm4);        /* 00 00 U3 00 00 00 U2 00 */
      por_r2r (mm4, mm0);        /* 00 Y7 U3 Y6 00 Y5 U2 Y4 */

      punpckhwd_r2r (mm7, mm5);  /* 00 00 00 V3 00 00 00 V2 */
      psllq_i2r (24, mm5);       /* V3 00 00 00 V2 00 00 00 */
      por_r2r (mm5, mm0);        /* V3 Y7 U3 Y6 V2 Y5 U2 Y4 */

      movntq_r2m (mm0, *(yuy2+8));



      movq_m2r (*(yp+8), mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */

      punpckhbw_r2r (mm7, mm1);  /* 00 U7 00 U6 00 U5 00 U4 */
      movq_r2r (mm1, mm6);       /* 00 U7 00 U6 00 U5 00 U4 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U5 00 00 00 U4 */
      psllq_i2r (8, mm6);        /* 00 00 U5 00 00 00 U4 00 */
      por_r2r (mm6, mm3);        /* 00 Y3 U5 Y2 00 Y1 U4 Y0 */

      punpckhbw_r2r (mm7, mm2);  /* 00 V7 00 V6 00 V5 00 V4 */
      movq_r2r (mm2, mm6);       /* 00 V7 00 V6 00 V5 00 V4 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V5 00 00 00 V4 */
      psllq_i2r (24, mm6);       /* V5 00 00 00 V4 00 00 00 */
      por_r2r (mm6, mm3);        /* V5 Y3 U5 Y2 V4 Y1 U4 Y0 */

      movntq_r2m (mm3, *(yuy2+16));


      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */

      punpckhwd_r2r (mm7, mm1);  /* 00 00 00 U7 00 00 00 U6 */
      psllq_i2r (8, mm1);        /* 00 00 U7 00 00 00 U6 00 */
      por_r2r (mm1, mm0);        /* 00 Y7 U7 Y6 00 Y5 U6 Y4 */

      punpckhwd_r2r (mm7, mm2);  /* 00 00 00 V7 00 00 00 V6 */
      psllq_i2r (24, mm2);       /* V7 00 00 00 V6 00 00 00 */
      por_r2r (mm2, mm0);        /* V7 Y7 U7 Y6 V6 Y5 U6 Y4 */

      movntq_r2m (mm0, *(yuy2+24));

      yp += 16;
      up+=8;
      vp+=8;
      yuy2 += 32;
    }
    for (; w < width; w+=8)
    {
      movq_m2r (*yp, mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movd_m2r (*up, mm1);       /* 00 00 00 00 U3 U2 U1 U0 */
      movd_m2r (*vp, mm2);       /* 00 00 00 00 V3 V2 V1 V0 */

      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */

      punpcklbw_r2r (mm7, mm1);  /* 00 U3 00 U2 00 U1 00 U0 */
      movq_r2r (mm1, mm6);       /* 00 U3 00 U2 00 U1 00 U0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U1 00 00 00 U0 */
      psllq_i2r (8, mm6);        /* 00 00 U1 00 00 00 U0 00 */
      por_r2r (mm6, mm3);        /* 00 Y3 U1 Y2 00 Y1 U0 Y0 */

      punpcklbw_r2r (mm7, mm2);  /* 00 V3 00 V2 00 V1 00 V0 */
      movq_r2r (mm2, mm6);       /* 00 V3 00 V2 00 V1 00 V0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V1 00 00 00 V0 */
      psllq_i2r (24, mm6);       /* V1 00 00 00 V0 00 00 00 */
      por_r2r (mm6, mm3);        /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */

      movntq_r2m (mm3, *(yuy2));

      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */

      punpckhwd_r2r (mm7, mm1);  /* 00 00 00 U3 00 00 00 U2 */
      psllq_i2r (8, mm1);        /* 00 00 U3 00 00 00 U2 00 */
      por_r2r (mm1, mm0);        /* 00 Y7 U3 Y6 00 Y5 U2 Y4 */

      punpckhwd_r2r (mm7, mm2);  /* 00 00 00 V3 00 00 00 V2 */
      psllq_i2r (24, mm2);       /* V3 00 00 00 V2 00 00 00 */
      por_r2r (mm2, mm0);        /* V3 Y7 U3 Y6 V2 Y5 U2 Y4 */

      movntq_r2m (mm0, *(yuy2+8));

      yp += 8;
      up+=4;
      vp+=4;
      yuy2 += 16;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    yuy2 += next_yuy2;
  }
  emms();
}

void csp_yuv422_to_uyvy_mmxext (uint8_t *yuv422_planes[],uint8_t *uyvy_planes[],
          int yuv422_stride[], int uyvy_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *uyvy;
  int next_y, next_u, next_v, next_uyvy;
  int h, w, tw;

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  uyvy = uyvy_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_uyvy = (uyvy_stride[0] - width) * 2;

//  width >>= 1;
  tw = width - 16;
  pxor_r2r (mm7, mm7);       /* 00 00 00 00 00 00 00 00 */
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=16)
    {
#if 0
      uyvy[0] = yp[0];
      uyvy[1] = up[0];
      uyvy[2] = yp[1];
      uyvy[3] = vp[0];
#endif
      movq_m2r (*yp, mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_m2r (*up, mm1);       /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movq_m2r (*vp, mm2);       /* V7 V6 V5 V4 V3 V2 V1 V0 */

      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_r2r (mm1, mm4);       /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movq_r2r (mm2, mm5);       /* V7 V6 V5 V4 V3 V2 V1 V0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      psllq_i2r (8, mm3);        /* Y3 00 Y2 00 Y1 00 Y0 00 */

      punpcklbw_r2r (mm7, mm4);  /* 00 U3 00 U2 00 U1 00 U0 */
      movq_r2r (mm4, mm6);       /* 00 U3 00 U2 00 U1 00 U0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U1 00 00 00 U0 */
      por_r2r (mm6, mm3);        /* Y3 00 Y2 U1 Y1 00 Y0 U0 */

      punpcklbw_r2r (mm7, mm5);  /* 00 V3 00 V2 00 V1 00 V0 */
      movq_r2r (mm5, mm6);       /* 00 V3 00 V2 00 V1 00 V0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V1 00 00 00 V0 */
      psllq_i2r (16, mm6);       /* 00 V1 00 00 00 V0 00 00 */
      por_r2r (mm6, mm3);        /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */

      movntq_r2m (mm3, *(uyvy));

      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      psllq_i2r (8, mm0);        /* Y7 00 Y6 00 Y5 00 Y4 00 */

      punpckhwd_r2r (mm7, mm4);  /* 00 00 00 U3 00 00 00 U2 */
      por_r2r (mm4, mm0);        /* Y7 00 Y6 U3 Y5 00 Y4 U2 */

      punpckhwd_r2r (mm7, mm5);  /* 00 00 00 V3 00 00 00 V2 */
      psllq_i2r (16, mm5);       /* 00 V3 00 00 00 V2 00 00 */
      por_r2r (mm5, mm0);        /* Y7 V3 Y6 U3 Y5 V2 Y4 U2 */

      movntq_r2m (mm0, *(uyvy+8));



      movq_m2r (*(yp+8), mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      psllq_i2r (8, mm3);        /* Y3 00 Y2 00 Y1 00 Y0 00 */

      punpckhbw_r2r (mm7, mm1);  /* 00 U7 00 U6 00 U5 00 U4 */
      movq_r2r (mm1, mm6);       /* 00 U7 00 U6 00 U5 00 U4 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U5 00 00 00 U4 */
      por_r2r (mm6, mm3);        /* Y3 00 Y2 U5 Y1 00 Y0 U4 */

      punpckhbw_r2r (mm7, mm2);  /* 00 V7 00 V6 00 V5 00 V4 */
      movq_r2r (mm2, mm6);       /* 00 V7 00 V6 00 V5 00 V4 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V5 00 00 00 V4 */
      psllq_i2r (16, mm6);       /* 00 V5 00 00 00 V4 00 00 */
      por_r2r (mm6, mm3);        /* Y3 V5 Y2 U5 Y1 V4 Y0 U4 */

      movntq_r2m (mm3, *(uyvy+16));


      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      psllq_i2r (8, mm0);        /* Y7 00 Y6 00 Y5 00 Y4 00 */

      punpckhwd_r2r (mm7, mm1);  /* 00 00 00 U7 00 00 00 U6 */
      por_r2r (mm1, mm0);        /* Y7 00 Y6 U7 Y5 00 Y4 U6 */

      punpckhwd_r2r (mm7, mm2);  /* 00 00 00 V7 00 00 00 V6 */
      psllq_i2r (16, mm2);       /* 00 V7 00 00 00 V6 00 00 */
      por_r2r (mm2, mm0);        /* Y7 V7 Y6 U7 Y5 V6 Y4 U6 */

      movntq_r2m (mm0, *(uyvy+24));

      yp += 16;
      up+=8;
      vp+=8;
      uyvy += 32;
    }
    for (; w < width; w+=8)
    {
      movq_m2r (*yp, mm0);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movd_m2r (*up, mm1);       /* 00 00 00 00 U3 U2 U1 U0 */
      movd_m2r (*vp, mm2);       /* 00 00 00 00 V3 V2 V1 V0 */

      movq_r2r (mm0, mm3);       /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */

      punpcklbw_r2r (mm7, mm3);  /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      psllq_i2r (8, mm3);        /* Y3 00 Y2 00 Y1 00 Y0 00 */

      punpcklbw_r2r (mm7, mm1);  /* 00 U3 00 U2 00 U1 00 U0 */
      movq_r2r (mm1, mm6);       /* 00 U3 00 U2 00 U1 00 U0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 U1 00 00 00 U0 */
      por_r2r (mm6, mm3);        /* Y3 00 Y2 U1 Y1 00 Y0 U0 */

      punpcklbw_r2r (mm7, mm2);  /* 00 V3 00 V2 00 V1 00 V0 */
      movq_r2r (mm2, mm6);       /* 00 V3 00 V2 00 V1 00 V0 */
      punpcklwd_r2r (mm7, mm6);  /* 00 00 00 V1 00 00 00 V0 */
      psllq_i2r (16, mm6);       /* 00 V1 00 00 00 V0 00 00 */
      por_r2r (mm6, mm3);        /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */

      movntq_r2m (mm3, *(uyvy));

      punpckhbw_r2r (mm7, mm0);  /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      psllq_i2r (8, mm0);        /* Y7 00 Y6 00 Y5 00 Y4 00 */

      punpckhwd_r2r (mm7, mm1);  /* 00 00 00 U3 00 00 00 U2 */
      por_r2r (mm1, mm0);        /* Y7 00 Y6 U3 Y5 00 Y4 U2 */

      punpckhwd_r2r (mm7, mm2);  /* 00 00 00 V3 00 00 00 V2 */
      psllq_i2r (16, mm2);       /* 00 V3 00 00 00 V2 00 00 */
      por_r2r (mm2, mm0);        /* Y7 V3 Y6 U3 Y5 V2 Y4 U2 */

      movntq_r2m (mm0, *(uyvy+8));

      yp += 8;
      up+=4;
      vp+=4;
      uyvy += 16;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    uyvy += next_uyvy;
  }
  emms();
}

void csp_yuy2_to_yuv422_mmxext (uint8_t *yuy2_planes[],uint8_t *yuv422_planes[],
          int yuy2_stride[], int yuv422_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *yuy2;
  int next_y, next_u, next_v, next_yuy2;
  int h, w, tw;
  static mmx_t ymask = {0x00FF00FF00FF00FFLL};
  static mmx_t uvmask = {0x000000FF000000FFLL};

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  yuy2 = yuy2_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_yuy2 = (yuy2_stride[0] - width) * 2;

//  width >>= 1;
  tw = width - 16;
  movq_m2r (ymask, mm7);
  movq_m2r (uvmask, mm6);
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=16)
    {
#if 0
      yuy2[0] = yp[0];
      yuy2[1] = up[0];
      yuy2[2] = yp[1];
      yuy2[3] = vp[0];
#endif
      movq_m2r (*yuy2, mm0);     /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */
      movq_m2r (*(yuy2+8), mm1); /* V3 Y7 U3 Y6 V2 Y5 U2 Y4 */
      movq_r2r (mm0, mm4);       /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */
      movq_r2r (mm0, mm5);       /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */
      movq_r2r (mm1, mm2);       /* V3 Y7 U3 Y6 V2 Y5 U2 Y4 */
      movq_r2r (mm1, mm3);       /* V3 Y7 U3 Y6 V2 Y5 U2 Y4 */
//      pand_m2r (ymask, mm0);
//      pand_m2r (ymask, mm1);
      pand_r2r (mm7, mm0);       /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      pand_r2r (mm7, mm1);       /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      packuswb_r2r (mm1, mm0);   /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movntq_r2m (mm0, *yp);

      psrlq_i2r (8, mm4);        /* 00 V1 Y3 U1 Y2 V0 Y1 U0 */
      psrlq_i2r (8, mm2);        /* 00 V3 Y7 U3 Y6 V2 Y5 U2 */
//      pand_m2r (uvmask, mm4);
//      pand_m2r (uvmask, mm2);
      pand_r2r (mm6, mm4);       /* 00 00 00 U1 00 00 00 U0 */
      pand_r2r (mm6, mm2);       /* 00 00 00 U3 00 00 00 U2 */
      packuswb_r2r (mm2, mm4);   /* 00 U3 00 U2 00 U1 00 U0 */

#if 0
      psrld_i2r (24, mm5);       /* 00 00 00 V1 00 00 00 V0 */
      psrld_i2r (24, mm3);       /* 00 00 00 V3 00 00 00 V2 */
#else
      psrlq_i2r (24, mm5);       /* 00 00 00 V1 00 00 00 V0 */
      psrlq_i2r (24, mm3);       /* 00 00 00 V3 00 00 00 V2 */
      pand_r2r (mm6, mm5);       /* 00 00 00 U1 00 00 00 U0 */
      pand_r2r (mm6, mm3);       /* 00 00 00 U3 00 00 00 U2 */
#endif
      packuswb_r2r (mm3, mm5);   /* 00 V3 00 V2 00 V1 00 V0 */

      movq_m2r (*(yuy2+16), mm0); /* V5 Y3 U5 Y2 V4 Y1 U4 Y0 */
      movq_m2r (*(yuy2+24), mm1); /* V7 Y7 U7 Y6 V6 Y5 U6 Y4 */
      movq_r2r (mm0, mm2);        /* V5 Y3 U5 Y2 V4 Y1 U4 Y0 */
      movq_r2r (mm1, mm3);        /* V7 Y7 U7 Y6 V6 Y5 U6 Y4 */
      pand_r2r (mm7, mm2);        /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      pand_r2r (mm7, mm3);        /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      packuswb_r2r (mm3, mm2);    /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movntq_r2m (mm2, *(yp+8));

      movq_r2r (mm0, mm2);        /* V5 Y3 U5 Y2 V4 Y1 U4 Y0 */
      movq_r2r (mm1, mm3);        /* V7 Y7 U7 Y6 V6 Y5 U6 Y4 */
      psrlq_i2r (8, mm2);         /* 00 V5 Y3 U5 Y2 V4 Y1 U4 */
      psrlq_i2r (8, mm3);         /* 00 V7 Y7 U7 Y6 V6 Y5 U6 */
      pand_r2r (mm6, mm2);        /* 00 00 00 U5 00 00 00 U4 */
      pand_r2r (mm6, mm3);        /* 00 00 00 U7 00 00 00 U6 */
      packuswb_r2r (mm3, mm2);    /* 00 U7 00 U6 00 U5 00 U4 */
      packuswb_r2r (mm2, mm4);    /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movntq_r2m (mm4, *up);

#if 0
      psrld_i2r (24, mm0);        /* 00 00 00 V5 00 00 00 V4 */
      psrld_i2r (24, mm1);        /* 00 00 00 V7 00 00 00 V6 */
#else
      psrlq_i2r (24, mm0);        /* 00 00 00 V5 00 00 00 V4 */
      psrlq_i2r (24, mm1);        /* 00 00 00 V7 00 00 00 V6 */
      pand_r2r (mm6, mm0);       /* 00 00 00 U1 00 00 00 U0 */
      pand_r2r (mm6, mm1);       /* 00 00 00 U3 00 00 00 U2 */
#endif
      packuswb_r2r (mm1, mm0);    /* 00 V7 00 V6 00 V5 00 V4 */
      packuswb_r2r (mm0, mm5);    /* V7 V6 V5 V4 V3 V2 V1 V0 */
      movntq_r2m (mm5, *vp);

      yp += 16;
      up+=8;
      vp+=8;
      yuy2 += 32;
    }
    for (; w < width; w+=8)
    {
      movq_m2r (*yuy2, mm0);
      movq_m2r (*(yuy2+8), mm1);
      movq_r2r (mm0, mm2);
      movq_r2r (mm1, mm3);
      pand_r2r (mm7, mm0);
      pand_r2r (mm7, mm1);
      packuswb_r2r (mm1, mm0);
      movntq_r2m (mm0, *yp);

      movq_r2r (mm2, mm0);
      movq_r2r (mm3, mm1);
      psrlq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm6, mm0);
      pand_r2r (mm6, mm1);
      packuswb_r2r (mm1, mm0);
      packuswb_r2r (mm1, mm0);
      movd_r2m (mm0, *up);

#if 1
      psrld_i2r (24, mm2);
      psrld_i2r (24, mm3);
#else
      psrlq_i2r (24, mm2);
      psrlq_i2r (24, mm3);
      pand_r2r (mm6, mm2);
      pand_r2r (mm6, mm3);
#endif
      packuswb_r2r (mm3, mm2);
      packuswb_r2r (mm3, mm2);
      movd_r2m (mm2, *vp);

      yp += 8;
      up+=4;
      vp+=4;
      yuy2 += 16;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    yuy2 += next_yuy2;
  }
  emms();
}

static void csp_yuv422pack_cpy_mmxext (uint8_t *src_planes[], uint8_t *dst_planes[],
          int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, w, tw;

  sp = src_planes[0];
  dp = dst_planes[0];

  next_src = (src_stride[0] - width) * 2;
  next_dst = (dst_stride[0] - width) * 2;

  width *= 2;
  tw = width - 32;
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=32) {
      movq_m2r (*sp, mm0);
      movntq_r2m (mm0, *dp);
      movq_m2r (*(sp+8), mm0);
      movntq_r2m (mm0, *(dp+8));
      movq_m2r (*(sp+16), mm0);
      movntq_r2m (mm0, *(dp+16));
      movq_m2r (*(sp+24), mm0);
      movntq_r2m (mm0, *(dp+24));
      sp += 32;
      dp += 32;
    }
    for (; w < width; w+=8) {
      movq_m2r (*sp, mm0);
      movntq_r2m (mm0, *dp);
      sp += 8;
      dp += 8;
    }
    sp += next_src;
    dp += next_dst;
  }
  emms();
}

void csp_yuy2_to_yuy2_mmxext (uint8_t *src_planes[], uint8_t *dst_planes[],
              int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422pack_cpy_mmxext(src_planes, dst_planes,
          src_stride, dst_stride, width, height);
}

void csp_uyvy_to_uyvy_mmxext (uint8_t *src_planes[], uint8_t *dst_planes[],
              int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv422pack_cpy_mmxext(src_planes, dst_planes,
          src_stride, dst_stride, width, height);
}

void csp_uyvy_to_yuv422_mmxext (uint8_t *uyvy_planes[],uint8_t *yuv422_planes[],
          int uyvy_stride[], int yuv422_stride[], int width, int height)
{
  uint8_t *yp, *up, *vp, *uyvy;
  int next_y, next_u, next_v, next_uyvy;
  int h, w, tw;
  static mmx_t ymask = {0x00FF00FF00FF00FFLL};
  static mmx_t uvmask = {0x000000FF000000FFLL};

  yp = yuv422_planes[0];
  up = yuv422_planes[1];
  vp = yuv422_planes[2];
  uyvy = uyvy_planes[0];

  next_y = yuv422_stride[0] - width;
  next_u = yuv422_stride[1] - (width>>1);
  next_v = yuv422_stride[2] - (width>>1);
  next_uyvy = (uyvy_stride[0] - width) * 2;

//  width >>= 1;
  tw = width - 16;
  movq_m2r (ymask, mm7);
  movq_m2r (uvmask, mm6);
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=16)
    {
      movq_m2r (*uyvy, mm0);     /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */
      movq_m2r (*(uyvy+8), mm1); /* Y7 V3 Y6 U3 Y5 V2 Y4 U2 */
      movq_r2r (mm0, mm4);       /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */
      movq_r2r (mm0, mm5);       /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */
      movq_r2r (mm1, mm2);       /* Y7 V3 Y6 U3 Y5 V2 Y4 U2 */
      movq_r2r (mm1, mm3);       /* Y7 V3 Y6 U3 Y5 V2 Y4 U2 */
      psrlq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm7, mm0);       /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      pand_r2r (mm7, mm1);       /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      packuswb_r2r (mm1, mm0);   /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movntq_r2m (mm0, *yp);

      pand_r2r (mm6, mm4);       /* 00 00 00 U1 00 00 00 U0 */
      pand_r2r (mm6, mm2);       /* 00 00 00 U3 00 00 00 U2 */
      packuswb_r2r (mm2, mm4);   /* 00 U3 00 U2 00 U1 00 U0 */

      psrlq_i2r (16, mm5);       /* 00 00 00 V1 00 00 00 V0 */
      psrlq_i2r (16, mm3);       /* 00 00 00 V3 00 00 00 V2 */
      pand_r2r (mm6, mm5);       /* 00 00 00 V1 00 00 00 V0 */
      pand_r2r (mm6, mm3);       /* 00 00 00 V3 00 00 00 V2 */
      packuswb_r2r (mm3, mm5);   /* 00 V3 00 V2 00 V1 00 V0 */

      movq_m2r (*(uyvy+16), mm0); /* Y3 V5 Y2 U5 Y1 V4 Y0 U4 */
      movq_m2r (*(uyvy+24), mm1); /* Y7 V7 Y6 U7 Y5 V6 Y4 U6 */
      movq_r2r (mm0, mm2);        /* Y3 V5 Y2 U5 Y1 V4 Y0 U4 */
      movq_r2r (mm1, mm3);        /* Y7 V7 Y6 U7 Y5 V6 Y4 U6 */
      psrlq_i2r (8, mm2);         /* 00 Y3 V5 Y2 U5 Y1 V4 Y0 */
      psrlq_i2r (8, mm3);         /* 00 Y7 V7 Y6 U7 Y5 V6 Y4 */
      pand_r2r (mm7, mm2);        /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      pand_r2r (mm7, mm3);        /* 00 Y7 00 Y6 00 Y5 00 Y4 */
      packuswb_r2r (mm3, mm2);    /* Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 */
      movntq_r2m (mm2, *(yp+8));

      movq_r2r (mm0, mm2);        /* Y3 V5 Y2 U5 Y1 V4 Y0 U4 */
      movq_r2r (mm1, mm3);        /* Y7 V7 Y6 U7 Y5 V6 Y4 U6 */
      pand_r2r (mm6, mm2);        /* 00 00 00 U5 00 00 00 U4 */
      pand_r2r (mm6, mm3);        /* 00 00 00 U7 00 00 00 U6 */
      packuswb_r2r (mm3, mm2);    /* 00 U7 00 U6 00 U5 00 U4 */
      packuswb_r2r (mm2, mm4);    /* U7 U6 U5 U4 U3 U2 U1 U0 */
      movntq_r2m (mm4, *up);

      psrlq_i2r (16, mm0);        /* 00 00 Y3 V5 Y2 U5 Y1 V4 */
      psrlq_i2r (16, mm1);        /* 00 00 Y7 V7 Y6 U7 Y5 V6 */
      pand_r2r (mm6, mm0);        /* 00 00 00 V5 00 00 00 V4 */
      pand_r2r (mm6, mm1);        /* 00 00 00 V7 00 00 00 V6 */
      packuswb_r2r (mm1, mm0);    /* 00 V7 00 V6 00 V5 00 V4 */
      packuswb_r2r (mm0, mm5);    /* V7 V6 V5 V4 V3 V2 V1 V0 */
      movntq_r2m (mm5, *vp);

      yp += 16;
      up+=8;
      vp+=8;
      uyvy += 32;
    }
    for (; w < width; w+=8)
    {
      movq_m2r (*uyvy, mm0);
      movq_m2r (*(uyvy+8), mm1);
      movq_r2r (mm0, mm2);
      movq_r2r (mm1, mm3);
      psrlq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm7, mm0);
      pand_r2r (mm7, mm1);
      packuswb_r2r (mm1, mm0);
      movntq_r2m (mm0, *yp);

      movq_r2r (mm2, mm0);
      movq_r2r (mm3, mm1);
      pand_r2r (mm6, mm0);
      pand_r2r (mm6, mm1);
      packuswb_r2r (mm1, mm0);
      packuswb_r2r (mm1, mm0);
      movd_r2m (mm0, *up);

      psrlq_i2r (16, mm2);
      psrlq_i2r (16, mm3);
      pand_r2r (mm6, mm2);
      pand_r2r (mm6, mm3);
      packuswb_r2r (mm3, mm2);
      packuswb_r2r (mm3, mm2);
      movd_r2m (mm2, *vp);

      yp += 8;
      up+=4;
      vp+=4;
      uyvy += 16;
    }
    yp += next_y;
    up += next_u;
    vp += next_v;
    uyvy += next_uyvy;
  }
  emms();
}

static void csp_yu_to_uy_mmxext (uint8_t *src_planes[], uint8_t *dst_planes[],
          int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, w, tw;

  static mmx_t mask = {0x00FF00FF00FF00FFLL};
  sp = src_planes[0];
  dp = dst_planes[0];

  next_src = (src_stride[0] - width) * 2;
  next_dst = (dst_stride[0] - width) * 2;

  width *= 2;
  tw = width - 32;
  movq_m2r (mask, mm2);
  for (h = height; h > 0; h--)
  {
    for (w = 0; w <= tw; w+=32) {
      movq_m2r (*sp, mm0);       /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */
      movq_r2r (mm0, mm1);       /* V1 Y3 U1 Y2 V0 Y1 U0 Y0 */
      pand_r2r (mm2, mm0);       /* 00 Y3 00 Y2 00 Y1 00 Y0 */
      psllq_i2r (8, mm0);        /* Y3 00 Y2 00 Y1 00 Y0 00 */
      psrlq_i2r (8, mm1);        /* 00 V1 Y3 U1 Y2 V0 Y1 U0 */
      pand_r2r (mm2, mm1);       /* 00 V1 00 U1 00 V0 00 U0 */
      por_r2r (mm1, mm0);        /* Y3 V1 Y2 U1 Y1 V0 Y0 U0 */
      movntq_r2m (mm0, *dp);

      movq_m2r (*(sp+8), mm0);
      movq_r2r (mm0, mm1);
      pand_r2r (mm2, mm0);
      psllq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm2, mm1);
      por_r2r (mm1, mm0);
      movntq_r2m (mm0, *(dp+8));

      movq_m2r (*(sp+16), mm0);
      movq_r2r (mm0, mm1);
      pand_r2r (mm2, mm0);
      psllq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm2, mm1);
      por_r2r (mm1, mm0);
      movntq_r2m (mm0, *(dp+16));

      movq_m2r (*(sp+24), mm0);
      movq_r2r (mm0, mm1);
      pand_r2r (mm2, mm0);
      psllq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm2, mm1);
      por_r2r (mm1, mm0);
      movntq_r2m (mm0, *(dp+24));

      sp += 32;
      dp += 32;
    }
    for (; w < width; w+=8) {
      movq_m2r (*sp, mm0);
      movq_r2r (mm0, mm1);
      pand_r2r (mm2, mm0);
      psllq_i2r (8, mm0);
      psrlq_i2r (8, mm1);
      pand_r2r (mm2, mm1);
      por_r2r (mm1, mm0);
      movntq_r2m (mm0, *dp);

      sp += 8;
      dp += 8;
    }
    sp += next_src;
    dp += next_dst;
  }
  emms();
}

void csp_yuy2_to_uyvy_mmxext(uint8_t *src_planes[],uint8_t *dst_planes[],
          int src_stride[], int dst_stride[], int width, int height)
{
  csp_yu_to_uy_mmxext(src_planes, dst_planes,
          src_stride, dst_stride, width, height);
}

void csp_uyvy_to_yuy2_mmxext(uint8_t *src_planes[],uint8_t *dst_planes[],
          int src_stride[], int dst_stride[], int width, int height)
{
  csp_yu_to_uy_mmxext(src_planes, dst_planes,
          src_stride, dst_stride, width, height);
}



/* YUV444 */

static void csp_yuv444_cpy_mmxext (uint8_t *src444[], uint8_t *dst444[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *sp, *dp;
  int next_src, next_dst;
  int h, w, k;

  for (k = 3; k > 0; k--)
  {
    if (k == 3) {
      sp = src444[0];
      dp = dst444[0];
      next_src = src_stride[0] - width;
      next_dst = dst_stride[0] - width;
    } else if (k == 2) {
      sp = src444[1];
      dp = dst444[1];
      next_src = src_stride[1] - width;
      next_dst = dst_stride[1] - width;
    } else {
      sp = src444[2];
      dp = dst444[2];
      next_src = src_stride[2] - width;
      next_dst = dst_stride[2] - width;
    }
    for (h = height; h > 0; h--)
    {
      for (w = 0; w < width; w+=8) {
	movq_m2r (*sp, mm0);
	movntq_r2m (mm0, *dp);
	sp += 8;
	dp += 8;
      }
      sp += next_src;
      dp += next_dst;
    }
  }
  emms();
}


void csp_yuv444_to_yuv444_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  csp_yuv444_cpy_mmxext (src, dst, src_stride, dst_stride, width, height);
}


#endif /* ARCH_X86 */

