/*
 * yv12_to_rgb.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "csp_conv_yuv_to_rgb.h"
#include "csp_conv_yv12_to_rgb.h"


void csp_yv12_to_argb32 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_argb32(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_rgb24 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb24(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_rgb565 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb565(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_rgb555 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb555(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgra32 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgra32(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgr24 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr24(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgr565 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr565(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgr555 (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr555(planes, dst, stride, dst_stride, width, height);
}

#ifdef ARCH_X86

void csp_yv12_to_argb32_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_argb32_mmxext(planes, dst, stride, dst_stride, width, height);
}

#if 0
void csp_yv12_to_rgb24_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb24_mmxext(planes, dst, stride, dst_stride, width, height);
}
#endif

void csp_yv12_to_rgb565_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb565_mmxext(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_rgb555_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb555_mmxext(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgra32_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgra32_mmxext(planes, dst, stride, dst_stride, width, height);
}

#if 0
void csp_yv12_to_bgr24_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr24_mmxext(planes, dst, stride, dst_stride, width, height);
}
#endif

void csp_yv12_to_bgr565_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr565_mmxext(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgr555_mmxext (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr555_mmxext(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_argb32_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_argb32_mmx(planes, dst, stride, dst_stride, width, height);
}

#if 0
void csp_yv12_to_rgb24_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb24_mmx(planes, dst, stride, dst_stride, width, height);
}
#endif

void csp_yv12_to_rgb565_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb565_mmx(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_rgb555_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_rgb555_mmx(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgra32_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgra32_mmx(planes, dst, stride, dst_stride, width, height);
}

#if 0
void csp_yv12_to_bgr24_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr24_mmx(planes, dst, stride, dst_stride, width, height);
}
#endif

void csp_yv12_to_bgr565_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr565_mmx(planes, dst, stride, dst_stride, width, height);
}

void csp_yv12_to_bgr555_mmx (uint8_t *src[], uint8_t *dst[],
	  int src_stride[], int dst_stride[], int width, int height)
{
  uint8_t *planes[3];
  int stride[3];
  planes[0] = src[0];
  planes[1] = src[2];
  planes[2] = src[1];
  stride[0] = src_stride[0];
  stride[1] = src_stride[2];
  stride[2] = src_stride[1];
  csp_yuv420_to_bgr555_mmx(planes, dst, stride, dst_stride, width, height);
}


#endif /* ARCH_X86 */

