/* ffmt.c */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
//#define __USE_GNU
#include <string.h>
#include <inttypes.h>
#include <errno.h>

#include <qtime.h>

#include "vcodec.h"
#include "vframe_mgr.h"
#include "venc.h"
#include "acodec.h"
#include "aframe_mgr.h"
#include "aenc.h"
#include "ffmt.h"


static const char *default_file_name = "movie";
//static const char *avi_suffix = "avi";
static const char *mov_suffix = "mov";


/* Thread function */

static void ffmt_error(ffmt_t *ffmt, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "ffmt: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);

  if (!ffmt) return;

  pthread_mutex_lock(&ffmt->mutex);
  ffmt->state = AENC_STATE_ERROR;
  pthread_mutex_unlock(&ffmt->mutex);
  if (ffmt->error_cb)
    ffmt->error_cb(ffmt->error_cb_data);

//  exit(1);
}

static int file_close(ffmt_t *ffmt);
static int
file_open(ffmt_t *ffmt)
{
  char name[MAX_FILE_NAME_LENGTH+1];
  struct stat st;

  if (ffmt->file_opend)
    file_close(ffmt);

  if (ffmt->file_num == 0) {
    strncpy(name, ffmt->name, MAX_FILE_NAME_LENGTH);
  } else {
    sprintf(name, "%s.%d", ffmt->name, ffmt->file_num);
  }
  ffmt->file_num++;
  while (stat(name, &st) == 0) {
    sprintf(name, "%s.%d", ffmt->name, ffmt->file_num);
    ffmt->file_num++;
  }

//  switch (ffmt->format) {
//    case FFMT_FORMAT_MOV:
      if (ffmt->qt == NULL) {
        if ((ffmt->qt = qtime_open_write(name)) == NULL) {
          ffmt_error(ffmt, "file_open: mov file open error, %s", name);
          return FFMT_FAIL;
        }

	if (ffmt->venc) {
          if ((ffmt->vqtrk = qtime_create_track(ffmt->qt)) == NULL) {
            ffmt_error(ffmt, "file_open: video track create error, %s", name);
            return FFMT_FAIL;
          }
          qtime_track_set_video(ffmt->vqtrk, ffmt->width, ffmt->height,
	      ffmt->fps, (char*)ffmt->vfourcc_str);
	}

	if (ffmt->aenc) {
          if ((ffmt->aqtrk = qtime_create_track(ffmt->qt)) == NULL) {
            ffmt_error(ffmt, "file_open: audio track create error, %s", name);
            return FFMT_FAIL;
          }
          qtime_track_set_audio(ffmt->aqtrk, ffmt->channels, ffmt->rate,
              ffmt->bitspspl, (char*)ffmt->afourcc_str);
	  if (ffmt->acodec == ACODEC_MP3)
	    qtime_track_set_audio_ext(ffmt->aqtrk, 1, -2, ffmt->aframe_sample_size, 0, 0, (ffmt->bitspspl/8)*ffmt->channels);
	}
      }
//      break;
//    default:
//      return FFMT_FAIL;
//  }
  pthread_mutex_lock(&ffmt->mutex);
  ffmt->record_time = 0;
  ffmt->audio_samples = 0;
  ffmt->video_frame_count = 0;
  ffmt->audio_frame_count = 0;
  ffmt->video_drop_frames = 0;
  ffmt->video_data_size = 0;
  ffmt->audio_data_size = 0;
  ffmt->total_data_size = 0;
  ffmt->file_opend = 0;
  pthread_mutex_unlock(&ffmt->mutex);
  ffmt->file_opend = 1;

fprintf(stdout, "\noutput file name: %s\n", name);

  return FFMT_OK;
}

static int
file_close(ffmt_t *ffmt)
{
  double fps, rec_time, samples;
//  long long file_size;
//  const char *unit;

  if (!ffmt->file_opend)
    return FFMT_OK;

  if (ffmt->venc) {
    if (ffmt->aenc && ffmt->sync_audio_flag) {
      samples = ffmt->audio_samples;
      rec_time = samples / (double)ffmt->rate;
      fps = (double)(ffmt->video_frame_count) / rec_time;
    } else if (ffmt->real_time_flag) {
      rec_time = ffmt->record_time / 1000000.0;
      fps = (double)(ffmt->video_frame_count) / rec_time;
    } else {
      fps = ffmt->fps;
    }
#ifndef NDEBUG
fprintf(stderr, "file_close: video_frame_count %d, record_time %llu, fps %f\n", ffmt->video_frame_count, ffmt->record_time, fps);
#endif /* NDEBUG */

    qtime_track_set_video_fps(ffmt->vqtrk, fps);
  }

  qtime_close(ffmt->qt);
  ffmt->qt = NULL;
  ffmt->vqtrk = NULL;
  ffmt->aqtrk = NULL;
#if 0
  switch (ffmt->format) {
    case FFMT_FORMAT_MOV:
      qtime_track_set_video_fps(ffmt->vqtrk, fps);
      qtime_close(ffmt->qt);
      ffmt->qt = NULL;
      ffmt->vqtrk = NULL;
      ffmt->aqtrk = NULL;
      break;
    default:
      return FFMT_FAIL;
  }
#endif
  fprintf(stdout, "\nfile write:  %s", ffmt->name);
  if (ffmt->file_num > 1)
    fprintf(stdout, ".%d\n", ffmt->file_num-1);
  else
    fprintf(stdout, "\n");
#if 0
  file_size = ffmt->total_data_size + ffmt->total_idx_size;
  if (file_size < 1024)
    unit = "";
  else {
    file_size /= 1024;
    if (file_size < 1024)
      unit = "k";
    else {
      file_size /= 1024;
      unit = "M";
    }
  }
  fprintf(stdout, "file size:  %lld%s bytes\n", file_size, unit);
  fprintf(stdout, "fps: %f\n", fps);
#endif

  pthread_mutex_lock(&ffmt->mutex);
  ffmt->record_time = 0;
  ffmt->audio_samples = 0;
  ffmt->video_drop_frames = 0;
  ffmt->video_frame_count = 0;
  ffmt->audio_frame_count = 0;
  ffmt->video_data_size = 0;
  ffmt->audio_data_size = 0;
  ffmt->total_data_size = 0;
  ffmt->file_opend = 0;
  pthread_mutex_unlock(&ffmt->mutex);
  return FFMT_OK;
}

static int
write_vframe(ffmt_t *ffmt, vframe_t *vfrm)
{
  int ret = 0;
  int is_key_frame = 0;

  if (vfrm->flag & VFRAME_FLAG_DROP) {
    ffmt->video_drop_frames++;
    fprintf(stderr, "frame %d drop, drop total %d\n", ffmt->video_frame_count, ffmt->video_drop_frames);
    vfrm->data_length = 0;
  }
#ifndef NDEBUG
  if (vfrm->data_length <= 0) {
    fprintf(stderr, "write_vframe: data %p\n", vfrm->data);
    ffmt_error(ffmt, "write_vframe: frame write size %d, flag %u\n", vfrm->data_length, vfrm->flag);
    return FFMT_FAIL;
  }
#endif

//  if (ffmt->format == FFMT_FORMAT_MOV) {
    if (vfrm->flag & VFRAME_FLAG_INTRA)
      is_key_frame = 1;
    else
      is_key_frame = 0;
    qtime_track_write_video(ffmt->vqtrk,
       	vfrm->data, vfrm->data_length, is_key_frame);
//  }

  pthread_mutex_lock(&ffmt->mutex);
  ffmt->video_frame_count++;
  ffmt->video_data_size += vfrm->data_length;
  ffmt->total_data_size += vfrm->data_length;
  ffmt->record_time += vfrm->duration;
  pthread_mutex_unlock(&ffmt->mutex);

  return ret;
}

static int
write_aframe(ffmt_t *ffmt, aframe_t *afrm)
{
  int ret = 0;

#ifndef NDEBUG
  if (afrm->stream_length <= 0)
    fprintf(stderr, "write_aframe: frame write size %d, flag %x\n", afrm->stream_length, afrm->flag);
#endif

//  if (ffmt->format == FFMT_FORMAT_MOV && afrm->stream_length > 0) {
  if (afrm->stream_length > 0) {
    qtime_track_write_audio(ffmt->aqtrk, afrm->stream, afrm->stream_length, afrm->samples);

    pthread_mutex_lock(&ffmt->mutex);
    ffmt->audio_frame_count++;
    ffmt->audio_data_size += afrm->stream_length;
    ffmt->total_data_size += afrm->stream_length;
    ffmt->audio_samples += afrm->samples;
    pthread_mutex_unlock(&ffmt->mutex);
    return afrm->stream_length;
  } else {
    ffmt_error(ffmt, "invalid aframe.\n");
    ret = FFMT_FAIL;
  }

  return ret;
}

static inline int
get_state(ffmt_t *ffmt)
{
  int state;
  pthread_mutex_lock(&ffmt->mutex);
  state = ffmt->state;
  pthread_mutex_unlock(&ffmt->mutex);
  return state;
}

static void
state_change_done(ffmt_t *ffmt)
{
  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_INTERFACE_WAIT) {
    ffmt->state &= ~FFMT_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&ffmt->interface_wait_cond);
  }
  pthread_mutex_unlock(&ffmt->mutex);
}

static void
set_state(ffmt_t *ffmt, int state)
{
  pthread_mutex_lock(&ffmt->mutex);
  ffmt->state &= ~(FFMT_STATE_START|FFMT_STATE_PAUSE|FFMT_STATE_STOP|FFMT_STATE_QUIT);
  ffmt->state |= state;
  if (ffmt->state & FFMT_STATE_INTERFACE_WAIT) {
    ffmt->state &= ~FFMT_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&ffmt->interface_wait_cond);
  }
  pthread_mutex_unlock(&ffmt->mutex);
}

static vframe_t*
sync_vframe(ffmt_t *ffmt, vframe_t *vfrm, uint32_t flag)
{
  while (!vfrm || !(vfrm->flag & flag)) {
    if (!vfrm) {
      pthread_mutex_lock(&ffmt->mutex);
      while ((vfrm = venc_get_vframe(ffmt->venc)) == NULL) {
	if (!(ffmt->state & FFMT_STATE_ERROR)) {
          ffmt->state |= FFMT_STATE_THREAD_WAIT;
          pthread_cond_wait(&ffmt->thread_wait_cond, &ffmt->mutex);
	}
        if (ffmt->state & FFMT_STATE_ERROR) {
          pthread_mutex_unlock(&ffmt->mutex);
	  return NULL;
        }
      }
      pthread_mutex_unlock(&ffmt->mutex);
    }

#ifndef NDEBUG
    if (vfrm->flag & ((VFRAME_FLAG_START|VFRAME_FLAG_END|VFRAME_FLAG_QUIT) & ~flag)) {
      ffmt_error(ffmt, "sync_vframe: invalid flag frame, %x.\n", vfrm->flag);
      exit(1);
    }
#endif /* NDEBUG */

    if (vfrm->flag & VFRAME_FLAG_NORMAL) {
      write_vframe(ffmt, vfrm);
      vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
      vfrm = NULL;
    } else {
      if (vfrm->flag & VFRAME_FLAG_DROP) {
        write_vframe(ffmt, vfrm);
        vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
        vfrm = NULL;
      }
#ifndef NDEBUG
      else if (!(vfrm->flag & flag)) {
        ffmt_error(ffmt, "sync_vframe: invalid ext frame, %x.\n", vfrm->flag);
        exit(1);
      }
#endif /* NDEBUG */
    }
  }

  return vfrm;
}

static aframe_t*
sync_aframe(ffmt_t *ffmt, aframe_t *afrm, uint32_t flag)
{
  while (!afrm || !(afrm->flag & flag)) {
    if (!afrm) {
      pthread_mutex_lock(&ffmt->mutex);
      while ((afrm = aenc_get_aframe(ffmt->aenc)) == NULL) {
	if (!(ffmt->state & FFMT_STATE_ERROR)) {
          ffmt->state |= FFMT_STATE_THREAD_WAIT;
          pthread_cond_wait(&ffmt->thread_wait_cond, &ffmt->mutex);
	}
        if (ffmt->state & FFMT_STATE_ERROR) {
          pthread_mutex_unlock(&ffmt->mutex);
	  return NULL;
        }
      }
      pthread_mutex_unlock(&ffmt->mutex);
    }

    if (afrm->flag & AFRAME_FLAG_NORMAL) {
      write_aframe(ffmt, afrm);
      aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
      afrm = NULL;
    }
    else {
      if (!(afrm->flag & flag)) {
        ffmt_error(ffmt, "sync_aframe: invalid ext frame, %x, %x.\n", flag, afrm->flag);
        aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
        afrm = NULL;
      }
    }
  }

  return afrm;
}

/* Thread */

static void*
ffmt_thread (void *data)
{
  ffmt_t *ffmt = (ffmt_t*)data;
  vframe_t *vfrm = NULL;
  aframe_t *afrm = NULL;
  int quit_flag = 0;
  uint32_t state;

  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_INTERFACE_WAIT) {
    pthread_cond_signal(&ffmt->interface_wait_cond);
  }
  ffmt->state = FFMT_STATE_STOP;
  pthread_mutex_unlock(&ffmt->mutex);

  while (1) {
    state = get_state(ffmt);

    if (quit_flag || state & FFMT_STATE_ERROR) {
      break;
    }

    do {
      if (ffmt->aenc)
        afrm = aenc_get_aframe(ffmt->aenc);
      if (ffmt->venc)
        vfrm = venc_get_vframe(ffmt->venc);
      if (afrm || vfrm || ffmt->state & FFMT_STATE_ERROR)
	break;
      pthread_mutex_lock(&ffmt->mutex);
      ffmt->state |= FFMT_STATE_THREAD_WAIT;
      pthread_cond_wait(&ffmt->thread_wait_cond, &ffmt->mutex);
      state = ffmt->state;
      pthread_mutex_unlock(&ffmt->mutex);
    } while (!(state & FFMT_STATE_ERROR));

    if (ffmt->state & FFMT_STATE_ERROR)
      break;

    if (vfrm) {
      if (vfrm->flag & VFRAME_FLAG_START) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: start video frame.\n");
#endif /* NDEBUG */
	if (ffmt->aenc) {
	  afrm = sync_aframe(ffmt, afrm, AFRAME_FLAG_START);
	  if (!afrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	  afrm = NULL;
	}
	file_close(ffmt);
	file_open(ffmt);
	set_state(ffmt, FFMT_STATE_START);
	state_change_done(ffmt);
      } else if (vfrm->flag & VFRAME_FLAG_END) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: end video frame.\n");
#endif /* NDEBUG */
	if (ffmt->aenc) {
	  afrm = sync_aframe(ffmt, afrm, AFRAME_FLAG_END);
	  if (!afrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	  afrm = NULL;
	}
	file_close(ffmt);
	set_state(ffmt, FFMT_STATE_STOP);
	state_change_done(ffmt);
      } else if (vfrm->flag & VFRAME_FLAG_QUIT) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: quit video frame.\n");
#endif /* NDEBUG */
	if (ffmt->aenc) {
	  afrm = sync_aframe(ffmt, afrm, AFRAME_FLAG_QUIT);
	  if (!afrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	  afrm = NULL;
	}
	vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
	vfrm = NULL;
	file_close(ffmt);
	quit_flag = 1;
	set_state(ffmt, FFMT_STATE_QUIT);
	break;
      } else if (vfrm->flag & VFRAME_FLAG_ERROR) {
	file_close(ffmt);
	if (afrm) {
	  aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	  afrm = NULL;
	}
	vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
	vfrm = NULL;
	quit_flag = 1;
	set_state(ffmt, FFMT_STATE_ERROR);
	break;
      } else {
	write_vframe(ffmt, vfrm);
      }
      vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
      vfrm = NULL;
    }

    if (afrm) {
      if (afrm->flag & AFRAME_FLAG_START) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: start audio frame.\n");
#endif /* NDEBUG */
	if (ffmt->venc) {
	  vfrm = sync_vframe(ffmt, vfrm, VFRAME_FLAG_START);
	  if (!vfrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
	  vfrm = NULL;
	}
	file_close(ffmt);
	file_open(ffmt);
	set_state(ffmt, FFMT_STATE_START);
	state_change_done(ffmt);
      } else if (afrm->flag & AFRAME_FLAG_END) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: end audio frame.\n");
#endif /* NDEBUG */
	if (ffmt->venc) {
	  vfrm = sync_vframe(ffmt, vfrm, VFRAME_FLAG_END);
	  if (!vfrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
	  vfrm = NULL;
        }
	file_close(ffmt);
	set_state(ffmt, FFMT_STATE_STOP);
	state_change_done(ffmt);
      } else if (afrm->flag & AFRAME_FLAG_QUIT) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: quit audio frame.\n");
#endif /* NDEBUG */
	if (ffmt->venc) {
	  vfrm = sync_vframe(ffmt, vfrm, VFRAME_FLAG_QUIT);
	  if (!vfrm) {
	    set_state(ffmt, FFMT_STATE_ERROR);
	    break;
	  }
	  vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
	  vfrm = NULL;
        }
	file_close(ffmt);
	set_state(ffmt, FFMT_STATE_QUIT);
	aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	afrm = NULL;
	quit_flag = 1;
	break;
      } else if (afrm->flag & AFRAME_FLAG_ERROR) {
#ifndef NDEBUG
        fprintf(stderr, "ffmt_thread: error audio frame.\n");
#endif /* NDEBUG */
	file_close(ffmt);
	set_state(ffmt, FFMT_STATE_ERROR);
	aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
	afrm = NULL;
	quit_flag = 1;
	break;
      } else {
	write_aframe(ffmt, afrm);
      }
      aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
      afrm = NULL;
    }
  }

  file_close(ffmt);

  if (vfrm) {
    vframe_mgr_add_free_frame(ffmt->vfrm_mgr, vfrm);
    ffmt_error(ffmt, "vframe remains.\n");
  }
  if (afrm) {
    aframe_mgr_add_free_frame(ffmt->afrm_mgr, afrm);
    ffmt_error(ffmt, "aframe remains.\n");
  }

  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_ERROR)
    fprintf(stderr, "ffmt_thread: STATE_ERROR break\n");
  if (ffmt->state & FFMT_STATE_INTERFACE_WAIT) {
    ffmt->state &= ~FFMT_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&ffmt->interface_wait_cond);
  }
  ffmt->state &= (FFMT_STATE_QUIT|AENC_STATE_ERROR);
  ffmt->state |= FFMT_STATE_THREAD_FINISH;
  pthread_mutex_unlock(&ffmt->mutex);
//  state_change_done(ffmt);

#ifndef NDEBUG
  fprintf(stderr, "ffmt_thread: done\n");
#endif /* NDEBUG */
  return ffmt;
}

/* Interface functions */

double
ffmt_get_bitrate(ffmt_t *ffmt)
{
  double data, time, bitrate;

  pthread_mutex_lock(&ffmt->mutex);
  data = ffmt->total_data_size;
  if (!ffmt->venc || ffmt->sync_audio_flag)
    time = (double)ffmt->audio_samples / (double)ffmt->rate;
  else
    time = (double)ffmt->record_time / U_TIME_SECOND;
  pthread_mutex_unlock(&ffmt->mutex);

  if (time == 0) {
    bitrate = 0;
  } else {
    data = data * 8;
    bitrate = data / time;
  }
  return bitrate;
}

long long
ffmt_get_total_data_size(ffmt_t *ffmt)
{
  long long size;

  pthread_mutex_lock(&ffmt->mutex);
  size = ffmt->total_data_size;
  pthread_mutex_unlock(&ffmt->mutex);
  return size;
}

int
ffmt_get_video_frame_count(ffmt_t *ffmt)
{
  int count;
  pthread_mutex_lock(&ffmt->mutex);
  count = ffmt->video_frame_count;
  pthread_mutex_unlock(&ffmt->mutex);
  return count;
}

double
ffmt_get_record_time(ffmt_t *ffmt)
{
  double time;

  pthread_mutex_lock(&ffmt->mutex);
  if (!ffmt->venc || ffmt->sync_audio_flag)
    time = (double)ffmt->audio_samples / (double)ffmt->rate;
  else
    time = (double)ffmt->record_time / U_TIME_SECOND;
  pthread_mutex_unlock(&ffmt->mutex);
  return time;
}

static int
ffmt_frame_ready_cb(void *cb_data)
{
  ffmt_t *ffmt = (ffmt_t*)cb_data;

  if (pthread_mutex_trylock(&ffmt->wait_cb_mutex) == EBUSY)
    return FFMT_FAIL;
  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_THREAD_WAIT) {
    ffmt->state &= ~FFMT_STATE_THREAD_WAIT;
    pthread_cond_signal(&ffmt->thread_wait_cond);
  }
  pthread_mutex_unlock(&ffmt->mutex);
  pthread_mutex_unlock(&ffmt->wait_cb_mutex);

  return FFMT_OK;
}

int
ffmt_set_vframe_mgr(ffmt_t *ffmt, vframe_mgr_t *vfrm_mgr)
{
  pthread_mutex_lock(&ffmt->mutex);
  ffmt->vfrm_mgr = vfrm_mgr;
  pthread_mutex_unlock(&ffmt->mutex);
  return VENC_OK;
}

int
ffmt_set_aframe_mgr(ffmt_t *ffmt, aframe_mgr_t *afrm_mgr)
{
  pthread_mutex_lock(&ffmt->mutex);
  ffmt->afrm_mgr = afrm_mgr;
  pthread_mutex_unlock(&ffmt->mutex);
  return VENC_OK;
}

int
ffmt_set_error_cb(ffmt_t *ffmt, error_cb_t error_cb, void *error_cb_data)
{
  pthread_mutex_lock(&ffmt->mutex);
  ffmt->error_cb = error_cb;
  ffmt->error_cb_data = error_cb_data;
  pthread_mutex_unlock(&ffmt->mutex);
  return FFMT_OK;
}

int
ffmt_state_error(ffmt_t *ffmt)
{
  int ret = FFMT_FAIL;

  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_THREAD_WAIT) {
    ffmt->state &= ~FFMT_STATE_THREAD_WAIT;
    pthread_cond_signal(&ffmt->thread_wait_cond);
  }
  if (ffmt->state & (FFMT_STATE_ERROR|FFMT_STATE_THREAD_FINISH)) {
    ret = FFMT_OK;
  } else {
    ffmt->state &= ~(FFMT_STATE_START|FFMT_STATE_PAUSE|FFMT_STATE_STOP);
    ffmt->state |= FFMT_STATE_ERROR;
#if 0
    ffmt->state |= FFMT_STATE_INTERFACE_WAIT;
    while (ffmt->state & FFMT_STATE_INTERFACE_WAIT)
      pthread_cond_wait(&ffmt->interface_wait_cond, &ffmt->mutex);
#endif
    ret = FFMT_OK;
  }
  pthread_mutex_unlock(&ffmt->mutex);
  return ret;
}

static int
ffmt_state_change_wait(ffmt_t *ffmt, uint32_t state)
{
  int ret = FFMT_FAIL;
  pthread_mutex_lock(&ffmt->mutex);
  if (ffmt->state & FFMT_STATE_THREAD_WAIT) {
    ffmt->state &= ~FFMT_STATE_THREAD_WAIT;
    pthread_cond_signal(&ffmt->thread_wait_cond);
  }
  if (ffmt->state & (state|FFMT_STATE_QUIT|FFMT_STATE_ERROR|FFMT_STATE_THREAD_FINISH))
    ret = FFMT_OK;
  else {
    while (!(ffmt->state & state)) {
      ffmt->state |= FFMT_STATE_INTERFACE_WAIT;
      pthread_cond_wait(&ffmt->interface_wait_cond, &ffmt->mutex);
    }
    ret = FFMT_OK;
  }
  pthread_mutex_unlock(&ffmt->mutex);
  return ret;
}

int
ffmt_state_start(ffmt_t *ffmt)
{
  return ffmt_state_change_wait(ffmt, FFMT_STATE_START);
}

int
ffmt_state_stop(ffmt_t *ffmt)
{
  return ffmt_state_change_wait(ffmt, FFMT_STATE_STOP);
}

int
ffmt_state_quit(ffmt_t *ffmt)
{
  return ffmt_state_change_wait(ffmt, FFMT_STATE_QUIT);
}

int
ffmt_state_toggle_pause(ffmt_t *ffmt)
{
  if (!ffmt) return FFMT_FAIL;
  return FFMT_OK;
}

int
ffmt_error_quit (ffmt_t *ffmt)
{
  int retcode, state;
  void* retval;

  ffmt_state_error(ffmt);

  file_close(ffmt);

  state = get_state(ffmt);
  if (!(state & FFMT_STATE_THREAD_FINISH) &&
      !(state & FFMT_STATE_NOT_READY)) {
    retcode = pthread_join(ffmt->ffmt_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "ffmt_quit: thread join error %d\n", retcode);
  }

  if (ffmt->name)
    free(ffmt->name);
  free(ffmt);

//  fprintf(stderr, "ffmt_quit: done\n");
  return FFMT_OK;
}

int
ffmt_quit (ffmt_t *ffmt)
{
  int retcode, state;
  void* retval;

  ffmt_state_quit(ffmt);

  file_close(ffmt);

  state = get_state(ffmt);
  if (!(state & FFMT_STATE_THREAD_FINISH) &&
      !(state & FFMT_STATE_NOT_READY)) {
    retcode = pthread_join(ffmt->ffmt_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "ffmt_quit: thread join error %d\n", retcode);
  }

  if (ffmt->name)
    free(ffmt->name);
  free(ffmt);

//  fprintf(stderr, "ffmt_quit: done\n");
  return FFMT_OK;
}

ffmt_t*
ffmt_init (char *name, int width, int height, double fps, unsigned long vcodec, const char *vfourcc_str, int acodec, int rate, int channels, int bitspspl, const char *afourcc_str, int sync_audio, venc_t *venc, aenc_t *aenc)
{
  char buf[MAX_FILE_NAME_LENGTH];
  int retcode;
  ffmt_t *ffmt;

  ffmt = (ffmt_t*)malloc(sizeof(ffmt_t));
  if (!ffmt) {
    ffmt_error(ffmt, "memory error.\n");
    return NULL;
  }

  ffmt->state = FFMT_STATE_NOT_READY;
  ffmt->name = NULL;
  ffmt->file_num = 0;
  ffmt->file_opend = 0;
  ffmt->error_cb = NULL;
  ffmt->error_cb_data = NULL;

  ffmt->qt = NULL;
  ffmt->vqtrk = NULL;
  ffmt->aqtrk = NULL;

  if (name == NULL) {
#if 0
    const char *suff;
    switch (ffmt->format) {
      case FFMT_FORMAT_AVI:
	suff = avi_suffix;
	break;
      case FFMT_FORMAT_RAW:
	suff = raw_suffix;
	break;
      case FFMT_FORMAT_MOV:
	suff = mov_suffix;
	break;
      default:
	suff = "xxx";
	break;
    }
    sprintf(buf, "%s.%s", default_file_name, suff);
#endif
    sprintf(buf, "%s.%s", default_file_name, mov_suffix);
  }
  else {
    strncpy(buf, name, MAX_FILE_NAME_LENGTH-8);
  }
  ffmt->name = (char*)strndup(buf, MAX_FILE_NAME_LENGTH-8);

  ffmt->width = width;
  ffmt->height = height;
  ffmt->fps = fps;
  ffmt->vcodec = vcodec;
  if (venc && !vfourcc_str)
    ffmt->vfourcc_str = vcodec_codec_to_fourcc_str(vcodec);
  else
    ffmt->vfourcc_str = vfourcc_str;
  ffmt->real_time_flag = 0;
  if (fps == 0)
    ffmt->real_time_flag = 1;

  ffmt->acodec = acodec;
  ffmt->rate = rate;
  ffmt->channels = channels;
  ffmt->bitspspl = bitspspl;
  if (aenc && !afourcc_str)
    ffmt->afourcc_str = acodec_codec_to_fourcc(acodec);
  else
    ffmt->afourcc_str = afourcc_str;
  if (aenc)
    ffmt->aframe_sample_size = acodec_encode_get_frame_size();

  ffmt->vfrm_mgr = NULL;
  ffmt->venc = venc;
  ffmt->afrm_mgr = NULL;
  ffmt->aenc = aenc;

  ffmt->record_time = 0;
  ffmt->audio_samples = 0;
  ffmt->video_drop_frames = 0;
  ffmt->video_frame_count = 0;
  ffmt->audio_frame_count = 0;
  ffmt->video_data_size = 0;
  ffmt->audio_data_size = 0;
  ffmt->total_data_size = 0;

  ffmt->sync_audio_flag = 0;
  if (venc && aenc && sync_audio)
    ffmt->sync_audio_flag = 1;

  pthread_mutex_init (&ffmt->mutex, NULL);
  pthread_mutex_init (&ffmt->wait_cb_mutex, NULL);
  pthread_cond_init (&ffmt->interface_wait_cond, NULL);
  pthread_cond_init (&ffmt->thread_wait_cond, NULL);

  if (venc)
    venc_set_vframe_ready_cb(ffmt->venc, ffmt_frame_ready_cb, ffmt);
  if (aenc)
    aenc_set_aframe_ready_cb(ffmt->aenc, ffmt_frame_ready_cb, ffmt);

  retcode = pthread_create (&ffmt->ffmt_th, NULL, ffmt_thread, ffmt);
  if (retcode != 0) {
    ffmt_error(ffmt, "ffmt_init() thread create failed %d\n", retcode);
    return NULL;
  }

  pthread_mutex_lock(&ffmt->mutex);
  if (!ffmt->state & FFMT_STATE_NOT_READY) {
    ffmt->state |= FFMT_STATE_INTERFACE_WAIT;
    pthread_cond_wait(&ffmt->interface_wait_cond, &ffmt->mutex);
  }
  pthread_mutex_unlock(&ffmt->mutex);

  return ffmt;
}

void
ffmt_print_param(ffmt_t *ffmt)
{
  fprintf (stderr, "file name:              %s\n", ffmt->name);
}

