/* ffmt.h */

#ifndef FFMT_H
#define FFMT_H

#include <inttypes.h>
#include <limits.h>
#include <qtime.h>

#include "get_cur_u_time.h"
#include "vframe_mgr.h"
#include "venc.h"
#include "aframe_mgr.h"
#include "aenc.h"
#include "error_mgr.h"


#define MAX_FILE_NAME_LENGTH 1024


#define FFMT_STATE_NOT_READY       (1UL<<0)
#define FFMT_STATE_STOP            (1UL<<1)
#define FFMT_STATE_START           (1UL<<2)
#define FFMT_STATE_PAUSE           (1UL<<3)
#define FFMT_STATE_QUIT            (1UL<<4)
#define FFMT_STATE_ERROR           (1UL<<5)
#define FFMT_STATE_INTERFACE_WAIT  (1UL<<6)
#define FFMT_STATE_THREAD_WAIT     (1UL<<7)
#define FFMT_STATE_THREAD_FINISH   (1UL<<8)

typedef struct {
  pthread_t ffmt_th;
  pthread_mutex_t mutex;
  pthread_mutex_t wait_cb_mutex;
  pthread_cond_t thread_wait_cond;
  pthread_cond_t interface_wait_cond;

  uint32_t state;

  char *name;
  int file_num;
  int file_opend;
//  int format;

  qtime_t *qt;
  qtime_track_t *vqtrk;
  qtime_track_t *aqtrk;
  int real_time_flag;
  int sync_audio_flag;

  vframe_mgr_t *vfrm_mgr;
  aframe_mgr_t *afrm_mgr;

  venc_t *venc;
  aenc_t *aenc;

  unsigned long vcodec;
  const char *vfourcc_str;
  int    width;
  int    height;
  double fps;

  int acodec;
  const char *afourcc_str;
  int rate;
  int channels;
  int bitspspl;
  int aframe_sample_size;

  u_time_t record_time;
//  long long total_sample;
  long long audio_samples;
  int video_drop_frames;
  int video_frame_count;
  int audio_frame_count;
  long long video_data_size;
  long long audio_data_size;
  long long total_data_size;

  error_cb_t error_cb;
  void* error_cb_data;
} ffmt_t;

#define FFMT_OK    0
#define FFMT_FAIL -1


double ffmt_get_bitrate(ffmt_t *ffmt);
long long ffmt_get_total_data_size(ffmt_t *ffmt);
int ffmt_get_video_frame_count(ffmt_t *ffmt);
double ffmt_get_record_time(ffmt_t *ffmt);
int ffmt_set_vframe_mgr(ffmt_t *ffmt, vframe_mgr_t *vfrm_mgr);
int ffmt_set_aframe_mgr(ffmt_t *ffmt, aframe_mgr_t *afrm_mgr);
int ffmt_set_error_cb(ffmt_t *ffmt, error_cb_t error_cb, void *error_cb_data);
int ffmt_state_error(ffmt_t *ffmt);
int ffmt_state_start(ffmt_t *ffmt);
int ffmt_state_stop(ffmt_t *ffmt);
int ffmt_state_quit(ffmt_t *ffmt);
int ffmt_state_toggle_pause(ffmt_t *ffmt);
int ffmt_quit (ffmt_t *ffmt);
int ffmt_error_quit (ffmt_t *ffmt);
ffmt_t *ffmt_init (char *name, int width, int height, double fps, unsigned long vcodec, const char *vforcc_str, int acodec, int rate, int channels, int bitspspl, const char *afourcc_str, int sync_audio, venc_t *venc, aenc_t *aenc);
void ffmt_print_param(ffmt_t *ffmt);

#endif /* FFMT_H */

