/* mgr.h */

#ifndef MGR_H
#define MGR_H

#include <inttypes.h>
#include <pthread.h>

#include "vframe_mgr.h"
#include "vgrab.h"
#include "venc.h"
#include "aframe_mgr.h"
#include "aread.h"
#include "aenc.h"
#include "ffmt.h"


#define MGR_STATE_NOT_READY       (1UL<<0)
#define MGR_STATE_STOP            (1UL<<1)
#define MGR_STATE_START           (1UL<<2)
#define MGR_STATE_PAUSE           (1UL<<3)
#define MGR_STATE_QUIT            (1UL<<4)
#define MGR_STATE_ERROR           (1UL<<5)
#define MGR_STATE_INTERFACE_WAIT  (1UL<<6)
#define MGR_STATE_THREAD_WAIT     (1UL<<7)


#define MGR_THREAD   (1UL<<0)
#define VGRAB_THREAD (1UL<<1)
#define VENC_THREAD  (1UL<<2)
#define AREAD_THREAD (1UL<<3)
#define AENC_THREAD  (1UL<<4)
#define FFMT_THREAD  (1UL<<5)

typedef struct {
  int width;
  int height;
  double fps;
  int grab_csp;
  int out_csp;
  int vcodec;
  int max_buffers;
  const char *use_vfourcc;

  const char *vdev_name;
  int vdev_source_channel;
  int vdev_tuner;
  int vdev_norm;
  int vdev_audio;
  int vdev_brightness;
  int vdev_hue;
  int vdev_colour;
  int vdev_contrast;
  int vdev_whiteness;
  int vdev_freq;

  int acodec;
  int audio_rate;
  int audio_out_rate;
  int audio_channels;
  int audio_bits;
  int audio_frame_sample_size;
  const char *use_afourcc;

  const char *adev_name;
  const char *adev_mixer_name;
  const char *adev_source;

#define SYNC_NONE   0
#define SYNC_SYSTEM 1
#define SYNC_AUDIO  2
  int sync_mode;

#define BOTH_MEDIA  0
#define VIDEO_MEDIA 1
#define AUDIO_MEDIA 2
  int record_media;

  char *output_fname;
  int file_format;
} MGR_PARAM;

typedef struct {
  pthread_mutex_t mutex;
  uint32_t state;
  uint32_t error_state;

  vgrab_t *vgrab;
  aread_t *aread;
  venc_t *venc;
  aenc_t *aenc;
  ffmt_t *ffmt;

  vframe_mgr_t *vfrm_mgr;
  aframe_mgr_t *afrm_mgr;
} mgr_t;

#define MGR_OK    0
#define MGR_FAIL -1

double mgr_get_bitrate(mgr_t *mgr);
double mgr_get_record_time(mgr_t *mgr);
long long mgr_get_total_data_size(mgr_t *mgr);
int mgr_get_video_frame_count(mgr_t *mgr);
int mgr_get_record_media(mgr_t *mgr);
int mgr_set_output_file_name (const char *arg);

int mgr_error_check(mgr_t *mgr);
int mgr_state_start(mgr_t *mgr);
int mgr_state_toggle_pause(mgr_t *mgr);
int mgr_state_stop(mgr_t *mgr);
int mgr_state_quit(mgr_t *mgr);
int mgr_state_error(mgr_t *mgr);
int mgr_error_quit(mgr_t *mgr);
int mgr_quit(mgr_t *mgr);
mgr_t* mgr_init(void);

void mgr_print_param(mgr_t *mgr);


#endif /* MGR_H */

