/* deno2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <malloc.h>
#include <inttypes.h>

#define ARCH_X86

#ifdef ARCH_X86
#include "attributes.h"
#include "mmx.h"
#else
#undef emms
#define emms()
#endif /* ARCH_X86 */


#include "denoise3.h"

#define BUF_NUM 3

typedef struct VFRAME_T {
  int frame_width[3];
  int frame_height[3];
  int horz_border[3];
  int vert_border[3];
  int width[3];
  int height[3];
  int stride[3];
  int csp;
  unsigned char *bufp;
  unsigned char *plane[3];
  unsigned char *offset[3];
} vframe_t;

typedef struct {
  int x, y;
  int mvx, mvy;
  unsigned int sad;
} motion_data_t;

typedef struct {
  unsigned int vari_avgthre;
  unsigned int vari_motthre;

  unsigned int vari_bk;
  unsigned int vari_fw;

  unsigned int vari_mbk;
  unsigned int vari_mfw;
  motion_data_t *mibk;
  motion_data_t *mifw;

#define BLOCK_FLAG_NONE                0
#define BLOCK_FLAG_AVG                 (1UL<<0)
#define BLOCK_FLAG_MOTION              (1UL<<1)
#define BLOCK_FLAG_BACK_AVG            (1UL<<2)
#define BLOCK_FLAG_FORW_AVG            (1UL<<3)
#define BLOCK_FLAG_MOTION_FORW_AVG     (1UL<<4)
#define BLOCK_FLAG_MOTION_BACK_AVG     (1UL<<5)
  unsigned int flag;
  int errcnt;
} block_info_t;

#define DEBUG

#ifdef DEBUG
#define PIX_CLIP(v) pix_clip_check((v))
#else
#define PIX_CLIP(v) pix_clip[(v)]
#endif

/* lowpass variable */
typedef double (*FILT_FUNC)(double x);
typedef void (*lowpass_func_t)(int , int , unsigned char *, int , unsigned char*, int , int *);
static int filt_coef0[5];
static int filt_coef1[5];
static lowpass_func_t lowpass_func0;
static lowpass_func_t lowpass_func1;


/* pixel range clip */
static unsigned char *pix_clip = NULL;
static unsigned char *pix_clip_buf = NULL;

static int uninited = 1;

/* parameters */
static int Threshold;
static int Threshold_uv;
static int Avg_threshold = 64;
static int Motion_threshold = 64;
static int Motion_radius = 8;
static int Field_flag = 0;

static int Threshold2;
static int Width, Height;
static int Width2, Height2;
//static int Width4, Height4;

/* motion search variables */
static unsigned char *bufp2;
static unsigned char *bufp4;
static unsigned char *sub2[3];
static unsigned char *sub4[3];
static motion_data_t *motion_data_back[3];
static motion_data_t *motion_data_forw[3];

static block_info_t *block_info[3];

static vframe_t *vframe[BUF_NUM];
#define PIX_FLAG_NONE               0
#define PIX_FLAG_BACK_AVG           (1<<0)
#define PIX_FLAG_FORW_AVG           (1<<1)
#define PIX_FLAG_MOTION_BACK_AVG    (1<<2)
#define PIX_FLAG_MOTION_FORW_AVG    (1<<3)
#define PIX_FLAG_DONE               (1<<4)
static unsigned char *pix_flag[3];
static unsigned char *mref_frame;
static unsigned char *mref_frame_forward;


/* motion search helper functions */
static uint32_t (*Calc_SAD)  (uint8_t *, int, uint8_t *, int, uint32_t);
static uint32_t (*Calc_SAD_half2)  (uint8_t *, int, uint8_t *, uint8_t*, int, uint32_t);
static uint32_t (*Calc_SAD_half4)  (uint8_t *, int, uint8_t *, uint8_t*, uint8_t*, uint8_t*, int, uint32_t);



/* PIX_CLIP FUNCTIONS */

#ifdef DEBUG
static int
pix_clip_check(int v)
{
  if (v < -348 || v >= 640) {
    fprintf(stderr, "pix_clip_check: value %d\n", v);
    v = (v<0) ? 0 : (v>255) ? 255 : v;
  }
  return pix_clip[v];
}
#endif

static int
pix_clip_init(void)
{
  int i;
  if (!pix_clip) {
//    pix_clip_buf = (unsigned char*)memalign(16, 1024);
    pix_clip_buf = (unsigned char*)malloc(1024);
    if (!pix_clip_buf) {
      fprintf(stderr, "pix_clip_init: memory error.\n");
      return -1;
    }
    pix_clip =  pix_clip_buf + 384;
    for (i = -384; i < 640; i++)
      pix_clip[i] = (i<0) ? 0 : (i>255) ? 255 : i;
  }
  return 0;
}

static void
pix_clip_quit(void)
{
  free(pix_clip_buf);
  pix_clip_buf = pix_clip = NULL;
}


/* VFRAME FUNCTIONS */

static vframe_t*
vframe_free(vframe_t *vfrm)
{
  if (vfrm->bufp) {
    free(vfrm->bufp);
//    vfrm->plane[0] = vfrm->plane[1] = vfrm->plane[2] = NULL;
  } else {
    free(vfrm->plane[0]);
    free(vfrm->plane[1]);
    free(vfrm->plane[2]);
  }
  free(vfrm);
  return NULL;
}

static vframe_t*
vframe_alloc(int width, int height, int horz_border, int vert_border, int csp)
{
  int bufsize, offset;
  vframe_t *vfrm;

  vfrm = (vframe_t*)malloc(sizeof(vframe_t));
  if (!vfrm) {
    fprintf(stderr, "vframe_alloc: memory error.\n");
    return NULL;
  }
  vfrm->width[0] = width;
  vfrm->height[0] = height;
  vfrm->frame_width[0] = width + horz_border*2;
  vfrm->frame_height[0] = height + vert_border*2;
  vfrm->horz_border[0] = horz_border;
  vfrm->vert_border[0] = vert_border;

  switch (csp) {
    default:
      vfrm->width[1] = width/2;
      vfrm->width[2] = width/2;
      vfrm->height[1] = height/2;
      vfrm->height[2] = height/2;
      vfrm->frame_width[1] = width/2 + horz_border*2/2;
      vfrm->frame_width[2] = width/2 + horz_border*2/2;
      vfrm->frame_height[1] = height/2 + vert_border*2/2;
      vfrm->frame_height[2] = height/2 + vert_border*2/2;
      vfrm->horz_border[1] = horz_border/2;
      vfrm->horz_border[2] = horz_border/2;
      vfrm->vert_border[1] = vert_border/2;
      vfrm->vert_border[2] = vert_border/2;
      break;
  }
  vfrm->csp = csp;

  vfrm->stride[0] = vfrm->frame_width[0];
  vfrm->stride[1] = vfrm->frame_width[1];
  vfrm->stride[2] = vfrm->frame_width[2];

  bufsize =  vfrm->frame_width[0] * vfrm->frame_height[0]
           + vfrm->frame_width[1] * vfrm->frame_height[1]
           + vfrm->frame_width[2] * vfrm->frame_height[2];
  vfrm->bufp = (unsigned char*)memalign(16, bufsize);
  if (!vfrm->bufp) {
    free(vfrm);
    fprintf(stderr, "vframe_alloc: memory error.\n");
    return NULL;
  }

  vfrm->plane[0] = vfrm->bufp;
  offset = vfrm->frame_width[0] * vfrm->frame_height[0];
  vfrm->plane[1] = vfrm->bufp + offset;
  offset += vfrm->frame_width[1] * vfrm->frame_height[1];
  vfrm->plane[2] = vfrm->bufp + offset;

  vfrm->offset[0] =  vfrm->plane[0]
                   + (vfrm->frame_width[0] * vfrm->vert_border[0])
		   + vfrm->horz_border[0];
  vfrm->offset[1] =  vfrm->plane[1]
                   + (vfrm->frame_width[1] * vfrm->vert_border[1])
		   + vfrm->horz_border[1];
  vfrm->offset[2] =  vfrm->plane[2]
                   + (vfrm->frame_width[2] * vfrm->vert_border[2])
		   + vfrm->horz_border[2];

  bufsize = vfrm->frame_width[0] * vfrm->frame_height[0];
  memset(vfrm->plane[0], 0, bufsize);
  switch (csp) {
    default:
      bufsize = vfrm->frame_width[1] * vfrm->frame_height[1];
      memset(vfrm->plane[1], 128, bufsize);
      bufsize = vfrm->frame_width[2] * vfrm->frame_height[2];
      memset(vfrm->plane[2], 128, bufsize);
      break;
  }

  return vfrm;
}

/* LOWPASS FILTER FUNCTIONS */

static double filt_cubic(double x)         /* 4th order (cubic) b-spline */
{
    double t;

    if (x<-2.) return 0.;
    if (x<-1.) {t = 2.+x; return t*t*t/6.;}
    if (x<0.) return (4.+x*x*(-6.+x*-3.))/6.;
    if (x<1.) return (4.+x*x*(-6.+x*3.))/6.;
    if (x<2.) {t = 2.-x; return t*t*t/6.;}
    return 0.;
}

static double filt_catrom(double x)        /* Catmull-Rom spline, Overhauser spline */
{
    if (x<-2.) return 0.;
    if (x<-1.) return .5*(4.+x*(8.+x*(5.+x)));
    if (x<0.) return .5*(2.+x*x*(-5.+x*-3.));
    if (x<1.) return .5*(2.+x*x*(-5.+x*3.));
    if (x<2.) return .5*(4.+x*(-8.+x*(5.-x)));
    return 0.;
}

static int
make_filt_coef(double blur, FILT_FUNC func, int coefs[], double supp)
{
  double filt_scale;
  double den, sc, tr;
  int i, c, ti, sum;

  if (func == NULL)
    func = filt_catrom;
  filt_scale = blur;
  if (supp == 0.)
    supp = 2.0;

  den = 0;
  for (i = -2; i <= 2; i++)
    den += func(((double)i) / filt_scale);
  sc = den==0. ? 1 : 1/den;
  sum = 0;
  c = 0;
  for (i = -2; i <= 2; i++) {
    tr = func(((double)i) / filt_scale);
    tr = sc * tr * 256;
    ti = tr + 0.5;
    sum = sum + ti;
    coefs[c++] = ti;
  }

  if (sum != 256)
    coefs[2] += 256 - sum;

  return 0;
}

static void
lowpass_filt3(int width, int height, unsigned char *src, int src_stride, unsigned char *dst, int dst_stride, int coef[])
{
  int i, j;
  unsigned char *sp, *dp;
  unsigned char *bpm1, *bp, *bpp1;
  unsigned char buf[width*height];
  int v;

  sp = src;
  bp = buf;
  for (j = 0; j < height; j++) {
    i = 0;
    v = (  coef[1]*sp[i]   + coef[2]*sp[i] + coef[3]*sp[i+1] + 128) >> 8;
    bp[i++] = PIX_CLIP(v);
    for (; i < width-1; i++) {
      v     = (  coef[1]*sp[i-1]
               + coef[2]*sp[i]
               + coef[3]*sp[i+1] + 128) >> 8;
      bp[i] = PIX_CLIP(v);
    }
    v = (  coef[1]*sp[i-1] + coef[2]*sp[i] + coef[3]*sp[i] + 128) >> 8;
    bp[i] = PIX_CLIP(v);
    sp += src_stride;
    bp += width;
  }

  dp = dst;
  bpm1 = bp = buf;
  bpp1 = buf + width;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      v     = (  coef[1]*bpm1[i]
               + coef[2]*bp[i]
               + coef[3]*bpp1[i] + 128) >> 8;
      dp[i] = PIX_CLIP(v);
    }
    dp += dst_stride;
    bpm1 = bp; bp = bpp1;
    if (j < height-1)
      bpp1 += width;
  }
}

static void
lowpass_filt5(int width, int height, unsigned char *src, int src_stride, unsigned char *dst, int dst_stride, int coef[])
{
  int i, j;
  unsigned char *sp, *dp;
  unsigned char *bpm2, *bpm1, *bp, *bpp1, *bpp2;
  unsigned char buf[width*height];
  int v;

  sp = src;
  bp = buf;
  for (j = 0; j < height; j++) {
    i = 0;
    v = (  coef[0]*sp[i]   + coef[1]*sp[i]   + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+2] + 128) >> 8;
    bp[i++] = PIX_CLIP(v);
    v = (  coef[0]*sp[i-1] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+2] + 128) >> 8;
    bp[i++] = PIX_CLIP(v);
    for (; i < width-2; i++) {
      v     = (  coef[0]*sp[i-2]
               + coef[1]*sp[i-1]
               + coef[2]*sp[i]
               + coef[3]*sp[i+1]
               + coef[4]*sp[i+2] + 128) >> 8;
      bp[i] = PIX_CLIP(v);
    }
    v = (  coef[0]*sp[i-2] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+1] + 128) >> 8;
    bp[i++] = PIX_CLIP(v);
    v = (  coef[0]*sp[i-2] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i] + coef[4]*sp[i] + 128) >> 8;
    bp[i] = PIX_CLIP(v);
    sp += src_stride;
    bp += width;
  }

  dp = dst;
  bpm2 = bpm1 = bp = buf;
  bpp1 = buf + width;
  bpp2 = bpp1 + width;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      v     = (  coef[0]*bpm2[i]
	       + coef[1]*bpm1[i]
	       + coef[2]*bp[i]
	       + coef[3]*bpp1[i]
	       + coef[4]*bpp2[i] + 128) >> 8;
      dp[i] = PIX_CLIP(v);
    }
    dp += dst_stride;
    bpm2 = bpm1; bpm1 = bp; bp = bpp1; bpp1 = bpp2;
    if (j < height-2)
      bpp2 += width;
  }
}

/* MEDIAN FILTER FUNCTIONS */

/* num must be odd */
static inline int
med_sort(int data[], int num, int shift)
{
  int i;
  int big_num, small_num;
  int big_count, small_count;
  int *big, *small, *src, *tmp;
  int arr[3][num];
  int val;

  big_num = small_num = num >> 1;
  if (!(num & 1)) {
    big_num -= 1;
  }
  big_num -= shift;
  small_num += shift;
  if (big_num <= 0 || small_num <= 0) {
    fprintf(stderr, "med_sort: invalid number, big %d, small %d.\n", big_num, small_num);
    exit(1);
  }

  src = arr[0];
  big = arr[1];
  small = arr[2];

  val = data[0];
  big_count = small_count = 0;
  for (i = 1; i < num; i++) {
    if (data[i] > val) {
      big[big_count] = data[i];
      big_count++;
//    } else if (data[i] <= val) {
    } else {
      small[small_count] = data[i];
      small_count++;
    }
  }

  if (big_count == big_num || small_count == small_num) {
    return val;
  }
  else if (big_count < big_num) {
    num = num - big_count - 1;
    big_num = num - small_num - 1;
    tmp = src; src = small; small = tmp;
  }
//  else if (small_count < small_num) {
  else {
    num = num - small_count - 1;
    small_num = num - big_num - 1;
    tmp = src; src = big; big = tmp;
  }

  do {
    val = src[0];
    big_count = small_count = 0;
    for (i = 1; i < num; i++) {
      if (src[i] > val) {
        big[big_count] = src[i];
        big_count++;
//    } else if (src[i] <= val) {
      } else {
        small[small_count] = src[i];
        small_count++;
      }
    }

    if (big_count == big_num || small_count == small_num) {
      return val;
    }
    else if (big_count < big_num) {
      num = num - big_count - 1;
      big_num = num - small_num - 1;
      tmp = src; src = small; small = tmp;
    }
//  else if (small_count < small_num) {
    else {
      num = num - small_count - 1;
      small_num = num - big_num - 1;
      tmp = src; src = big; big = tmp;
    }
  } while (1);

  // never reached
  return val;
}

static int
median_field(unsigned char *frm, int stride)
{
  int point[13];
  int v;
  int thre = Threshold2;

  point[0]  = *frm;

  point[7]  = point[0] - thre;
  point[8]  = point[0] + thre;

  point[1]  = frm[-1];
  point[2]  = frm[1];

  point[3]  = frm[-stride];
  point[4]  = frm[stride];
  point[5]  = frm[-stride*2];
  point[6]  = frm[stride*2];

  v = med_sort(point, 9, 0);
#if 0
  point[9]  = frm[-1] - thre;
  point[10] = frm[-1] + thre;
  point[11] = frm[1] - thre;
  point[12] = frm[1] + thre;
  v = med_sort(point, 13, 0);
#endif
  return v;
}

static int
median_frame(unsigned char *frm, int stride)
{
  int point[11];
  int v;
  int thre = Threshold2;

#if 0
  point[0]  = frm[-1];
  point[1]  = *frm;
  point[2]  = frm[1];

  point[3]  = frm[-stride-1];
  point[4]  = frm[-stride];
  point[5]  = frm[-stride+1];

  point[6]  = frm[stride-1];
  point[7]  = frm[stride];
  point[8]  = frm[stride+1];

  point[9]  = point[1] - thre;
  point[10] = point[1] + thre;
  v = med_sort(point, 11, 0);
#endif

  point[0]  = *frm;
  point[1]  = frm[-1];
  point[2]  = frm[1];

  point[3]  = frm[-stride];
  point[4]  = frm[stride];

  point[5]  = point[0] - thre;
  point[6]  = point[0] + thre;

  point[7]  = frm[-stride*2];
  point[8]  = frm[stride*2];

  v = med_sort(point, 9, 0);
  return v;
}

/* LAPLACIAN FUNCTIONS */

static void
laplacian(int width, int height, unsigned char *src, int src_stride, unsigned char *dst, int dst_stride)
{
  unsigned char *sp, *spm1, *spp1, *dp;
  int v, i, j;

  spm1 = sp = src;
  spp1 = sp + src_stride;
  dp = dst;
  for (j = 0; j < height; j++)
  {
    i = 0;
    v = (12*sp[i] - sp[i] - sp[i+1] - spm1[i] - spp1[i])>>3;
    //v = (5*sp[i] - sp[i] - sp[i+1] - spm1[i] - spp1[i]);
    dp[i++] = PIX_CLIP(v);
    for (i = 0; i < width-1; i++)
    {
      v = (12*sp[i] - sp[i-1] - sp[i+1] - spm1[i] - spp1[i])>>3;
      //v = (5*sp[i] - sp[i-1] - sp[i+1] - spm1[i] - spp1[i]);
      dp[i] = PIX_CLIP(v);
    }
    v = (12*sp[i] - sp[i-1] - sp[i] - spm1[i] - spp1[i])>>3;
    //v = (5*sp[i] - sp[i-1] - sp[i] - spm1[i] - spp1[i])>>3;
    dp[i++] = PIX_CLIP(v);
    dp += dst_stride;
    spm1 = sp; sp = spp1;
    if (j < height-1)
      spp1 += src_stride;
  }
}

/* MOTION SEARCH FUNCTIONS */


static void 
Mb_search_44 (uint8_t *cur, int cur_stride, uint8_t *ref, int ref_stride, motion_data_t *mi)
{
  uint32_t best_SAD=0x00ffffff;
  uint32_t SAD=0x00ffffff; 
  int radius = Motion_radius>>2;       /* search radius /4 in pixels */
  int32_t  x, y;
  uint8_t *rp;

  rp = ref - ((radius+3) * ref_stride) - 3;
  for(y = -radius; y <= radius; y++)
  {
    for(x = -radius; x <= radius; x++)
    {
      SAD = Calc_SAD (cur, cur_stride, &rp[x], ref_stride, best_SAD);
      SAD += x*x + y*y; /* favour center matches... */
      if(SAD < best_SAD) {
        best_SAD = SAD;
	mi->mvx = x;
	mi->mvy = y;
      }
    }
    rp += ref_stride;
  }
  mi->sad = best_SAD;
}

static void 
Mb_search_22 (uint8_t *cur, int cur_stride, uint8_t *ref, int ref_stride, motion_data_t *mi)
{
  uint32_t best_SAD=0x00ffffff;
  uint32_t SAD=0x00ffffff; 
  int32_t  x, y;
  uint8_t *rp;
  int vx;
  int vy;

  vx = mi->mvx<<1;
  vy = mi->mvy<<1;

  rp = ref + ((vy-2-2) * ref_stride + vx - 2);
  for(y = -2; y <= 2; y++)
  {
    for(x = -2; x <= 2; x++)
    {
      SAD = Calc_SAD (cur, cur_stride, &rp[x], ref_stride, best_SAD);
      if(SAD < best_SAD) {
        best_SAD = SAD;
        mi->mvx = x + vx;
        mi->mvy = y + vy;
      }
    }
    rp += ref_stride;
  }

  SAD = Calc_SAD (cur, cur_stride, ref, ref_stride, best_SAD);
  if(SAD <= best_SAD) {
    best_SAD = SAD;
    mi->mvx = 0;
    mi->mvy = 0;
  }
  mi->sad = best_SAD;
}

static void 
Mb_search_11 (uint8_t *cur, int cur_stride, uint8_t *ref, int ref_stride, motion_data_t *mi)
{
  uint32_t best_SAD=0x00ffffff;
  uint32_t SAD=0x00ffffff; 
  int32_t  x, y;
  uint8_t *rp;
  int vx;
  int vy;

  vx = mi->mvx<<1;
  vy = mi->mvy<<1;

  rp = ref + ((vy-2) * ref_stride + vx);
  for(y = -2; y <= 2; y++)
  {
    for(x = -2; x <= 2; x++)
    {
      SAD = Calc_SAD (cur, cur_stride, &rp[x], ref_stride, best_SAD);
      if(SAD < best_SAD) {
        best_SAD = SAD;
        mi->mvx = x + vx;
        mi->mvy = y + vy;
      }
    }
    rp += ref_stride;
  }

  SAD = Calc_SAD (cur, cur_stride, ref, ref_stride, best_SAD);
  if(SAD <= best_SAD) {
    best_SAD = SAD;
    mi->mvx = 0;
    mi->mvy = 0;
  }
  mi->sad = best_SAD;
}

#if 0
static void 
Mb_search_00 (uint8_t *cur, int cur_stride, uint8_t *ref, int ref_stride, motion_data_t *mi)
{
  uint32_t best_SAD = 0x00ffffff;
  uint32_t SAD, SAD1;
  uint8_t *rp;
  int x, y;
  int vx, vy;

  vx = mi->mvx;
  vy = mi->mvy;

  rp = ref + ((vy-1) * ref_stride + vx);
  vx <<= 1;
  vy <<= 1;
  best_SAD = mi->sad;
  for(y = -2; y <= 2; y++)
  {
    if (y & 1) {
      for(x = -2; x <= 2; x++)
      {
	if (x &1) {
          SAD = Calc_SAD_half4 (cur, cur_stride,
                                &rp[x>>1], &rp[(x>>1) + ref_stride],
                                &rp[(x>>1)+1], &rp[(x>>1) + ref_stride + 1],
			        ref_stride, best_SAD);
	} else {
          SAD = Calc_SAD_half2 (cur, cur_stride,
                                &rp[x>>1], &rp[(x>>1) + ref_stride],
			        ref_stride, best_SAD);
	}
        if(SAD<best_SAD)
        {
          best_SAD = SAD;
          mi->mvx = x + vx;
          mi->mvy = y + vy;
        }
      }
      rp += ref_stride;
    } else {
      SAD = Calc_SAD_half2 (cur, cur_stride,
                            &rp[-1], &rp[0],
                            ref_stride, best_SAD);
      SAD1 = Calc_SAD_half2 (cur, cur_stride,
                            &rp[0], &rp[1],
                            ref_stride, best_SAD);
      if (SAD < SAD1) {
	x = -1;
      } else {
	x = 1;
	SAD = SAD1;
      }
      if(SAD < best_SAD) {
        best_SAD = SAD;
        mi->mvx = x + vx;
        mi->mvy = y + vy;
      }
    }
    rp += ref_stride;
  }
  mi->sad = best_SAD;
}
#endif

static void 
Mb_search_00 (uint8_t *cur, int cur_stride, uint8_t *ref, int ref_stride, motion_data_t *mi)
{
  uint32_t best_SAD = 0x00ffffff;
  uint32_t SAD, SAD1;
  uint8_t *rp;
  int x, y;
  int vx, vy;

  vx = mi->mvx;
  vy = mi->mvy;

  rp = ref + ((vy-1) * ref_stride + vx);
  vx <<= 1;
  vy <<= 1;
  mi->mvx = vx;
  mi->mvy = vy;
  best_SAD = mi->sad;
  for(y = -2; y <= 2; y+=2)
  {
    SAD = Calc_SAD_half2 (cur, cur_stride,
                          &rp[-1], &rp[0],
                          ref_stride, best_SAD);
    SAD1 = Calc_SAD_half2 (cur, cur_stride,
                           &rp[0], &rp[1],
                           ref_stride, best_SAD);
    if (SAD < SAD1) {
      x = -1;
    } else {
      x = 1;
      SAD = SAD1;
    }
    if(SAD < best_SAD) {
      best_SAD = SAD;
      mi->mvx = x + vx;
      mi->mvy = y + vy;
    }
    rp += ref_stride;
  }
  mi->sad = best_SAD;
}

#ifdef ARCH_X86

static uint32_t
Calc_SAD_mmxext(uint8_t *src0, int stride0, uint8_t *src1, int stride1, uint32_t best_SAD)
{
  register uint32_t a;
  int y;
  
  pxor_r2r(mm0, mm0);
  for (y = 0; y < 8; y++) {
    movq_m2r(*src0, mm1);
    psadbw_m2r(*src1, mm1);
    paddd_r2r(mm1, mm0);
    movd_r2v(mm0, a);
    if (a > best_SAD)
      return a;
    src0 += stride0;
    src1 += stride1;
  }
  movd_r2v(mm0, a);

  return a;
}

static uint32_t
Calc_SAD_half2_mmxext(uint8_t *cur, int cur_stride, uint8_t *ref1, uint8_t *ref2, int ref_stride, uint32_t best_SAD) 
{
  register uint32_t a;
  int y;

  pxor_r2r(mm0, mm0);
  for (y = 0; y < 8; y++) {
    movq_m2r(*ref1, mm1);
    pavgb_m2r(*ref2, mm1);
    psadbw_m2r(*cur, mm1);
    paddd_r2r(mm1, mm0);
    movd_r2v(mm0, a);
    if (a > best_SAD)
      return a;
    ref1 += ref_stride;
    ref2 += ref_stride;
    cur += cur_stride;
  }
  movd_r2v(mm0, a);
  return a;
}

static uint32_t
Calc_SAD_half4_mmxext(uint8_t *cur, int cur_stride, uint8_t *ref1, uint8_t *ref2, uint8_t *ref3, uint8_t *ref4, int ref_stride, uint32_t best_SAD) 
{
  register uint32_t a;
  int y;

  pxor_r2r(mm0, mm0);
  for (y = 0; y < 8; y++) {
    movq_m2r(*ref1, mm1);
    pavgb_m2r(*ref2, mm1);
    pavgb_m2r(*ref3, mm1);
    pavgb_m2r(*ref4, mm1);
    psadbw_m2r(*cur, mm1);
    paddd_r2r(mm1, mm0);
    movd_r2v(mm0, a);
    if (a > best_SAD)
      return a;
    ref1 += ref_stride;
    ref2 += ref_stride;
    ref3 += ref_stride;
    ref4 += ref_stride;
    cur += cur_stride;
  }
  movd_r2v(mm0, a);
  return a;
}

#endif /* ARCH_X86 */

static uint32_t
Calc_SAD_c(uint8_t *src0, int stride0, uint8_t *src1, int stride1, uint32_t best_SAD)
{
  int dx, dy;
  int d = 0;
  uint32_t a = 0;

  for (dy = 0; dy < 8; dy++) {
    for (dx = 0; dx < 8; dx++) {
      d = src0[dx] - src1[dx];
      a += (d<0) ? -d:d;
      if (a > best_SAD)
	return a;
    }
    src0 += stride0;
    src1 += stride1;
  }
  return a;
}

static uint32_t
Calc_SAD_half2_c(uint8_t *cur, int cur_stride, uint8_t *ref1, uint8_t *ref2, int ref_stride, uint32_t best_SAD) 
{
  int dx, dy;
  int d = 0;
  uint32_t a = 0;

  for (dy = 0; dy < 8; dy++) {
    for (dx = 0; dx < 8; dx++) {
      d = ((ref1[dx] + ref2[dx] + 1) >> 1) - cur[dx];
      a += (d<0) ? -d:d;
      if (a > best_SAD)
	return a;
    }
    ref1 += ref_stride;
    ref2 += ref_stride;
    cur  += cur_stride;
  }
  return a;
}

static uint32_t
Calc_SAD_half4_c(uint8_t *cur, int cur_stride, uint8_t *ref1, uint8_t *ref2, uint8_t *ref3, uint8_t *ref4, int ref_stride, uint32_t best_SAD) 
{
  int dx, dy;
  int d = 0;
  uint32_t a = 0;

  for (dy = 0; dy < 8; dy++) {
    for (dx = 0; dx < 8; dx++) {
      d = ((ref1[dx] + ref2[dx] + ref3[dx] + ref4[dx] + 2) >> 2) - cur[dx];
      a += (d<0) ? -d:d;
      if (a > best_SAD)
	return a;
    }
    ref1 += ref_stride;
    ref2 += ref_stride;
    ref3 += ref_stride;
    ref4 += ref_stride;
    cur  += cur_stride;
  }
  return a;
}

static void
subsample(uint8_t *src, int src_stride, uint8_t *dst, int dst_stride, int width, int height)
{
  int x, y;
  int dpos;
  uint8_t *s1, *s2;

  s1 = src;
  s2 = src + src_stride;
  src_stride <<= 1;
  height >>= 1;
  for (y = 0; y < height; y++) {
    for (x = dpos = 0; x < width; x += 2, dpos++) {
      dst[dpos] = (s1[x] + s1[x+1] + s2[x] + s2[x+1] + 2) >> 2;
    }
    s1 += src_stride;
    s2 += src_stride;
    dst += dst_stride;
  }
}

static void
motion_search(int width, int height, vframe_t *curfrm, vframe_t *frmm1, vframe_t *frmp1, block_info_t *blkinfo)
{
  int x, y;
  int width2 = width>>1;
  int height2 = height>>1;
  int width4 = width2>>1;
//  int stride4 = width4*(8/4);
//  int stride2 = width2*(8/2);
//  int stride1 = width*8;
//  int height4 = height2>>1;
  unsigned char *cur1, *cur2, *cur4;
  unsigned char *fm11, *fm12, *fm14;
  unsigned char *fp11, *fp12, *fp14;
  unsigned char *tmp;

  cur1 = curfrm->offset[0];
  cur2 = sub2[1];
  cur4 = sub4[1];
  fm11 = frmm1->offset[0];
  fm12 = sub2[2];
  fm14 = sub4[2];
  fp11 = frmp1->offset[0];
  fp12 = sub2[0];
  fp14 = sub4[0];

  subsample(fp11, frmp1->stride[0], fp12, width2, width, height);
  subsample(fp12, width2, fp14, width4, width2, height2);

  for (y = 0; y < height; y += 8) {
    for (x = 0; x < width; x += 8) {
      if (blkinfo->flag & BLOCK_FLAG_MOTION) {
        Mb_search_44(cur4+(x>>2), width4, fm14+(x>>2), width4, blkinfo->mibk);
        Mb_search_22(cur2+(x>>1), width2, fm12+(x>>1), width2, blkinfo->mibk);
        Mb_search_11(cur1+ x,     width,  fm11+ x,     width,  blkinfo->mibk);
        Mb_search_00(cur1+ x,     width,  fm11+ x,     width,  blkinfo->mibk);

        Mb_search_44(cur4+(x>>2), width4, fp14+(x>>2), width4, blkinfo->mifw);
        Mb_search_22(cur2+(x>>1), width2, fp12+(x>>1), width2, blkinfo->mifw);
        Mb_search_11(cur1+ x,     width,  fp11+ x,     width,  blkinfo->mifw);
        Mb_search_00(cur1+ x,     width,  fp11+ x,     width,  blkinfo->mifw);
	emms();
      } else {
	blkinfo->mibk->mvx = blkinfo->mibk->mvy = 0;
	blkinfo->mifw->mvx = blkinfo->mifw->mvy = 0;
	blkinfo->errcnt = 0;
      }
      blkinfo++;
    }
    cur4 += width4*(8/4);
    cur2 += width2*(8/2);
    cur1 += width *8;
    fm14 += width4*(8/4);
    fm12 += width2*(8/2);
    fm11 += width *8;
    fp14 += width4*(8/4);
    fp12 += width2*(8/2);
    fp11 += width *8;
  }

  tmp = sub2[2]; sub2[2] = sub2[1]; sub2[1] = sub2[0]; sub2[0] = tmp;
  tmp = sub4[2]; sub4[2] = sub4[1]; sub4[1] = sub4[0]; sub4[0] = tmp;
}

static int
motion_search_init(int width, int height)
{
  int h1, h2, h4;
  int w1, w2, w4;
  int vborder;
  int bufsize;
  int offset;

  vborder = ((Motion_radius + 16 + 15) / 16) * 16;
  h1 = height + vborder*2;
  h2 = (h1 >> 1) + 4*2;
  h4 = (h2 >> 1) + 3*2;
  w1 = width;
  w2 = w1 >> 1;
  w4 = w2 >> 1;

  bufsize = h2 * w2;
  bufp2 = (unsigned char*)memalign(16, bufsize*3 + 16);
  offset = w2 * ((vborder>>1)+4);
  sub2[0] = bufp2 + (((offset+15)/16)*16);
  sub2[1] = sub2[0] + bufsize;
  sub2[2] = sub2[1] + bufsize;
  sub2[0] = (unsigned char*)((int)(sub2[0]) & ~0xf);
  sub2[1] = (unsigned char*)((int)(sub2[1]) & ~0xf);
  sub2[2] = (unsigned char*)((int)(sub2[2]) & ~0xf);

  memset(bufp2, 0, bufsize*3 + 16);

  bufsize = h4 * w4;
  bufp4 = (unsigned char*)memalign(16, bufsize*3 + 16);
  offset = w4 * ((vborder>>2)+3);
  sub4[0] = bufp4 + (((offset+15)/16)*16);
  sub4[1] = sub4[0] + bufsize;
  sub4[2] = sub4[1] + bufsize;
  sub4[0] = (unsigned char*)((int)(sub4[0]) & ~0xf);
  sub4[1] = (unsigned char*)((int)(sub4[1]) & ~0xf);
  sub4[2] = (unsigned char*)((int)(sub4[2]) & ~0xf);

  memset(bufp4, 0, bufsize*3 + 16);

  Calc_SAD = Calc_SAD_mmxext;
  Calc_SAD_half2 = Calc_SAD_half2_mmxext;
  Calc_SAD_half4 = Calc_SAD_half4_mmxext;

  return 0;
}

static void
motion_search_quit(void)
{
  free(bufp2);
  free(bufp4);
}

static void
make_motion_frame(int width, int height, unsigned char *refm1, unsigned char *refp1, int ref_stride, unsigned char *dstm1, unsigned char *dstp1, int dst_stride, block_info_t *blkinfo)
{
  int w, h;
//  int i;
  int j;
  int vx, vy;
  int halfx, halfy, dir;
  unsigned char *dp, *rp0, *rp1, *rp2, *rp3;

  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      if (blkinfo->flag & BLOCK_FLAG_MOTION)
      {
	for (dir = 0; dir < 2; dir++)
       	{
	  if (dir == 0) {
	    vx = blkinfo->mibk->mvx;
	    vy = blkinfo->mibk->mvy;
	    halfx = (vx & 1);
	    halfy = (vy & 1);
	    vx >>= 1;
	    vy >>= 1;
            rp0 = refm1 + w + (vy * ref_stride + vx);
	    dp = dstm1 + w;
	  } else {
	    vx = blkinfo->mifw->mvx;
	    vy = blkinfo->mifw->mvy;
	    halfx = (vx & 1);
	    halfy = (vy & 1);
	    vx >>= 1;
	    vy >>= 1;
            rp0 = refp1 + w + (vy * ref_stride + vx);
	    dp = dstp1 + w;
	  }

	  if (!halfx && !halfy) {
            for (j = 0; j < 8; j++) {
#ifdef ARCH_X86
              movq_m2r(*rp0, mm0);
              movq_r2m(mm0, *dp);
#else
              for (i = 0; i < 8; i++) {
                dp[i] = rp0[i];
              }
#endif
              dp += dst_stride;
              rp0 += ref_stride;
            }
	  }
	  else if (halfx && !halfy) {
	    rp1 = rp0 + 1;
            for (j = 0; j < 8; j++) {
#ifdef ARCH_X86
              movq_m2r(*rp0, mm0);
              pavgb_m2r(*rp1, mm0);
              movq_r2m(mm0, *dp);
#else
              for (i = 0; i < 8; i++) {
                dp[i] = (rp0[i] + rp1[i] + 1) >> 1;
              }
#endif
              dp += dst_stride;
              rp0 += ref_stride;
              rp1 += ref_stride;
            }
	  }
	  else if (!halfx && halfy) {
	    rp2 = rp0 + ref_stride;
            for (j = 0; j < 8; j++) {
#ifdef ARCH_X86
              movq_m2r(*rp0, mm0);
              pavgb_m2r(*rp2, mm0);
              movq_r2m(mm0, *dp);
#else
              for (i = 0; i < 8; i++) {
                dp[i] = (rp0[i] + rp2[i] + 1) >> 1;
              }
#endif
              dp += dst_stride;
              rp0 += ref_stride;
              rp2 += ref_stride;
            }
	  }
	  else if (halfx && halfy) {
	    rp1 = rp0 + 1;
	    rp2 = rp0 + ref_stride;
	    rp3 = rp2 + 1;
            for (j = 0; j < 8; j++) {
#ifdef ARCH_X86
              movq_m2r(*rp0, mm0);
              pavgb_m2r(*rp1, mm0);
              pavgb_m2r(*rp2, mm0);
              pavgb_m2r(*rp3, mm0);
              movq_r2m(mm0, *dp);
#else
              for (i = 0; i < 8; i++) {
                dp[i] = (rp0[i] + rp1[i] + rp2[i] + rp3[i] + 2) >> 2;
              }
#endif
              dp += dst_stride;
              rp0 += ref_stride;
              rp1 += ref_stride;
              rp2 += ref_stride;
              rp3 += ref_stride;
            }
	  }
	}
      }
#if 0
      else {
	for (dir = 0; dir < 2; dir++)
       	{
	  if (dir == 0) {
            dp = dstm1 + w;
	    rp0 = refm1 + w;
	  } else {
            dp = dstp1 + w;
	    rp0 = refp1 + w;
	  }
          for (j = 0; j < 8; j++) {
            for (i = 0; i < 8; i++) {
              dp[i] = rp0[i];
            }
            dp += dst_stride;
            rp0 += ref_stride;
	  }
	}
      }
#endif
      blkinfo++;
    }
    dstm1 += dst_stride*8;
    dstp1 += dst_stride*8;
    refm1 += ref_stride*8;
    refp1 += ref_stride*8;
  }
  emms();
}


/* DENOISE FUNCTIONS */

//#define FW_AVG 1

//#define BIAVG(c,b,f) (((c)*6 + (b)*7 + (f)*3)>>4)
//#define BIAVG(c,b,f) (((c)*1 + (b)*-1 + (f)*1))            /* 1 */
//#define BIAVG(c,b,f) (((c)*1 + (b)*0 + (f)*1)>>1)          /* 2 */
//#define BIAVG(c,b,f) (((c)*451 + (b)*122 + (f)*451)>>10)   /* 3 */
//#define BIAVG(c,b,f) (((c)*424 + (b)*176 + (f)*424)>>10)   /* 4 */
//#define BIAVG(c,b,f) (((c)*408 + (b)*208 + (f)*408)>>10)   /* 5 */
//#define BIAVG(c,b,f) (((c)*397 + (b)*230 + (f)*397)>>10)   /* 6 */
//#define BIAVG(c,b,f) (((c)*390 + (b)*244 + (f)*390)>>10)   /* 7 */
//#define BIAVG(c,b,f) (((c)*384 + (b)*256 + (f)*384)>>10)   /* 8 */
//#define BIAVG(c,b,f) (((c)*379 + (b)*266 + (f)*379)>>10)   /* 9 */
//#define BIAVG(c,b,f) (((c)*376 + (b)*272 + (f)*376)>>10)   /* 10 */
//#define BIAVG(c,b,f) (((c)*363 + (b)*298 + (f)*363)>>10)   /* 16 */
//#define BIAVG(c,b,f) (((c)*352 + (b)*320 + (f)*352)>>10)   /* 32 */
//#define BIAVG(c,b,f) (((c)*347 + (b)*330 + (f)*347)>>10)   /* 64 */
//#define BIAVG(c,b,f) (((c)*344 + (b)*336 + (f)*344)>>10)   /* 128 */
//#define BIAVG(c,b,f) (((c)*343 + (b)*338 + (f)*343)>>10)   /* 256 */
#define BIAVG(c,b,f) (((c)*342 + (b)*340 + (f)*342)>>10)   /* 2048 */

//#define BIAVG(c,b,f) (((c)*3 + (b)*12 + (f)*1)>>4)   /* 2048 */
//#define BIAVG(c,b,f) (((c)*3 + (b)*28 + (f)*1)>>5)   /* 2048 */


//#define BKAVG(c,b) (((c)*9 + (b)*7)>>4)
//#define BKAVG(c,b) (((c)*2 + (b)*-1))                     /* 1 */
//#define BKAVG(c,b) (((c)*1 + (b)*0))                      /* 2 */
//#define BKAVG(c,b) (((c) + (b))>>1)            /* 2 */
//#define BKAVG(c,b) (((c)*902 + (b)*122)>>10)   /* 3 */
//#define BKAVG(c,b) (((c)*848 + (b)*176)>>10)   /* 4 */
//#define BKAVG(c,b) (((c)*768 + (b)*256)>>10)   /* 8 */
//#define BKAVG(c,b) (((c)*758 + (b)*266)>>10)   /* 9 */
//#define BKAVG(c,b) (((c)*726 + (b)*298)>>10)   /* 16 */
//#define BKAVG(c,b) (((c)*704 + (b)*320)>>10)   /* 32 */
//#define BKAVG(c,b) (((c)*694 + (b)*330)>>10)   /* 64 */
//#define BKAVG(c,b) (((c)*688 + (b)*336)>>10)   /* 128 */
//#define BKAVG(c,b) (((c)*686 + (b)*338)>>10)   /* 256 */
//#define BKAVG(c,b) (((c)*684 + (b)*340)>>10)   /* 2048 */

//#define BKAVG(c,b) (((c)*(344+172) + (b)*(336+172))>>10)   /* 128 */
#define BKAVG(c,b) (((c)*(342+171) + (b)*(340+171))>>10)   /* 2048 */


//#define FWAVG(c,f) (((c)*13 + (f)*3)>>4)
//#define FWAVG(c,f) (((c) + (f))>>1)            /* 2 */
//#define FWAVG(c,f) (((c)*902 + (f)*122)>>10)   /* 3 */
//#define FWAVG(c,f) (((c)*848 + (f)*176)>>10)   /* 4 */
//#define FWAVG(c,f) (((c)*768 + (f)*256)>>10)   /* 8 */
//#define FWAVG(c,f) (((c)*758 + (f)*266)>>10)   /* 9 */
//#define FWAVG(c,f) (((c)*704 + (f)*320)>>10)   /* 32 */
//#define FWAVG(c,f) (((c)*694 + (f)*330)>>10)   /* 64 */
//#define FWAVG(c,f) (((c)*688 + (f)*336)>>10)   /* 128 */
//#define FWAVG(c,f) (((c)*686 + (f)*338)>>10)   /* 256 */
#define FWAVG(c,f) (((c)*684 + (f)*340)>>10)   /* 2048 */

//#define FWAVG(c,f) (((c)*(344+172) + (f)*(336+172))>>10)   /* 128 */
//#define FWAVG(c,f) (((c)*(342+171) + (f)*(340+171))>>10)   /* 2048 */

static inline int
avg_pixel4(unsigned char *src, int stride, int thre)
{
  int dl, dr, du, dd, l, r, u, d, v;

  v = *src;
  l = src[-1];
  r = src[1];
  u = src[-stride];
  d = src[stride];

  du = v - u;
  du = (du<0) ? -du : du;

  dd = v - d;
  dd = (dd<0) ? -dd : dd;

  if (du <= thre && dd <= thre)
    v = (v*2 + u + d + 2) >> 2;
  else if (du <= thre)
    v = (v*3 + u + 2) >> 2;
  else if (dd <= thre)
    v = (v*3 + d + 2) >> 2;

  dl = v - l;
  dl = (dl<0) ? -dl : dl;

  dr = v - r;
  dr = (dr<0) ? -dr : dr;

  if (dl <= thre && dr <= thre)
    v = (v*2 + l + r + 2) >> 2;
  else if (dl <= thre)
    v = (v*3 + l + 2) >> 2;
  else if (dr <= thre)
    v = (v*3 + r + 2) >> 2;

  return v;
}

static void
deno_frame1(int width, int height, unsigned char *frmm1, unsigned char *frm, unsigned char *frmp1, int frm_stride, unsigned char *pix_flag, block_info_t *blk_info, int uv_flag)
{
  int w, h;
  int i, j;
  int v, errcnt, d0, d1;
  int thre;
  int offset, frm_offset;
  unsigned char *fp, *fpm1, *fpp1;
  unsigned char *pix;
  unsigned int vari_bk, vari_fw;
  unsigned int vari_avgthre, vari_motthre;
  block_info_t *blk;

  if (!uv_flag)
    thre = Threshold;
  else
    thre = Threshold_uv;
  vari_avgthre = (thre*thre)*Avg_threshold;
  vari_motthre = (thre*thre)*Motion_threshold;
  blk = blk_info;
  offset = frm_offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      fp = &frm[frm_offset];
      fpm1 = &frmm1[frm_offset];
      fpp1 = &frmp1[frm_offset];
      pix = &pix_flag[offset];
      vari_bk = vari_fw = 0;
      errcnt = 0;
      for (j = h; j < h+8; j++)
      {
	for (i = w; i < w+8; i++)
	{
          v = fp[i];
	  d0 = v - fpm1[i];
          d0 = (d0<0) ? -d0 : d0;
	  vari_bk += d0*d0;

          d1 = v - fpp1[i];
          d1 = (d1<0) ? -d1 : d1;
	  vari_fw += d1*d1;

	  pix[i] = PIX_FLAG_NONE;

	  if (d0 <= thre) {
	    pix[i] |= PIX_FLAG_BACK_AVG;
	  }
	  if (d1 <= thre) {
	    pix[i] |= PIX_FLAG_FORW_AVG;
	  }

	  if (d0 > thre || d1 > thre) {
	    errcnt++;
	  }
        }
        fp += frm_stride;
        fpm1 += frm_stride;
        fpp1 += frm_stride;
	pix += width;
      }

      blk->flag = BLOCK_FLAG_NONE;
      if (vari_bk <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_BACK_AVG;
      if (vari_fw <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_FORW_AVG;

      if (vari_bk > vari_motthre || vari_fw > vari_motthre)
	blk->flag |= BLOCK_FLAG_MOTION;
      else if (vari_bk <= vari_avgthre && vari_fw <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_AVG;
      blk->vari_bk = vari_bk;
      blk->vari_fw = vari_fw;
      blk->vari_avgthre = vari_avgthre;
      blk->vari_motthre = vari_motthre;
      blk->errcnt = errcnt;
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }

  blk = blk_info;
  offset = frm_offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      fp = &frm[frm_offset];
      fpm1 = &frmm1[frm_offset];
      fpp1 = &frmp1[frm_offset];
      pix = &pix_flag[offset];
      if (blk->flag & BLOCK_FLAG_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
            if (pix[i] == (PIX_FLAG_BACK_AVG|PIX_FLAG_FORW_AVG)) {
	      v = fp[i];
	      v = BIAVG(v, fpm1[i], fpp1[i]);
              fp[i] = PIX_CLIP(v);
	      pix[i] = PIX_FLAG_DONE;
            }
	  }
          fp += frm_stride;
          fpm1 += frm_stride;
          fpp1 += frm_stride;
	  pix += width;
	}
      }
#if 0
      else if (blk->flag & BLOCK_FLAG_BACK_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (pix[i] & PIX_FLAG_BACK_AVG) {
	      v = fp[i];
	      v = BKAVG(v, fpm1[i]);
              fp[i] = PIX_CLIP(v);
	      pix[i] = PIX_FLAG_DONE;
	    }
	  }
          fp += frm_stride;
          fpm1 += frm_stride;
	  pix += width;
	}
	blk->flag &= ~BLOCK_FLAG_MOTION;
      }
#endif
#if 0
      else if (blk->flag & BLOCK_FLAG_FORW_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (pix[i] & PIX_FLAG_FORW_AVG) {
	      v = fp[i];
	      v = FWAVG(v, fpp1[i]);
              fp[i] = PIX_CLIP(v);
	      pix[i] = PIX_FLAG_DONE;
	    }
	  }
          fp += frm_stride;
          fpp1 += frm_stride;
	  pix += width;
	}
	blk->flag &= ~BLOCK_FLAG_MOTION;
      }
#endif
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }
}

static void
deno_frame_motion1(int width, int height, unsigned char *frmm1, unsigned char *frm, unsigned char *frmp1, int frm_stride, unsigned char *mref, unsigned char *mref_forw, unsigned char *pix_flag, block_info_t *blk_info)
{
  int w, h;
  int i, j;
  int v;
  int d0, d1;
  int thre;
  int frm_offset, offset;
  unsigned int vari_mbk, vari_mfw, vari_avgthre;
  unsigned char *mp, *mpf, *fp, *fpm1, *fpp1;
  unsigned char *pix;
  unsigned int blkflag;
  block_info_t *blk;

  thre = Threshold;
  vari_avgthre = (thre*thre)*Avg_threshold;
  blk = blk_info;
  offset = frm_offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      if (blk->flag & BLOCK_FLAG_MOTION)
      {
        fp = &frm[frm_offset];
        mp = &mref[offset];
        mpf = &mref_forw[offset];
        pix = &pix_flag[offset];
        vari_mbk = vari_mfw = 0;
        for (j = h; j < h+8; j++)
        {
	  for (i = w; i < w+8; i++)
	  {
            v = fp[i];
	    d0 = v - mp[i];
            d0 = (d0<0) ? -d0 : d0;
	    vari_mbk += d0*d0;

            d1 = v - mpf[i];
            d1 = (d1<0) ? -d1 : d1;
	    vari_mfw += d1*d1;

	    if (d0 <= thre) {
	      pix[i] |= PIX_FLAG_MOTION_BACK_AVG;
	    }
	    if (d1 <= thre) {
	      pix[i] |= PIX_FLAG_MOTION_FORW_AVG;
	    }

          }
          fp += frm_stride;
          mp += width;
          mpf += width;
	  pix += width;
        }
        if (vari_mbk <= vari_avgthre) {
	  if (blk->flag & BLOCK_FLAG_BACK_AVG) {
	    if (vari_mbk < blk->vari_bk) {
	      blk->flag &= ~BLOCK_FLAG_BACK_AVG;
	      blk->flag |= BLOCK_FLAG_MOTION_BACK_AVG;
	    }
	  } else {
	    blk->flag |= BLOCK_FLAG_MOTION_BACK_AVG;
	  }
	}
        if (vari_mfw <= vari_avgthre) {
	  if (blk->flag & BLOCK_FLAG_FORW_AVG) {
	    if (vari_mfw < blk->vari_fw) {
	      blk->flag &= ~BLOCK_FLAG_FORW_AVG;
	      blk->flag |= BLOCK_FLAG_MOTION_FORW_AVG;
	    }
	  } else {
	    blk->flag |= BLOCK_FLAG_MOTION_FORW_AVG;
	  }
	}
	blk->vari_mbk = vari_mbk;
	blk->vari_mfw = vari_mfw;
      }
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }


  blk = blk_info;
  frm_offset = offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      if (blk->flag & BLOCK_FLAG_MOTION)
      {
        fp = &frm[frm_offset];
        fpm1 = &frmm1[frm_offset];
        fpp1 = &frmp1[frm_offset];
        mp = &mref[offset];
        mpf = &mref_forw[offset];
        pix = &pix_flag[offset];

        blkflag = blk->flag &
              (BLOCK_FLAG_BACK_AVG|BLOCK_FLAG_FORW_AVG
	      |BLOCK_FLAG_MOTION_BACK_AVG|BLOCK_FLAG_MOTION_FORW_AVG);

	if (blkflag == (BLOCK_FLAG_BACK_AVG|BLOCK_FLAG_MOTION_FORW_AVG)) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] &
		    (PIX_FLAG_DONE|PIX_FLAG_BACK_AVG|PIX_FLAG_MOTION_FORW_AVG))
		  == (PIX_FLAG_BACK_AVG|PIX_FLAG_MOTION_FORW_AVG)) {
	        v = fp[i];
		v = BIAVG(v, fpm1[i], mpf[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    fpm1 += frm_stride; mpf += width;
	  }
	}
	else if (blkflag == (BLOCK_FLAG_MOTION_BACK_AVG|BLOCK_FLAG_FORW_AVG)) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] &
		    (PIX_FLAG_DONE|PIX_FLAG_MOTION_BACK_AVG|PIX_FLAG_FORW_AVG))
		  == (PIX_FLAG_MOTION_BACK_AVG|PIX_FLAG_FORW_AVG)) {
	        v = fp[i];
		v = BIAVG(v, mp[i], fpp1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    mp += width; fpp1 += frm_stride;
	  }
	}
	else if (blkflag == ( BLOCK_FLAG_MOTION_BACK_AVG
	                     |BLOCK_FLAG_MOTION_FORW_AVG )) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] &
		    (PIX_FLAG_DONE|PIX_FLAG_MOTION_BACK_AVG|
		     PIX_FLAG_MOTION_FORW_AVG))
		  == (PIX_FLAG_MOTION_BACK_AVG|PIX_FLAG_MOTION_FORW_AVG)) {
	        v = fp[i];
		v = BIAVG(v, mp[i], mpf[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    mp += width; mpf += width;
	  }
	}
	else if (blkflag == BLOCK_FLAG_BACK_AVG) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_BACK_AVG))
		  == (PIX_FLAG_BACK_AVG)) {
	        v = fp[i];
		v = BKAVG(v, fpm1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    fpm1 += width;
	  }
	}
	else if (blkflag == BLOCK_FLAG_MOTION_BACK_AVG) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_MOTION_BACK_AVG))
		  == (PIX_FLAG_MOTION_BACK_AVG)) {
	        v = fp[i];
		v = BKAVG(v, mp[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    mp += width;
	  }
	}
#if 0
#ifdef FW_AVG
	else if (blkflag == BLOCK_FLAG_FORW_AVG) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_FORW_AVG))
		  == (PIX_FLAG_FORW_AVG)) {
	        v = fp[i];
		v = FWAVG(v, fpp1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    fpp1 += width;
	  }
	}
	else if (blkflag == BLOCK_FLAG_MOTION_FORW_AVG) {
	  for (j = h; j < h+8; j++) {
	    for (i = w; i < w+8; i++) {
	      if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_MOTION_FORW_AVG))
		  == (PIX_FLAG_MOTION_FORW_AVG)) {
	        v = fp[i];
		v = FWAVG(v, mpf[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	    }
	    pix += width; fp += frm_stride;
	    mpf += width;
	  }
	}
#endif
#endif
      }
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }
}

static void
deno_median(int width, int height, unsigned char *frm, int frm_stride, unsigned char *pix_flag, int pix_stride, int field_flag)
{
  int w, h;
  int v;
//  int thre = Threshold;
  int (*median_func)(unsigned char*, int);

  if (field_flag)
    median_func = median_field;
  else
    median_func = median_frame;

  for (h = 0; h < height; h++)
  {
    for (w = 0; w < width; w++)
    {
      if (!(pix_flag[w] & PIX_FLAG_DONE)) {
        v = median_func(&frm[w], frm_stride);
//        v = (v*(256-thre) + frm[w]*(thre)) >> 8;
        frm[w] = PIX_CLIP(v);
      }
    }
    frm += frm_stride;
    pix_flag += pix_stride;
  }
}

static void
deno_frame_motion2(int width, int height, unsigned char *frmm1, unsigned char *frm, unsigned char *frmp1, int frm_stride, unsigned char *mref, unsigned char *mref_forw, unsigned char *pix_flag, block_info_t *blk_info)
{
  int w, h;
  int i, j;
  int v;
  int d0, d1;
  int frm_offset, offset;
  int thre;
  unsigned char *fp, *fpm1, *fpp1, *mp, *mpf;
  unsigned char *pix;
  unsigned int blkflag;
  block_info_t *blk;

  thre = Threshold;
  blk = blk_info;
  frm_offset = offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      fp = &frm[frm_offset];
      fpm1 = &frmm1[frm_offset];
      fpp1 = &frmp1[frm_offset];
      mp = &mref[offset];
      mpf = &mref_forw[offset];
      pix = &pix_flag[offset];
      blkflag = blk->flag &
              (BLOCK_FLAG_BACK_AVG|BLOCK_FLAG_FORW_AVG
	      |BLOCK_FLAG_MOTION_BACK_AVG|BLOCK_FLAG_MOTION_FORW_AVG);
      if (blkflag == (BLOCK_FLAG_BACK_AVG|BLOCK_FLAG_FORW_AVG)) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - fpm1[i];
              d0 = (d0<0) ? -d0 : d0;
              d1 = v - fpp1[i];
              d1 = (d1<0) ? -d1 : d1;

              if (d0 <= thre && d1 <= thre) {
	        v = BIAVG(v, fpm1[i], fpp1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d0 <= thre) {
	        v = BKAVG(v, fpm1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d1 <= thre) {
	        v = FWAVG(v, fpp1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          fpm1 += frm_stride; fpp1 += frm_stride;
	}
      }
      else if (blkflag == (BLOCK_FLAG_BACK_AVG|BLOCK_FLAG_MOTION_FORW_AVG)) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - fpm1[i];
              d0 = (d0<0) ? -d0 : d0;
              d1 = v - mpf[i];
              d1 = (d1<0) ? -d1 : d1;

              if (d0 <= thre && d1 <= thre) {
	        v = BIAVG(v, fpm1[i], mpf[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d0 <= thre) {
	        v = BKAVG(v, fpm1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d1 <= thre) {
	        v = FWAVG(v, mpf[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          fpm1 += frm_stride; mpf += frm_stride;
	}
      }
      else if (blkflag == (BLOCK_FLAG_MOTION_BACK_AVG|BLOCK_FLAG_FORW_AVG)) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - mp[i];
              d0 = (d0<0) ? -d0 : d0;
              d1 = v - fpp1[i];
              d1 = (d1<0) ? -d1 : d1;

              if (d0 <= thre && d1 <= thre) {
	        v = BIAVG(v, mp[i], fpp1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d0 <= thre) {
	        v = BKAVG(v, mp[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d1 <= thre) {
	        v = FWAVG(v, fpp1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          mp += frm_stride; fpp1 += frm_stride;
	}
      }
      else if (blkflag == (BLOCK_FLAG_MOTION_BACK_AVG
	                  |BLOCK_FLAG_MOTION_FORW_AVG)) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - mp[i];
              d0 = (d0<0) ? -d0 : d0;
              d1 = v - mpf[i];
              d1 = (d1<0) ? -d1 : d1;

              if (d0 <= thre && d1 <= thre) {
	        v = BIAVG(v, mp[i], mpf[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d0 <= thre) {
	        v = BKAVG(v, mp[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	      else if (d1 <= thre) {
	        v = FWAVG(v, mpf[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          mp += frm_stride; mpf += frm_stride;
	}
      }
      else if (blkflag & BLOCK_FLAG_BACK_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - fpm1[i];
              d0 = (d0<0) ? -d0 : d0;

	      if (d0 <= thre) {
	        v = BKAVG(v, fpm1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          fpm1 += frm_stride;
	}
      }
      else if (blkflag & BLOCK_FLAG_MOTION_BACK_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - mp[i];
              d0 = (d0<0) ? -d0 : d0;

	      if (d0 <= thre) {
	        v = BKAVG(v, mp[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          mp += frm_stride;
	}
      }
      else if (blkflag & BLOCK_FLAG_FORW_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - fpp1[i];
              d0 = (d0<0) ? -d0 : d0;

	      if (d0 <= thre) {
	        v = BKAVG(v, fpp1[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          fpp1 += frm_stride;
	}
      }
      else if (blkflag & BLOCK_FLAG_MOTION_FORW_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
	      v = fp[i];
              d0 = v - mpf[i];
              d0 = (d0<0) ? -d0 : d0;

	      if (d0 <= thre) {
	        v = BKAVG(v, mpf[i]);
                fp[i] = PIX_CLIP(v);
                pix[i] = PIX_FLAG_DONE;
	      }
	    }
          }
	  pix += width; fp += frm_stride;
          mpf += frm_stride;
	}
      }
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }
}

static void
deno_frame2_uv(int width, int height, unsigned char *frmm1, unsigned char *frm, unsigned char *frmp1, int frm_stride, unsigned char *pix_flag, block_info_t *blk_info)
{
  int w, h;
  int i, j;
  int v, errcnt, d0, d1;
  int thre;
  int offset, frm_offset;
  unsigned char *fp, *fpm1, *fpp1;
  unsigned char *pix;
  unsigned int vari_bk, vari_fw;
  unsigned int vari_avgthre;
  block_info_t *blk;

  thre = Threshold_uv;
  vari_avgthre = (thre*thre)*Avg_threshold;
  blk = blk_info;
  offset = frm_offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      fp = &frm[frm_offset];
      fpm1 = &frmm1[frm_offset];
      fpp1 = &frmp1[frm_offset];
      pix = &pix_flag[offset];
      vari_bk = vari_fw = 0;
      errcnt = 0;
      for (j = h; j < h+8; j++)
      {
	for (i = w; i < w+8; i++)
	{
          v = fp[i];
	  d0 = v - fpm1[i];
          d0 = (d0<0) ? -d0 : d0;
	  vari_bk += d0*d0;

          d1 = v - fpp1[i];
          d1 = (d1<0) ? -d1 : d1;
	  vari_fw += d1*d1;

//	  pix[i] = PIX_FLAG_NONE;

	  if (d0 <= thre) {
	    pix[i] |= PIX_FLAG_BACK_AVG;
	  }
	  if (d1 <= thre) {
	    pix[i] |= PIX_FLAG_FORW_AVG;
	  }

	  if (d0 > thre || d1 > thre) {
	    errcnt++;
	  }
        }
        fp += frm_stride;
        fpm1 += frm_stride;
        fpp1 += frm_stride;
	pix += width;
      }

      blk->flag = BLOCK_FLAG_NONE;
      if (vari_bk <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_BACK_AVG;
      if (vari_fw <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_FORW_AVG;

      if (vari_bk <= vari_avgthre && vari_fw <= vari_avgthre)
	blk->flag |= BLOCK_FLAG_AVG;
      blk->vari_bk = vari_bk;
      blk->vari_fw = vari_fw;
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }

  blk = blk_info;
  offset = frm_offset = 0;
  for (h = 0; h < height; h+=8)
  {
    for (w = 0; w < width; w+=8)
    {
      fp = &frm[frm_offset];
      fpm1 = &frmm1[frm_offset];
      fpp1 = &frmp1[frm_offset];
      pix = &pix_flag[offset];
      if (blk->flag & BLOCK_FLAG_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if (!(pix[i] & PIX_FLAG_DONE)) {
              if ((pix[i] & (PIX_FLAG_BACK_AVG|PIX_FLAG_FORW_AVG))
                   == (PIX_FLAG_BACK_AVG|PIX_FLAG_FORW_AVG)) {
	        v = fp[i];
	        v = BIAVG(v, fpm1[i], fpp1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	      else if (pix[i] & PIX_FLAG_BACK_AVG) {
	        v = fp[i];
	        v = BKAVG(v, fpm1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
	      else if (pix[i] & PIX_FLAG_FORW_AVG) {
	        v = fp[i];
	        v = FWAVG(v, fpp1[i]);
                fp[i] = PIX_CLIP(v);
	        pix[i] = PIX_FLAG_DONE;
	      }
            }
	  }
          fp += frm_stride;
          fpm1 += frm_stride;
          fpp1 += frm_stride;
	  pix += width;
	}
      }
      else if (blk->flag & BLOCK_FLAG_BACK_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_BACK_AVG))
	       	==  PIX_FLAG_BACK_AVG) {
	      v = fp[i];
	      v = BKAVG(v, fpm1[i]);
              fp[i] = PIX_CLIP(v);
	      pix[i] = PIX_FLAG_DONE;
	    }
	  }
          fp += frm_stride;
          fpm1 += frm_stride;
	  pix += width;
	}
      }
      else if (blk->flag & BLOCK_FLAG_FORW_AVG) {
        for (j = h; j < h+8; j++) {
	  for (i = w; i < w+8; i++) {
	    if ((pix[i] & (PIX_FLAG_DONE|PIX_FLAG_FORW_AVG))
	       	==  PIX_FLAG_FORW_AVG) {
	      v = fp[i];
	      v = BKAVG(v, fpp1[i]);
              fp[i] = PIX_CLIP(v);
	      pix[i] = PIX_FLAG_DONE;
	    }
	  }
          fp += frm_stride;
          fpp1 += frm_stride;
	  pix += width;
	}
      }
      blk++;
    }
    frm_offset += frm_stride*8;
    offset += width*8;
  }
}

static void
deno_frame3(int width, int height, unsigned char *frm, int stride)
{
  int i, j, d0, d1, v;
  int thre = Threshold;

  for (j = 1; j < height-1; j++)
  {
    for (i = 1; i < width-1; i++)
    {
      v = frm[i];
      d0 = v - frm[i-1];
      d0 = (d0<0) ? -d0:d0;
      d1 = v - frm[i+1];
      d1 = (d1<0) ? -d1:d1;
      if (d0 <= thre && d1 <= thre) {
        v = (v*2 + frm[i-1] + frm[i+1])>>2;
      }
      d0 = v - frm[i-stride];
      d0 = (d0<0) ? -d0:d0;
      d1 = v - frm[i+stride];
      d1 = (d1<0) ? -d1:d1;
      if (d0 <= thre && d1 <= thre) {
        v = (v*2 + frm[i-stride] + frm[i+stride])>>2;
      }
      frm[i] = PIX_CLIP(v);
    }
    frm += stride;
  }
}


static void
rotate_buf(void)
{
  int i;
  vframe_t *tmp;
  tmp = vframe[BUF_NUM-1];
  for (i = BUF_NUM-2; i >= 0; i--)
    vframe[i+1] = vframe[i];
  vframe[0] = tmp;
}

int
denoise3(unsigned char *src, unsigned char *dst)
{
  int i, j, offset, stride;
  int Y_width, Y_height, Y_stride;
  unsigned char *frm;
  unsigned char *dp;

  if (Field_flag) {
    Y_width = Width * 2;
    Y_height = Height >> 1;
    Y_stride = Width * 2;
  } else {
    Y_width = Width;
    Y_height = Height;
    Y_stride = Width;
  }

#if 0
  frm = vframe[0]->offset[0];
//  stride = vframe[0]->stride[0] * 2;
  /* Y top and bottom field */
  lowpass_func0(Y_width, Y_height, src, Y_stride, frm, Y_stride, filt_coef0);
#else
  memcpy(vframe[0]->offset[0], src, Y_width*Y_height);
#endif

  /* U */
  offset = Width*Height;
  frm = vframe[0]->offset[1];
  stride = vframe[0]->stride[1];
  lowpass_func1(Width2, Height2, src+offset, Width2, frm, stride, filt_coef1);

  /* V */
  offset += offset>>2;
  frm = vframe[0]->offset[2];
  stride = vframe[0]->stride[2];
  lowpass_func1(Width2, Height2, src+offset, Width2, frm, stride, filt_coef1);

  if (uninited == 2) {
    rotate_buf();
    uninited = 0;
    return DENOISE3_NEED_MORE_DATA;
  }

  /* Denoise Y plane */
  deno_frame1(Y_width, Y_height,
      vframe[2]->offset[0],
      vframe[1]->offset[0],
      vframe[0]->offset[0], Y_stride,
      pix_flag[0], block_info[0], 0);

  motion_search(Y_width, Y_height,
      vframe[1], vframe[2], vframe[0], block_info[0]);

//  for (i = 0; i < Y_width*Y_height; i++)
//    mref_frame[i] = mref_frame_forward[i] = 0;
//    mref_frame[i] = mref_frame_forward[i] = vframe[1]->offset[0][i] >> 2;

  make_motion_frame(Y_width, Y_height,
      vframe[2]->offset[0], vframe[0]->offset[0], Y_stride,
      mref_frame, mref_frame_forward, Y_stride,
      block_info[0]);

  deno_frame_motion1(Y_width, Y_height,
      vframe[2]->offset[0],
      vframe[1]->offset[0],
      vframe[0]->offset[0], Y_stride,
      mref_frame, mref_frame_forward,
      pix_flag[0], block_info[0]);

  if (Field_flag)
    deno_median(Width, Height, vframe[1]->offset[0], vframe[1]->stride[0],
        pix_flag[0], Width, 1);
  else
    deno_median(Width, Height, vframe[1]->offset[0], vframe[1]->stride[0],
        pix_flag[0], Width, 0);

  deno_frame_motion2(Y_width, Y_height,
      vframe[2]->offset[0],
      vframe[1]->offset[0],
      vframe[0]->offset[0], Y_stride,
      mref_frame, mref_frame_forward,
      pix_flag[0], block_info[0]);

  deno_frame3(Width, Height, vframe[1]->offset[0], vframe[1]->stride[0]);

  /* Denoise U plane */
  deno_frame1(Width2, Height2,
      vframe[2]->offset[1],
      vframe[1]->offset[1],
      vframe[0]->offset[1], vframe[0]->stride[1],
      pix_flag[1], block_info[1], 1);
  deno_median(Width2, Height2, vframe[1]->offset[1], vframe[1]->stride[1],
      pix_flag[1], Width2, 0);
  deno_frame2_uv(Width2, Height2,
      vframe[2]->offset[1],
      vframe[1]->offset[1],
      vframe[0]->offset[1], vframe[0]->stride[1],
      pix_flag[1], block_info[1]);

  deno_frame3(Width2, Height2, vframe[1]->offset[1], vframe[1]->stride[1]);

  /* Denoise V plane */
  deno_frame1(Width2, Height2,
      vframe[2]->offset[2],
      vframe[1]->offset[2],
      vframe[0]->offset[2], vframe[0]->stride[2],
      pix_flag[2], block_info[2], 1);
  deno_median(Width2, Height2, vframe[1]->offset[2], vframe[1]->stride[2],
      pix_flag[2], Width2, 0);
  deno_frame2_uv(Width2, Height2,
      vframe[2]->offset[2],
      vframe[1]->offset[2],
      vframe[0]->offset[2], vframe[0]->stride[2],
      pix_flag[2], block_info[2]);

  deno_frame3(Width2, Height2, vframe[1]->offset[2], vframe[1]->stride[2]);

  laplacian(Width, Height, vframe[1]->offset[0], vframe[1]->stride[0], dst, Width);
#if 0
  offset = Width * Height;
  laplacian(Width2, Height2, frmbuf[1][1]+start_offset[1], stride[1], dst+offset, Width2);
  offset = Width * Height * 5 / 4;
  laplacian(Width2, Height2, frmbuf[1][2]+start_offset[2], stride[2], dst+offset, Width2);
#endif

#if 0
  frm = vframe[1]->offset[0];
  stride = vframe[1]->stride[0];
  dp = dst;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width; i++)
      dp[i] = frm[i];
    dp += Width;
    frm += stride;
  }
#endif

  offset = Width * Height;
  frm = vframe[1]->offset[1];
  stride = vframe[1]->stride[1];
  dp = dst + offset;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++)
      dp[i] = frm[i];
    dp += Width2;
    frm += stride;
  }

  offset = Width * Height * 5 / 4;
  frm = vframe[1]->offset[2];
  stride = vframe[1]->stride[2];
  dp = dst + offset;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++)
      dp[i] = frm[i];
    dp += Width2;
    frm += stride;
  }

  rotate_buf();

  return DENOISE3_OK;
}

void
denoise3_quit(void)
{
  int i;

  if (uninited == 1)
    return;

  free(pix_flag[0]);
  free(mref_frame);
  free(mref_frame_forward);
  for (i = 0; i < 3; i++) {
    free(motion_data_back[i]);
    free(motion_data_forw[i]);
    free(block_info[i]);
  }
  motion_search_quit();
  pix_clip_quit();

  for (i = 0; i < BUF_NUM; i++) {
    vframe[i] = vframe_free(vframe[i]);
  }

  uninited = 1;
}

void
denoise3_print_param(void)
{
  printf("DENOISE3:                       ON\n");
  printf("DENOISE3_THRESHOLD:             %d\n", Threshold);
  printf("DENOISE3_THRESHOLD_UV:          %d\n", Threshold_uv);
  printf("DENOISE3_AVG_THRESHOLD:         %d\n", Avg_threshold);
  printf("DENOISE3_MOTION_THRESHOLD:      %d\n", Motion_threshold);
  printf("DENOISE3_MOTION_RADIUS:         %d\n", Motion_radius);
}

int
denoise3_init(int width, int height, int threshold, int avg_threshold, int motion_threshold, int motion_radius, int field_flag)
{
  double thre, thre2;
  int i, hborder, vborder;
  int num_blocks;

  if (uninited != 1) {
    denoise3_quit();
  }

  Width = width;
  Width2 = width>>1;
  Height = height;
  Height2 = height>>1;

  Threshold = threshold;
  Threshold_uv = threshold;
  Avg_threshold = avg_threshold;
  Motion_threshold = motion_threshold;
  Motion_radius = motion_radius;

  if (field_flag)
    Field_flag = 1;
  else
    Field_flag = 0;

  Threshold2 = threshold>>1;

  thre = threshold;
  if (Field_flag)
    thre2 = thre * 2.0;
  else
    thre2 = thre * 1.0;
  if (thre2 > 255.)
    thre2 = 255.;
  make_filt_coef((255.+thre2)/255., filt_catrom, filt_coef0, 2.0);
  if (filt_coef0[0] == 0)
    lowpass_func0 = lowpass_filt3;
  else
    lowpass_func0 = lowpass_filt5;
  make_filt_coef((255.+thre)/255., filt_cubic, filt_coef1, 2.0);
  if (filt_coef1[0] == 0)
    lowpass_func1 = lowpass_filt3;
  else
    lowpass_func1 = lowpass_filt5;

  hborder = 0;
  vborder = ((Motion_radius + 16 + 15)/16) * 16;
  for (i = 0; i < BUF_NUM; i++)
    vframe[i] = vframe_alloc(width, height, hborder, vborder, 0);

  pix_clip_init();

  pix_flag[0] = (unsigned char*)memalign(16, Width*Height*3/2);
  pix_flag[1] = pix_flag[0] + Width * Height;
  pix_flag[2] = pix_flag[1] + Width2 * Height2;

  num_blocks = (Width/8) * (Height/8);
  motion_data_back[0] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  motion_data_forw[0] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  block_info[0] = (block_info_t*)malloc(sizeof(block_info_t) * num_blocks);
  for (i = 0; i < num_blocks; i++) {
    block_info[0][i].mibk = &motion_data_back[0][i];
    block_info[0][i].mifw = &motion_data_forw[0][i];
  }

  num_blocks = (Width2/8) * (Height2/8);
  motion_data_back[1] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  motion_data_forw[1] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  block_info[1] = (block_info_t*)malloc(sizeof(block_info_t) * num_blocks);
  for (i = 0; i < num_blocks; i++) {
    block_info[1][i].mibk = &motion_data_back[1][i];
    block_info[1][i].mifw = &motion_data_forw[1][i];
  }

  num_blocks = (Width2/8) * (Height2/8);
  motion_data_back[2] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  motion_data_forw[2] =(motion_data_t*)malloc(sizeof(motion_data_t)*num_blocks);
  block_info[2] = (block_info_t*)malloc(sizeof(block_info_t) * num_blocks);
  for (i = 0; i < num_blocks; i++) {
    block_info[2][i].mibk = &motion_data_back[2][i];
    block_info[2][i].mifw = &motion_data_forw[2][i];
  }


  mref_frame = (unsigned char*)memalign(16, Width*Height*3/2);
  mref_frame_forward = (unsigned char*)memalign(16, Width*Height*3/2);
  for (i = 0; i < Width*Height; i++)
    mref_frame[i] = mref_frame_forward[i] = 0;
  for (; i < Width*Height*3/2; i++)
    mref_frame[i] = mref_frame_forward[i] = 128;

  motion_search_init(Width, Height);

  uninited = 2;

  denoise3_print_param();

  return DENOISE3_OK;
}


