/* xwin.h */

#ifndef _XWIN_H
#define _XWIN_H

#include <inttypes.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>


#define XWIN_VO_X11  0
#define XWIN_VO_XSHM 1
#define XWIN_VO_XV   2

#define XWIN_CSP_UNKNOWN 0
#define XWIN_CSP_RGB24   1
#define XWIN_CSP_RGB32   2
#define XWIN_CSP_RGB565  3
#define XWIN_CSP_RGB555  4
#define XWIN_CSP_YV12    5
#define XWIN_CSP_I420    6
#define XWIN_CSP_YUY2    7
#define XWIN_CSP_UYVY    8

#define XWIN_CSP_BGRA32  9
#define XWIN_CSP_BGR24   10
#define XWIN_CSP_ARGB32  11
#define XWIN_CSP_YUV420  12
#define XWIN_CSP_YUV422  13

typedef struct {
  const char *title;
  const char *dpyname;
  Display *dpy;
  Window win;
  GC gc;
  Atom wm_protocols;
  Atom wm_delete_window;
  XShmSegmentInfo shm_info;
  XImage *shm_image;
  XImage *x11_image;
  XvImage *xv_image;
  XvPortID port;

  int vo_type;
  const char *vo_name;
  int shm_flag;
  int win_width, win_height;

  int disp_csp;
  int disp_depth;
  int disp_line_stride;
  int disp_width, disp_height;
  int disp_bytes_per_line;
  int disp_data_size;
  int disp_mode_rgb;
  uint8_t *disp_planes[3];

  int disp_stride[3];
  int disp_offset[3];
  int disp_size[3];

  int image_csp;
  int image_depth;
  int image_line_stride;
  int image_width, image_height;

  int image_stride[3];
  int image_offset[3];
  int image_size[3];
} xwin_t;



xwin_t* xwin_init (int width, int height, int in_csp, int vo_type,
    const char *display_name, const char *title);
int xwin_quit (xwin_t *xw);
int xwin_put_pic (xwin_t *xw, void *pic_data);
int xwin_dispatch_event (xwin_t *xw);
int xwin_get_disp_depth(xwin_t *xw);

int xwin_put_image (xwin_t *xw, uint8_t *image_planes[], int image_stride[]);


#endif /* _XWIN_H */

