/*
 * simple quicktime movie player, needs libquicktime (or quicktime4linux).
 *
 *  (c) 2002 Gerd Knorr <kraxel@bytesex.org>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


#include <avilib.h>
#include <qtime.h>

#include "csp.h"
#include "vcodec.h"
#include "acodec.h"

#include "xwin.h"


/* ------------------------------------------------------------------------ */
/* oss code                                                                 */

#ifndef AFMT_S16_NE
# if BYTE_ORDER == BIG_ENDIAN
#  define AFMT_S16_NE AFMT_S16_BE
# else
#  define AFMT_S16_NE AFMT_S16_LE
# endif
#endif

static int oss_fd = -1;
static int oss_sr,oss_hr;
static int rate, channels, bits;
static int oss_out_delay;
static int oss_fragments, oss_fragsize, oss_fragstotal;

static int
oss_setformat(int chan, int rate)
{
    int hw_afmt = AFMT_S16_NE;
    int hw_chan = chan;
    int hw_rate = rate;
    audio_buf_info abinfo;

    ioctl(oss_fd, SNDCTL_DSP_SETFMT, &hw_afmt);
    if (AFMT_S16_LE != hw_afmt) {
	fprintf(stderr,"ERROR: can't set sound format\n");
	exit(1);
    }
    ioctl(oss_fd, SNDCTL_DSP_CHANNELS, &hw_chan);
    if (chan != hw_chan) {
	fprintf(stderr,"ERROR: can't set sound channels, %d\n", chan);
	exit(1);
    }
    ioctl(oss_fd, SNDCTL_DSP_SPEED, &hw_rate);
    if (rate != hw_rate) {
	oss_sr = rate;
	oss_hr = hw_rate;
	fprintf(stderr,"WARNING: sample rate mismatch (need %d, got %d)\n",
		rate,hw_rate);
    }
    if (ioctl(oss_fd, SNDCTL_DSP_GETODELAY, &oss_out_delay) < 0) {
      perror("SNDCTL_DSP_GETODELAY");
      exit(1);
    }
    if (ioctl(oss_fd, SNDCTL_DSP_GETOSPACE, &abinfo) < 0) {
      perror("SNDCTL_DSP_GETOSPACE");
      exit(1);
    }
    oss_fragments = abinfo.fragments;
    oss_fragsize = abinfo.fragsize;
    oss_fragstotal = abinfo.fragstotal;
    printf("oss_out_delay %d\n", oss_out_delay);
    return 0;
}

static int oss_init(char *dev, int chan, int r)
{
    int trigger;

    oss_fd = open(dev,O_WRONLY | O_NONBLOCK);
    if (-1 == oss_fd) {
	fprintf(stderr,"WARNING: open %s: %s\n",dev,strerror(errno));
	return -1;
    }
    oss_setformat(chan,r);
    trigger = PCM_ENABLE_OUTPUT;
    ioctl(oss_fd,SNDCTL_DSP_SETTRIGGER,&trigger);
    return 0;
}

/* ------------------------------------------------------------------------ */
/* qtime code                                                           */

static avi_t *avi;
static int hasaudio, hasvideo;
static qtime_t *qt = NULL;
static qtime_track_t *vqtrk = NULL;
static qtime_track_t *aqtrk = NULL;
#define FILE_FORMAT_UNKNOWN 0
#define FILE_FORMAT_AVI     1
#define FILE_FORMAT_MOV     2
static int file_format;

static xwin_t *xw;
static int width = 320, height = 240;
static int drop = 0, droptotal = 0;
static double fps = 0;
static unsigned char *frame_buf;
static unsigned char *read_buf;
static unsigned char *pic_data;
static uint32_t vcodec;
static int csp;
static int max_frame_num;
static int frame_num;

static uint32_t acodec = ACODEC_UNKNOWN;
//static int audio_frame_samples = 0;
static int bytespspl = 0;
//static int frame_size = 0;
static short *pcm_buf[2];
static unsigned char *audio_read_buf;
static unsigned char *audio_buf;
static int audio_buf_pos = 0;
static int audio_read_bytes = 0;
static int audio_samples = 0;
static int audio_gap = 0;
//static int avsync_step = 3;
//static int avsync_next_second = 0;
static struct timeval start;

int
check_file_format_avi(const char *name)
{
  int fd;
  int format;
  unsigned char head_buf1[8];
  unsigned char head_buf2[8];

  format = FILE_FORMAT_UNKNOWN;

  if ((fd = open(name, O_RDONLY)) < 0) {
    fprintf(stderr, "check_file_format: cannot open %s.\n", name);
    exit(1);
//    return FILE_FORMAT_UNKNOWN;
  }
  if (read(fd, head_buf1, 8) == 8) {
    if (strncasecmp(head_buf1, "RIFF", 4) == 0) {
      if (read(fd, head_buf2, 8) == 8) {
        if (strncasecmp(head_buf2, "AVI ", 4) == 0) {
          format = FILE_FORMAT_AVI;
        }
      }
    }
  }
  close(fd);

  return format;
}

static void file_init(FILE *fp, char *filename)
{
  const char *comp, *acomp;
  int audio_bytes;

  file_format = check_file_format_avi(filename);

  if (file_format == FILE_FORMAT_AVI) {
    avi = AVI_open_input_file(filename, 1);
    if (NULL == avi) {
      fprintf(stderr,"ERROR: can't open file: %s\n",filename);
      exit(1);
    }
    max_frame_num = AVI_video_frames(avi);
    if (max_frame_num > 0) {
      hasvideo = 1;
      width = AVI_video_width(avi);
      height = AVI_video_height(avi);
      fps = AVI_frame_rate(avi);
      comp = AVI_video_compressor(avi);
      vcodec = vcodec_fourcc_str_to_codec(comp);
    } else {
      hasvideo = 0;
      width = 0;
      height = 0;
      fps = 0;
      comp = NULL;
      vcodec = VCODEC_UNKNOWN;
      max_frame_num = 0;
    }

    audio_bytes = AVI_audio_bytes(avi);
    if (audio_bytes > 0) {
      hasaudio = 1;
      rate = AVI_audio_rate(avi);
      channels = AVI_audio_channels(avi);
      bits = AVI_audio_bits(avi);
      bytespspl = ((bits+7)/8) * channels;
      acodec = acodec_code_to_codec(AVI_audio_format(avi));
      switch (acodec) {
	case ACODEC_PCM:
	case ACODEC_PCM_LE:
	  audio_read_bytes = 576 * bytespspl;
	  break;
	case ACODEC_MP3:
	  audio_read_bytes = 288;
	  break;
      }
//      audio_frame_samples = 1152 * 4 * bytespspl;
    } else {
      hasaudio = 0;
      rate = 0;
      channels = 0;
      bits = 0;
      acodec = ACODEC_UNKNOWN;
      audio_read_bytes = 0;
//      audio_frame_samples = 0;
    }
  } else {
    qt = qtime_open_read(filename);
    if (NULL == qt) {
	fprintf(stderr,"ERROR: can't open file: %s\n",filename);
	exit(1);
    }
    file_format = FILE_FORMAT_MOV;

    vqtrk = qtime_get_next_video_track(qt, NULL);
    aqtrk = qtime_get_next_sound_track(qt, NULL);

    if (vqtrk) {
      hasvideo = 1;
      width  = qtime_track_get_video_width(vqtrk);
      height = qtime_track_get_video_height(vqtrk);
      fps = qtime_track_get_video_fps(vqtrk);
      comp = qtime_track_get_compressor(vqtrk);
      max_frame_num = qtime_track_get_video_frames(vqtrk);
      vcodec = vcodec_fourcc_str_to_codec(comp);
    } else {
      hasvideo = 0;
      width = 0;
      height = 0;
      fps = 0;
      comp = NULL;
      vcodec = VCODEC_UNKNOWN;
      max_frame_num = 0;
    }

    if (aqtrk) {
      hasaudio = 1;
      rate = qtime_track_get_audio_rate(aqtrk);
      channels = qtime_track_get_audio_channels(aqtrk);
      bits = qtime_track_get_audio_bits(aqtrk);
      acomp = qtime_track_get_compressor(aqtrk);
      acodec = acodec_fourcc_to_codec(acomp);
      bytespspl = ((bits+7)/8) * channels;
      switch (acodec) {
	case ACODEC_PCM:
	case ACODEC_PCM_LE:
	case ACODEC_PCM_BE:
	  audio_read_bytes = 576 * bytespspl;
	  break;
	case ACODEC_MP3:
	  audio_read_bytes = 288;
	  break;
      }
    } else {
      hasaudio = 0;
      rate = 0;
      channels = 0;
      bits = 0;
      acomp = NULL;
      acodec = ACODEC_UNKNOWN;
      bytespspl = 0;
//      audio_frame_samples = 0;
    }
  }

  if (hasvideo) {
    if (vcodec == VCODEC_UNKNOWN) {
      fprintf(fp, "fourcc '%.4s' is unsupported.\n", comp);
      exit(1);
    }
    if (vcodec_decode_set_codec(vcodec) == VCODEC_FAIL) {
      fprintf(fp, "fourcc '%.4s' is unsupported.\n", comp);
      exit(1);
    }
    csp = vcodec_decode_csp_cap(CSP_UNKNOWN);

    fprintf(fp,"INFO: playing %s\n",filename);
    /* print video info */
    fprintf(fp, "  video:\n");
    fprintf(fp, "      width : %d\n"
		"      height: %d\n"
		"      rate  : %.2f fps\n"
		"      codec : %s\n",
		width, height, fps, comp);
  }

  if (hasaudio) {
    if (acodec_decode_set_codec(acodec) != ACODEC_OK) {
      fprintf(stderr,"WARNING: unsupported audio codec\n");
      hasaudio = 0;
      return;
    }
    if (acodec_get_cap(acodec, ACODEC_CAP_ANALYSIS) != ACODEC_CAP_NONE) {
      int ret, size = 0;
      int r, c, b;
      uint8_t rbuf[640*1024];
      int num_samples;

      ret = ACODEC_NEED_MORE_DATA;
      while (ret == ACODEC_NEED_MORE_DATA) {
	if (file_format == FILE_FORMAT_AVI)
	  size = AVI_read_audio(avi, rbuf, audio_read_bytes);
	else if (file_format == FILE_FORMAT_MOV) {
	  if (acodec == ACODEC_MP3)
	    num_samples = 1;
	  else if (acodec == ACODEC_PCM || acodec == ACODEC_PCM_LE ||
	           acodec == ACODEC_PCM_BE)
	    num_samples = audio_read_bytes / bytespspl;
          size = qtime_track_read_audio(aqtrk, rbuf, &num_samples);
	}
        ret = acodec_decode_analyze_stream(rbuf, size, &r, &c, &b);
        if (ret == ACODEC_FAIL) {
          fprintf(stderr, "acodec_decode_analyze_stream() error.\n");
	  hasaudio = 0;
          return;
        }
      }
      if (bits != 16) {
        fprintf(stderr, "unsupported audio bits.\n");
	hasaudio = 0;
        return;
      }

      rate = r;
      channels = c;
      bits = b;
      bytespspl = ((bits+7)/8) * channels;

      if (file_format == FILE_FORMAT_AVI)
	AVI_set_audio_position(avi, 0);
      else if (file_format == FILE_FORMAT_MOV)
        qtime_track_set_audio_position(aqtrk, 0);
    }

    if (channels > 2) {
      fprintf(stderr, "unsupported channels %d.\n", channels);
      hasaudio = 0;
      return;
    }

    pcm_buf[0] = (short*)malloc(rate * sizeof(short) * 2);
    pcm_buf[1] = (short*)malloc(rate * sizeof(short) * 2);
    audio_buf = (unsigned char*)malloc(rate * bytespspl * 2);
    audio_read_buf = (unsigned char*)malloc(rate * bytespspl * 2);
    if (!pcm_buf[0] || !pcm_buf[1] || !audio_buf || !audio_read_buf * 2) {
      fprintf(stderr, "memory allocate error.\n");
      exit(1);
    }

    if (acodec_decode_init(acodec, rate, channels, bits) != ACODEC_OK) {
      fprintf(stderr, "acodec_decode_init error.\n");
      hasaudio = 0;
      return;
    }
    if (-1 == oss_init("/dev/dsp", channels, rate)) {
      hasaudio = 0;
      return;
    }

    /* print audio info */
    fprintf(fp,"  audio:\n");
    fprintf(fp, "      rate  : %d Hz\n"
		"      bits  : %d\n"
		"      chans : %d\n"
		"      codec : %s\n",
		rate, bits, channels, acodec_codec_to_fourcc(acodec));
  }

  if (!hasvideo && !hasaudio) {
    fprintf(stderr,"ERROR: no playable stream found\n");
    exit(1);
  }
}

static int frame_blit(void)
{
    int i;
    int keyframe, size=0, flag;
    
    if (!hasvideo)
      return 0;

    if (0 == frame_num) {
//      int xcsp;
	/* init */
        if (vcodec_decode_csp_cap(csp) != csp) {
	  fprintf(stderr, "cannot decode %s\n", csp_to_name(csp));
	  exit(1);
	}
	frame_buf = memalign(16, width * height * 4);
	read_buf = memalign(16, width * height * 4);
        pic_data = frame_buf;

#if 0
	switch (csp) {
	case CSP_RGB24:
	  xcsp = XWIN_CSP_RGB24;
	  break;
	case CSP_YUV422:
	  xcsp = XWIN_CSP_YUY2;
	  break;
	case CSP_YV12:
	  xcsp = XWIN_CSP_YV12;
	  break;
	case CSP_YUV420P:
	case CSP_I420:
	  xcsp = XWIN_CSP_YUV420;
          break;
	default:
	    fprintf(stderr,"ERROR: internal error at %s:%d\n",
		    __FILE__,__LINE__);
	    exit(1);
	}
	xw = xwin_init(width, height, xcsp, XWIN_VO_XV, NULL, "ndeplay");
#endif
	if (vcodec_decode_init(vcodec, width, height, csp) !=
	    VCODEC_OK) {
	  fprintf(stderr,"ERROR: video codec '%s' initialize failed.\n", vcodec_get_codec_name(vcodec));
	  exit(1);
	}
    }

    if (drop) {
	droptotal += drop;
	fprintf(stderr,"dropped %d frame(s)\r",droptotal);
	for (i = 0; i < drop; i++) {
	  flag = VCODEC_NULL_DECODE;
	  if (file_format == FILE_FORMAT_AVI)
            size = AVI_read_frame(avi, read_buf, &keyframe);
	  else if (file_format == FILE_FORMAT_MOV)
            size = qtime_track_read_video(vqtrk, read_buf, &keyframe);
	  if (size < 0)
	    return -1;
	  else if (size > 0)
            vcodec_decode(pic_data, read_buf, size, &flag);
	  frame_num++;
	}
	drop = 0;
    }
    if (file_format == FILE_FORMAT_AVI)
      size = AVI_read_frame(avi, read_buf, &keyframe);
    else if (file_format == FILE_FORMAT_MOV)
      size = qtime_track_read_video(vqtrk, read_buf, &keyframe);
    else
      size = 0;
    flag = 0;
    if (size < 0)
      return -1;
    else if (size > 0)
      vcodec_decode(pic_data, read_buf, size, &flag);
    frame_num++;

    xwin_put_pic (xw, pic_data);

    if (frame_num > max_frame_num)
	return -1;
    return 0;
}

static int audio_write(void)
{
  int rc, i, c;
  int num_spl, spls;
  int size;
  int put_bytes = 0;
  struct timeval now;

  if (!hasaudio)
    return 0;

  while (audio_buf_pos <= 0) {
    if (file_format == FILE_FORMAT_AVI)
      size = AVI_read_audio(avi, audio_read_buf, audio_read_bytes);
    else {
      num_spl = audio_read_bytes / bytespspl;
      size = qtime_track_read_audio(aqtrk, audio_read_buf, &num_spl);
    }
    if (size <= 0)
      return -1;

    spls = acodec_decode((void**)pcm_buf, audio_read_buf, size);
    if (spls == 0)
      continue;
    for (c = 0; c < channels; c++) {
      short *sp = (short*)audio_buf;
      sp += c;
      for (i = 0; i < spls; i++) {
        *sp = pcm_buf[c][i];
        sp += channels;
      }
    }
    audio_buf_pos = spls * bytespspl;
  }

  put_bytes = audio_buf_pos;
  rc = write(oss_fd, audio_buf, audio_buf_pos);
  if (rc >= 0) {
    if (rc != audio_buf_pos) {
      fprintf(stderr, "write_audio: put_bytes %d, rc %d.\n", put_bytes, rc);
      if (rc > 0) {
        memmove(audio_buf, audio_buf + rc, audio_buf_pos - rc);
        audio_buf_pos -= rc;
        if (audio_buf_pos < 0) {
          fprintf(stderr, "write_audio: audio_buf_pos negative %d.\n", audio_buf_pos);
          audio_buf_pos = 0;
        }
	return 0;
      }
    } else {
      audio_buf_pos = 0;
    }
    audio_samples += rc / bytespspl;
  } else if (errno == EAGAIN) {
    return 0;
  }

  switch (rc) {
    case -1:
      perror("write dsp");
      close(oss_fd);
      oss_fd = -1;
      aqtrk = NULL;
      break;
    case 0:
      fprintf(stderr,"write dsp: Huh? no data written?\n");
      close(oss_fd);
      oss_fd = -1;
      aqtrk = NULL;
      break;
    default:
      break;
  }

  {
  long msec, amsec;
  int out_delay;

  out_delay = (int)((double)(oss_fragsize*oss_fragstotal/bytespspl) / (double)rate * 1000);

  gettimeofday(&now,NULL);
  msec  = (start.tv_sec  - now.tv_sec)  * 1000;
  msec += (start.tv_usec - now.tv_usec) / 1000;
//  if (now.tv_sec > avsync_next_second) {
    amsec = (long)((double)(audio_samples) / (double)rate * 1000.0);
    amsec -= out_delay;
    amsec += msec;
    amsec += audio_gap;
    if (amsec < -150)
      audio_gap +=  -amsec / 4;
    else if (amsec > 150)
      audio_gap -= amsec / 4;
   // avsync_next_second = now.tv_sec + avsync_step;
//    avsync_next_second = now.tv_sec + 1;
//    printf("audio_gap %d.\n", audio_gap);
//    printf("amsec = %d, droptotal %d.\n", amsec, droptotal);
//    printf("out_delay %d\n", out_delay);
//  }
  }

  return 0;
}

static void frame_delay(struct timeval *start, struct timeval *wait)
{
    struct timeval now;
    long msec;

    gettimeofday(&now,NULL);
    msec  = (start->tv_sec  - now.tv_sec)  * 1000;
    msec += (start->tv_usec - now.tv_usec) / 1000;
    msec += frame_num * 1000 / fps;
    msec += audio_gap;
    if (msec < 0) {
	drop = -msec * fps / 1000;
	wait->tv_sec  = 0;
	wait->tv_usec = 0;
    } else {
	wait->tv_sec  = msec / 1000;
	wait->tv_usec = (msec % 1000) * 1000;
    }
}

/* ------------------------------------------------------------------------ */
/* main                                                                     */

static void usage(FILE *fp, char *prog)
{
    char *p;

    p = strrchr(prog,'/');
    if (p) prog = p+1;
    fprintf(fp,
	    "\n"
	    "Very simple quicktime movie player for X11.  Just playes\n"
	    "the movie and nothing else.  No fancy gui, no controls.\n"
	    "\n"
	    "You can quit with 'Q' and 'ESC' keys.\n"
	    "\n"
	    "usage: %s [ options ] <file>\n"
	    "options:\n"
	    "  -noxv   don't use the Xvideo extention\n"
	    "  -nogl   don't use OpenGL\n"
	    "\n",
	    prog);
}


int main(int argc, char *argv[])
{
    struct timeval wait;
    int video_blit = 0;
    
    /* open file */
    if (argc < 2) {
	usage(stderr,argv[0]);
	exit(1);
    }
    file_init(stdout,argv[1]);
    frame_num = 0;

    if (hasvideo)
      xw = xwin_init(width, height, csp, XWIN_VO_XV, NULL, "ndeplay");

    /* enter main loop */
    gettimeofday(&start,NULL);
    for (;;) {
	int rc,max;
	fd_set rd,wr;
//	XEvent event;

//	if (True == XCheckMaskEvent(dpy, ~0, &event)) {
//	    XtDispatchEvent(&event);
	if ('q' == xwin_dispatch_event(xw)) {
	  break;
	} else {
//	    XFlush(dpy);
	    FD_ZERO(&rd);
	    FD_ZERO(&wr);
//	    FD_SET(ConnectionNumber(dpy),&rd);
//	    max = ConnectionNumber(dpy);
	    max = 0;
	    if (hasaudio) {
		FD_SET(oss_fd,&wr);
		if (oss_fd > max)
		    max = oss_fd;
	    }
	    if (hasvideo) {
		frame_delay(&start,&wait);
		if (wait.tv_usec < 1000)
		  video_blit = 1;
	    } else {
		wait.tv_sec  = 1;
		wait.tv_usec = 0;
	    }
	    rc = select(max+1,&rd,&wr,NULL,&wait);
	    if (hasaudio && FD_ISSET(oss_fd,&wr))
	      if (0 != audio_write()) {
		hasaudio = 0;
//		printf("audio break\n");
//	        break;
	      }
//	    if (hasvideo && 0 == rc)
	    if (hasvideo)
	      if (0 == rc || video_blit) {
                if (0 != frame_blit()) {
		  hasvideo = 0;
//		  printf("video break\n");
//	          break;
		}
		video_blit = 0;
	      }
	}
	if (!hasvideo && !hasaudio)
	  break;
    }
    return 0;
}

