/* oss.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <linux/soundcard.h>
//#define __USE_GNU
#include <string.h>

#include "oss.h"


static const char *default_device_name = "/dev/dsp";
static const char *default_mixer_device_name = "/dev/mixer/";
static const char *mix_dev_names[] = SOUND_DEVICE_NAMES;

int oss_get_rate(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->rate;
}

int oss_get_channels(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->channels;
}

int oss_get_bitspspl(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->bitspspl;
}

int oss_get_bytespspl(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->bytespspl;
}

int oss_get_fragments(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->fragments;
}

int oss_get_fragsize(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->fragsize;
}

int oss_get_fragstotal(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->fragstotal;
}

int oss_get_format(oss_t *oss)
{
  if (oss->fd == -1) return OSS_FAIL;
  return oss->format;
}

#if 0
int
oss_trigger (void)
{
  int flag;
  int arg;

  if (oss->fd == -1)
    return OSS_FAIL;
  if (ioctl(oss->fd, SNDCTL_DSP_GETCAPS, &flag) < 0) {
    perror("SNDCTL_DSP_GETCAPS");
    return OSS_FAIL;
  }
  if (!(flag & DSP_CAP_TRIGGER)) {
    fprintf(stderr, "oss_trigger: Device does't support trigger.\n");
    return OSS_FAIL;
  }
  arg = 0;
  if (ioctl(oss->fd, SNDCTL_DSP_SETTRIGGER, &arg) < 0) {
    perror("SNDCTL_DSP_SETTRIGGER");
    return OSS_FAIL;
  }
  arg = PCM_ENABLE_INPUT;
  if (ioctl(oss->fd, SNDCTL_DSP_SETTRIGGER, &arg) < 0) {
    perror("SNDCTL_DSP_SETTRIGGER");
    return OSS_FAIL;
  }
  return OSS_OK;
}
#endif

int
oss_set_param(oss_t *oss, int rate, int channels, int bitspspl)
{
  audio_buf_info buf_info;
  int r = rate;
  int c = channels;
  int bits = bitspspl;
  int format;
  int frg;

  if (oss->fd == -1)
    return OSS_FAIL;

  switch (bitspspl) {
    case 8:
      format = AFMT_U8;
      break;
    case 16:
      format = AFMT_S16_LE;
      break;
    default:
      return OSS_FAIL;
  }

  if (ioctl(oss->fd, SNDCTL_DSP_RESET, NULL) < 0) {
    perror("SNDCTL_DSP_RESET");
    return OSS_FAIL;
  }

  frg = (0x7fff << 16) + 13;
  if (ioctl(oss->fd, SNDCTL_DSP_SETFRAGMENT, &frg) < 0) {
    perror("SNDCTL_DSP_SETFRAGMENT");
    return OSS_FAIL;
  }

  if (ioctl(oss->fd, SNDCTL_DSP_SETFMT, &format) < 0) {
    perror("SNDCTL_DSP_SETFMT");
    return OSS_FAIL;
  }
  switch (format) {
    case AFMT_U8:
      bits = 8;
      break;
    case AFMT_S16_LE:
      bits = 16;
      break;
    default:
      return OSS_FAIL;
  }
  if (bits != bitspspl) {
    fprintf(stderr, "oss_set_param: bitspspl error, bitspspl %d, %d\n", bitspspl, bits);
    return OSS_FAIL;
  }

  if (ioctl(oss->fd, SNDCTL_DSP_SPEED, &r) < 0) {
    perror("SNDCTL_DSP_SPEED");
    return OSS_FAIL;
  }
  if (r != rate) {
    fprintf(stderr, "oss_set_param: rate error, rate %d, %d\n", rate, r);
    return OSS_FAIL;
  }

  if (ioctl(oss->fd, SNDCTL_DSP_CHANNELS, &c) < 0) {
    perror("SNDCTL_DSP_CHANNELS");
    return OSS_FAIL;
  }
  if (c != channels) {
    fprintf(stderr, "oss_set_param: channels error, channels %d, %d\n", channels, c);
    return OSS_FAIL;
  }

  if (ioctl(oss->fd, SNDCTL_DSP_GETISPACE, &buf_info) < 0) {
    perror("SNDCTL_DSP_GETISPACE");
    return OSS_FAIL;
  }

  //int b;
  //if (ioctl (oss->fd, SNDCTL_DSP_GETBLKSIZE, &b) < 0) {
  //  perror("SNDCTL_DSP_GETBLKSIZE");
  //  return OSS_FAIL;
  //}

  oss->rate = r;
  oss->channels = c;
  oss->bytespspl = c * ((bits+7)/8);
  oss->bitspspl = bits;
  oss->format = format;
  oss->fragments = buf_info.fragments;
  oss->fragsize = buf_info.fragsize;
  oss->fragstotal = buf_info.fragstotal;
//  oss_trigger ();

  return OSS_OK;
}

int
oss_close(oss_t *oss)
{
  int fd;
  if (oss->fd == -1)
    return OSS_OK;
  fd = oss->fd;
  oss->fd = -1;
  if (close(fd) < 0) {
    perror("oss_close");
    return OSS_FAIL;
  }
  return OSS_OK;
}

int
oss_open (oss_t *oss, const char* device_name)
{
  const char *name;

  if (oss->fd != -1)
    oss_close(oss);

  if (device_name == NULL) {
    name = default_device_name;
  } else {
    name = (const char*)strndup(device_name, 1024);
  }
  oss->dev_name = name;
  if ((oss->fd = open(name, O_RDWR)) < 0) {
    fprintf(stderr, "oss_open: %s open failer ", name);
    perror ("");
    oss->fd = -1;
    return OSS_FAIL;
  }
  return OSS_OK;
}

int
oss_quit(oss_t *oss)
{
  int ret = OSS_FAIL;
  if (oss->fd != -1)
    ret = oss_close(oss);
  if (oss->dev_name && oss->dev_name != default_device_name)
    free((char*)oss->dev_name);
  if (oss->mix_dev_name && oss->mix_dev_name != default_mixer_device_name)
    free((char*)oss->mix_dev_name);
  free(oss);
  return ret;
}

oss_t*
oss_init(const char *device_name, int rate, int channels, int bitpspl)
{
  oss_t *oss;

  oss = (oss_t*)malloc(sizeof(oss_t));
  if (!oss) {
    fprintf(stderr, "oss_init: memory error.\n");
    return NULL;
  }
  oss->fd = -1;
  oss->dev_name = NULL;
  oss->rate       =
  oss->channels   =
  oss->bytespspl  =
  oss->bitspspl   =
  oss->format     =
  oss->fragments  =
  oss->fragsize   =
  oss->fragstotal = 0;

  oss->mix_fd = -1;
  oss->mix_dev_name = NULL;
  oss->device_mask =
  oss->reccap_mask =
  oss->stereo_mask =
  oss->recsrc_mask =
  oss->excl_input  = 0;

  if (oss_open(oss, device_name) == OSS_FAIL) {
    oss_quit(oss);
    return NULL;
  }

  if (oss_set_param(oss, rate, channels, bitpspl) == OSS_FAIL) {
    oss_quit(oss);
    return NULL;
  }
  return oss;
}

int
oss_read_data(oss_t *oss, char* buf, int buf_size)
{
  int result;
  int rd;

  rd = 0;
  while (buf_size > 0) {
    result = read (oss->fd, buf, buf_size);
    if (result < 0) {
      result = 0;
      perror("oss_read_data");
      return 0;
    }
    buf += result;
    buf_size -= result;
    rd += result;
  }
  return rd;
}

#if 0
int
oss_write (oss_t *oss, char* buf, int byte_length)
{
  return write(oss->fd, buf, byte_length);
}

int
oss_post (oss_t *oss)
{
  if (oss->fd == -1)
    return OSS_FAIL;

  if (ioctl (oss->fd, SNDCTL_DSP_POST) < 0) {
    perror ("SNDCTL_DSP_POST");
    return OSS_FAIL;
  }
  return OSS_OK;
}

int
oss_sync (oss_t *oss)
{
  if (oss->fd == -1)
    return OSS_FAIL;

  if (ioctl (oss->fd, SNDCTL_DSP_SYNC) < 0) {
    perror ("SNDCTL_DSP_SYNC");
    return OSS_FAIL;
  }
  return OSS_OK;
}
#endif

/* mixer functions */

int
oss_mixer_name_to_id(const char *name)
{
  int i, id = OSS_FAIL;

  for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
    if (strcasecmp(mix_dev_names[i], name) == 0) {
      id = i;
      break;
    }
  }
  return id;
}

const char*
oss_mixer_id_to_name(int id)
{
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return NULL;
  return mix_dev_names[id];
}

#define MCHK(m,n) ((m)&(1<<(n)))

int
oss_mixer_dev_id(oss_t *oss, const char *name)
{
  int i, id;

  if (!oss->device_mask)
    return OSS_FAIL;

  id = OSS_FAIL;
  for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
    if (MCHK(oss->device_mask, i)) {
      if (strcasecmp(mix_dev_names[i], name) == 0) {
	id = i;
	break;
      }
    }
  }
  return id;
}

int
oss_mixer_stereo_cap(oss_t *oss, int id)
{
  int ret = OSS_FAIL;

  if (!MCHK(oss->device_mask, id))
    return ret;
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (MCHK(oss->stereo_mask, id))
    ret = 1;
  else
    ret = 0;
  return ret;
}

int
oss_mixer_record_cap(oss_t *oss, int id)
{
  int ret = OSS_FAIL;

  if (!MCHK(oss->device_mask, id))
    return ret;

  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (MCHK(oss->reccap_mask, id))
    ret = 1;
  else
    ret = 0;
  return ret;
}

int
oss_mixer_record_enable(oss_t *oss, int id)
{
  int ret = OSS_FAIL;
  unsigned int mask;

  if (oss->mix_fd == -1)
    return ret;
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (!MCHK(oss->device_mask, id))
    return ret;
  if (!MCHK(oss->reccap_mask, id))
    return ret;
  if (oss->excl_input)
    mask = (1<<id);
  else {
    mask = oss->recsrc_mask;
    mask |= (1<<id);
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_WRITE_RECSRC, &mask) == -1) {
    perror("SOUND_MIXER_WRITE_RECSRC");
    return ret;
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_RECSRC, &mask) == -1) {
    perror("SOUND_MIXER_READ_RECSRC");
    return ret;
  }
  oss->recsrc_mask = mask;
  if (MCHK(oss->recsrc_mask, id))
    ret = OSS_OK;
  return ret;
}

int
oss_mixer_record_disable(oss_t *oss, int id)
{
  int ret = OSS_FAIL;
  int mask;

  if (oss->mix_fd == -1)
    return ret;
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (!MCHK(oss->device_mask, id))
    return ret;
  if (!MCHK(oss->reccap_mask, id))
    return ret;
  mask = oss->recsrc_mask;
  mask &= ~(1<<id);
  if (ioctl(oss->mix_fd, SOUND_MIXER_WRITE_RECSRC, &mask) == -1) {
    perror("SOUND_MIXER_WRITE_RECSRC");
    return ret;
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_RECSRC, &mask) == -1) {
    perror("SOUND_MIXER_READ_RECSRC");
    return ret;
  }
  oss->recsrc_mask = mask;
  if ((!MCHK(oss->recsrc_mask, id)) || (id == SOUND_MIXER_MIC))
    ret = OSS_OK;
  return ret;
}

int
oss_mixer_vol_set(oss_t *oss, int id, int left, int right)
{
  int ret = OSS_FAIL;
  int vol;

  if (oss->mix_fd == -1)
    return ret;
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (!MCHK(oss->device_mask, id))
    return ret;

  vol = (left&0xff);
  if (MCHK(oss->stereo_mask, id))
    vol |= ((right&0xff)<<8);

  if (ioctl(oss->mix_fd, MIXER_WRITE(id), &vol) == -1) {
    perror("MIXER_WRITE");
    return ret;
  }
  return OSS_OK;
}

int
oss_mixer_vol_get(oss_t *oss, int id, int *left, int *right)
{
  int ret = OSS_FAIL;
  int vol;

  if (oss->mix_fd == -1)
    return ret;
  if (id < 0 || id > SOUND_MIXER_NRDEVICES)
    return ret;
  if (!MCHK(oss->device_mask, id))
    return ret;

  if (ioctl(oss->mix_fd, MIXER_READ(id), &vol) == -1) {
    perror("MIXER_WRITE");
    return ret;
  }
  *left = (vol&0xff);
  if (MCHK(oss->stereo_mask, id))
    *right = ((vol>>8)&0xff);
  else if (right)
    *right = -1;
  return OSS_OK;
}

int
oss_mixer_close(oss_t *oss)
{
  int fd;

  if (oss->mix_fd == -1)
    return OSS_OK;
  fd = oss->mix_fd;
  oss->mix_fd = -1;
  if (close(fd) < 0) {
    perror("oss_mixer_close");
    return OSS_FAIL;
  }
  return OSS_OK;
}

int
oss_mixer_open (oss_t *oss, const char* device_name)
{
  const char *name;
  int cap_mask;

  if (oss->mix_fd != -1)
    oss_mixer_close(oss);

  if (device_name == NULL)
    name = default_mixer_device_name;
  else {
    name = device_name;
  }
  if (oss->mix_dev_name && oss->mix_dev_name != default_mixer_device_name)
    free((char*)oss->mix_dev_name);
  oss->mix_dev_name = name;
  if ((oss->mix_fd = open(name, O_RDWR)) < 0) {
    fprintf(stderr, "oss_mixer_open: %s open failer ", name);
    perror ("");
    oss->mix_fd = -1;
    return OSS_FAIL;
  }

  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_DEVMASK, &oss->device_mask)
      == -1) {
    perror("SOUND_MIXER_READ_DEVMASK");
    oss_mixer_close(oss);
    return OSS_FAIL;
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_RECMASK, &oss->reccap_mask)
      == -1) {
    perror("SOUND_MIXER_READ_RECMASK");
    oss_mixer_close(oss);
    return OSS_FAIL;
  }
  if (ioctl(oss->mix_fd,SOUND_MIXER_READ_STEREODEVS,&oss->stereo_mask)
      == -1) {
    perror("SOUND_MIXER_READ_STEREODEVS");
    oss_mixer_close(oss);
    return OSS_FAIL;
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_CAPS, &cap_mask)
      == -1) {
    perror("SOUND_MIXER_READ_CAPS");
    oss_mixer_close(oss);
    return OSS_FAIL;
  }
  oss->excl_input = 0;
  if (cap_mask & SOUND_CAP_EXCL_INPUT) {
    oss->excl_input = 1;
  }
  if (ioctl(oss->mix_fd, SOUND_MIXER_READ_RECSRC, &oss->recsrc_mask)
      == -1) {
    perror("SOUND_MIXER_READ_RECSRC");
    oss_mixer_close(oss);
    return OSS_FAIL;
  }
  return OSS_OK;
}

void
oss_print_param(oss_t *oss)
{
  int i;
  int first = 1;
  fprintf(stdout, "AUDIO DEVICE:            %s\n", oss->dev_name);
  fprintf(stdout, "AUDIO MIXER DEVICE:      %s\n", oss->mix_dev_name);
  fprintf(stdout, "AUDIO SOURCE:            ");
  for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
    if (MCHK(oss->recsrc_mask, i)) {
      if (first) {
	fprintf(stdout, "%s", mix_dev_names[i]);
	first = !first;
      } else
	fprintf(stdout, ",%s", mix_dev_names[i]);
    }
  }
  fprintf(stdout, "\n");
}

