/* oss.h */

#ifndef OSS_H
#define OSS_H

#define OSS_OK    0
#define OSS_FAIL -1

typedef struct {
  const char *device_name;
} oss_param_t;

typedef struct {
  const char *dev_name;
  int fd;
  int rate;
  int channels;
  int bytespspl;
  int bitspspl;
  int format;
  int fragments;
  int fragsize;
  int fragstotal;

  const char *mix_dev_name;
  int mix_fd;
  unsigned int device_mask;
  unsigned int reccap_mask;
  unsigned int stereo_mask;
  unsigned int recsrc_mask;
  unsigned int excl_input;
} oss_t;

int oss_get_rate(oss_t *oss);
int oss_get_channels(oss_t *oss);
int oss_get_bitspspl(oss_t *oss);
int oss_get_bytespspl(oss_t *oss);
int oss_get_fragments(oss_t *oss);
int oss_get_fragsize(oss_t *oss);
int oss_get_fragstotal(oss_t *oss);
int oss_get_format(oss_t *oss);
//int oss_trigger (oss_t *oss);
int oss_set_param(oss_t *oss, int rate, int channels, int bitspspl);
int oss_close(oss_t *oss);
int oss_open (oss_t *oss, const char* device_name);
int oss_quit(oss_t *oss);
oss_t* oss_init(const char *device_name, int rate, int channels, int bitpspl);
int oss_read_data(oss_t *oss, char* buf, int buf_size);
//int oss_write (oss_t *oss, char* buf, int byte_length);
//int oss_post (oss_t *oss);
//int oss_sync (oss_t *oss);

/* mixer functions */;
int oss_mixer_name_to_id(const char *name);
const char* oss_mixer_id_to_name(int id);
int oss_mixer_dev_id(oss_t *oss, const char *name);
int oss_mixer_stereo_cap(oss_t *oss, int id);
int oss_mixer_record_cap(oss_t *oss, int id);
int oss_mixer_record_enable(oss_t *oss, int id);
int oss_mixer_record_disable(oss_t *oss, int id);
int oss_mixer_vol_set(oss_t *oss, int id, int left, int right);
int oss_mixer_vol_get(oss_t *oss, int id, int *left, int *right);
int oss_mixer_close(oss_t *oss);
int oss_mixer_open (oss_t *oss, const char* device_name);

void oss_print_param(oss_t *oss);

#endif /* OSS_H */

