/* rtj1.h */

#ifndef RTJ1_H
#define RTJ1_H

#include <inttypes.h>

#define RTJ1_CSP_UNKNOWN 0
#define RTJ1_CSP_YUV444  1
#define RTJ1_CSP_YUV422  2
#define RTJ1_CSP_YUV420  3
#define RTJ1_CSP_YV12    4

#define RTJ1_CSP_INTERLACED        64
#define RTJ1_CSP_YUV444_INTERLACE  (RTJ1_CSP_YUV444|RTJ1_CSP_INTERLACED)
#define RTJ1_CSP_YUV422_INTERLACE  (RTJ1_CSP_YUV422|RTJ1_CSP_INTERLACED)
#define RTJ1_CSP_YUV420_INTERLACE  (RTJ1_CSP_YUV420|RTJ1_CSP_INTERLACED)
#define RTJ1_CSP_YV12_INTERLACE    (RTJ1_CSP_YV12  |RTJ1_CSP_INTERLACED)

#define RTJ1_CSP_INTERLACED_Y        128
#define RTJ1_CSP_YUV444_INTERLACE_Y  (RTJ1_CSP_YUV444|RTJ1_CSP_INTERLACED_Y)
#define RTJ1_CSP_YUV422_INTERLACE_Y  (RTJ1_CSP_YUV422|RTJ1_CSP_INTERLACED_Y)
#define RTJ1_CSP_YUV420_INTERLACE_Y  (RTJ1_CSP_YUV420|RTJ1_CSP_INTERLACED_Y)
#define RTJ1_CSP_YV12_INTERLACE_Y    (RTJ1_CSP_YV12  |RTJ1_CSP_INTERLACED_Y)

#define RTJ1_CSP_MASK (RTJ1_CSP_INTERLACED-1)
#define RTJ1_CSP_INTERLACED_MASK (~(RTJ1_CSP_MASK))

#define RTJ1_FLAG_NONE         (0)

typedef struct {
  int cc;
  int offset;
  int stride;
} comp_block_t;

#define RTJ1_MAGIC (('R')|(('T')<<8)|(('J')<<16)|(('1')<<24))

typedef struct {
  uint32_t magic;
  uint16_t width;
  uint16_t height;
  uint8_t  csp;
  uint8_t  quant;
  uint16_t method;
  uint32_t data_size;
} rtj1_head_t;

typedef struct {
  int width;
  int height;
  int chroma_horz_period;
  int chroma_vert_period;
  int u_offset;
  int v_offset;
  int num_comp;
  comp_block_t *comp_block;
  int csp;
  int image_csp;
  int quant;
  int dc_prec;

  int16_t *quant_mat;
  int16_t *inv_quant_mat;
  int16_t *quant_mul;
  int16_t *quant_add;
} rtj1_t;

#define RTJ1_OK    0
#define RTJ1_FAIL -1
#define RTJ1_NEED_MORE_DATA -2

#define RTJ1_MODE_ENCODE (1<<0)
#define RTJ1_MODE_DECODE (1<<1)

//rtj1_t* rtj1_init(int width, int height, int csp, int quant, int mode);
rtj1_t* rtj1_encode_init(int width, int height, int csp, int quant,int dc_prec);
rtj1_t* rtj1_decode_init(int width, int height, int csp);

rtj1_t* rtj1_quit(rtj1_t *rj);

int rtj1_encode(rtj1_t *rj, uint8_t *src, uint8_t *dst);

int rtj1_decode(rtj1_t *rj, uint8_t *src, uint8_t *dst, int size);
int rtj1_decode_analyze(rtj1_t *rj, uint8_t *src, int size);
int rtj1_set_decode_csp(rtj1_t *rj, int csp);
int rtj1_get_decode_csp(rtj1_t *rj);
int rtj1_get_csp(rtj1_t *rj);
int rtj1_get_width(rtj1_t *rj);
int rtj1_get_height(rtj1_t *rj);

#endif /* RTJ1_H */

