/* venc.h */

#ifndef VENC_H
#define VENC_H

#include <inttypes.h>
#include <pthread.h>

#include "vframe.h"
#include "vframe_mgr.h"
#include "vgrab.h"
#include "error_mgr.h"

#define VENC_STATE_NOT_READY       (1UL<<0)
#define VENC_STATE_STOP            (1UL<<1)
#define VENC_STATE_START           (1UL<<2)
#define VENC_STATE_PAUSE           (1UL<<3)
#define VENC_STATE_QUIT            (1UL<<4)
#define VENC_STATE_ERROR           (1UL<<5)
#define VENC_STATE_INTERFACE_WAIT  (1UL<<6)
#define VENC_STATE_THREAD_WAIT     (1UL<<7)
#define VENC_STATE_THREAD_FINISH   (1UL<<8)

//typedef int (*vframe_ready_cb_t)(void *);

typedef struct {
  pthread_t venc_th;
  pthread_mutex_t mutex;
  pthread_cond_t interface_wait_cond;
  pthread_cond_t thread_wait_cond;

  uint32_t state;

  vframe_t *src_vframe;
  vframe_t *dst_vframe;
  vframe_chain_t vframe_chain;

  vgrab_t *vgrab;
  vframe_mgr_t *vframe_mgr;
  vframe_ready_cb_t vframe_ready_cb;
  void *vframe_ready_cb_data;

  uint32_t vcodec;
  int    width;
  int    height;
  double fps;
  int    in_csp;
  int    out_csp;

  error_cb_t error_cb;
  void *error_cb_data;
} venc_t;

#define VENC_OK    0
#define VENC_FAIL -1

int venc_set_vframe_ready_cb(venc_t *venc,
              vframe_ready_cb_t vframe_ready_cb, void *vframe_ready_cb_data);
int venc_set_vframe_mgr(venc_t *venc, vframe_mgr_t *vfrm_mgr);
int venc_set_error_cb(venc_t *venc, error_cb_t error_cb, void *error_cb_data);
vframe_t *venc_get_vframe(venc_t *venc);
int venc_get_out_csp(venc_t *venc);
int venc_state_error(venc_t *venc);
int venc_state_start(venc_t *venc);
int venc_state_stop(venc_t *venc);
int venc_state_quit(venc_t *venc);
int venc_state_toggle_pause(venc_t *venc);
int venc_error_quit (venc_t *venc);
int venc_quit (venc_t *venc);
venc_t* venc_init (int width, int height, double fps, uint32_t vcodec, int in_csp, vgrab_t *vgrab);


#endif /* VENC_H */

