/* vframe.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "vframe.h"

vframe_t*
vframe_free(vframe_t *vfrm)
{
  if (vfrm->bufp) {
    free(vfrm->bufp);
//    vfrm->plane[0] = vfrm->plane[1] = vfrm->plane[2] = NULL;
  } else {
    free(vfrm->plane[0]);
    free(vfrm->plane[1]);
    free(vfrm->plane[2]);
  }
  free(vfrm);
  return NULL;
}

vframe_t*
vframe_alloc(int width, int height, int horz_border, int vert_border, int csp)
{
  int bufsize, offset;
  vframe_t *vfrm;

  vfrm = (vframe_t*)malloc(sizeof(vframe_t));
  if (!vfrm) {
    fprintf(stderr, "vframe_alloc: memory error.\n");
    return NULL;
  }
  vfrm->width[0] = width;
  vfrm->height[0] = height;
  vfrm->frame_width[0] = width + horz_border*2;
  vfrm->frame_height[0] = height + vert_border*2;
  vfrm->horz_border[0] = horz_border;
  vfrm->vert_border[0] = vert_border;

  switch (csp) {
    default:
      vfrm->width[1] = width/2;
      vfrm->width[2] = width/2;
      vfrm->height[1] = height/2;
      vfrm->height[2] = height/2;
      vfrm->frame_width[1] = width/2 + horz_border*2/2;
      vfrm->frame_width[2] = width/2 + horz_border*2/2;
      vfrm->frame_height[1] = height/2 + vert_border*2/2;
      vfrm->frame_height[2] = height/2 + vert_border*2/2;
      vfrm->horz_border[1] = horz_border/2;
      vfrm->horz_border[2] = horz_border/2;
      vfrm->vert_border[1] = vert_border/2;
      vfrm->vert_border[2] = vert_border/2;
      break;
  }
  vfrm->csp = csp;

  vfrm->stride[0] = vfrm->frame_width[0];
  vfrm->stride[1] = vfrm->frame_width[1];
  vfrm->stride[2] = vfrm->frame_width[2];

  bufsize =  vfrm->frame_width[0] * vfrm->frame_height[0]
           + vfrm->frame_width[1] * vfrm->frame_height[1]
           + vfrm->frame_width[2] * vfrm->frame_height[2];
  vfrm->bufp = (unsigned char*)memalign(16, bufsize);
  if (!vfrm->bufp) {
    free(vfrm);
    fprintf(stderr, "vframe_alloc: memory error.\n");
    return NULL;
  }

  vfrm->plane[0] = vfrm->bufp;
  offset = vfrm->frame_width[0] * vfrm->frame_height[0];
  vfrm->plane[1] = vfrm->bufp + offset;
  offset += vfrm->frame_width[1] * vfrm->frame_height[1];
  vfrm->plane[2] = vfrm->bufp + offset;

  vfrm->offset[0] =  vfrm->plane[0]
                   + (vfrm->frame_width[0] * vfrm->vert_border[0])
		   + vfrm->horz_border[0];
  vfrm->offset[1] =  vfrm->plane[1]
                   + (vfrm->frame_width[1] * vfrm->vert_border[1])
		   + vfrm->horz_border[1];
  vfrm->offset[2] =  vfrm->plane[2]
                   + (vfrm->frame_width[2] * vfrm->vert_border[2])
		   + vfrm->horz_border[2];

  bufsize = vfrm->frame_width[0] * vfrm->frame_height[0];
  memset(vfrm->plane[0], 0, bufsize);
  switch (csp) {
    default:
      bufsize = vfrm->frame_width[1] * vfrm->frame_height[1];
      memset(vfrm->plane[1], 128, bufsize);
      bufsize = vfrm->frame_width[2] * vfrm->frame_height[2];
      memset(vfrm->plane[2], 128, bufsize);
      break;
  }

  return vfrm;
}

void
plain_to_vframe(unsigned char *src, vframe_t *vfrm)
{
  int i, j, k, width, height, stride;
  unsigned char *dp;

  for (k = 0; k < 3; k++) {
    width = vfrm->width[k];
    height = vfrm->height[k];
    stride = vfrm->stride[k];
    dp = vfrm->offset[k];
    for (j = 0; j < height; j++) {
      for (i = 0; i < width; i++) {
	dp[i] = src[i];
      }
      src += width;
      dp += stride;
    }
  }
}

void
vframe_to_plain(vframe_t *vfrm, unsigned char *dst)
{
  int i, j, k, width, height, stride;
  unsigned char *sp;

  for (k = 0; k < 3; k++) {
    width = vfrm->width[k];
    height = vfrm->height[k];
    stride = vfrm->stride[k];
    sp = vfrm->offset[k];
    for (j = 0; j < height; j++) {
      for (i = 0; i < width; i++) {
	dst[i] = sp[i];
      }
      dst += width;
      sp += stride;
    }
  }
}

