/* vframe.h */

#ifndef VFRAME_H
#define VFRAME_H

#include <inttypes.h>

#define VFRAME_FLAG_NORMAL (1<<0)
#define VFRAME_FLAG_EXT    (1<<1)
#define VFRAME_FLAG_START  (1<<2)
#define VFRAME_FLAG_END    (1<<3)
#define VFRAME_FLAG_QUIT   (1<<4)
#define VFRAME_FLAG_ERROR  (1<<5)
#define VFRAME_FLAG_DROP   (1<<6)
#define VFRAME_FLAG_INTRA  (1<<7)
#define VFRAME_FLAG_INTER  (1<<8)

typedef struct VFRAME_T {
  uint8_t *planes[3];
  int32_t stride[3];
  int32_t width, height;

  uint8_t *data;
  int32_t data_length;
  int32_t data_buf_size;

  struct VFRAME_T *prev;
  struct VFRAME_T *next;

  uint32_t flag;
//  uint32_t exflag;
  uint32_t duration;
  uint32_t count;

  int32_t csp;
  int32_t horz_border, vert_border;
  uint8_t *bufp;
} vframe_t;


typedef int (*vframe_ready_cb_t)(void *);


typedef struct {
  int num_frames;
  vframe_t *vframes;
} vframe_chain_t;

static inline void
vframe_chain_init(vframe_chain_t *vfrm_chain)
{
  vfrm_chain->num_frames = 0;
  vfrm_chain->vframes = NULL;
}

static inline int
vframe_chain_add(vframe_chain_t *vfrm_chain, vframe_t *vfrm)
{
  if (vfrm_chain->vframes) {
    vfrm->prev = vfrm_chain->vframes->prev;
    vfrm->next = vfrm_chain->vframes;
    vfrm_chain->vframes->prev->next = vfrm;
    vfrm_chain->vframes->prev = vfrm;
  } else {
    vfrm->prev = vfrm->next = vfrm;
    vfrm_chain->vframes = vfrm;
  }
  vfrm_chain->num_frames++;
  return vfrm_chain->num_frames;
}

static inline vframe_t*
vframe_chain_remove(vframe_chain_t *vfrm_chain)
{
  vframe_t *vfrm;

  if (!vfrm_chain->vframes) {
    return NULL;
  }
  vfrm = vfrm_chain->vframes;
  if (vfrm->next == vfrm) {
    vfrm_chain->vframes = NULL;
  } else {
    vfrm_chain->vframes = vfrm->next;
    vfrm->prev->next = vfrm->next;
    vfrm->next->prev = vfrm->prev;
  }
  vfrm->prev = vfrm->next = NULL;
  vfrm_chain->num_frames--;
  return vfrm;
}

#endif /* VFRAME_H */

