/* vgrab.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
//#include <unistd.h>
//#include <sys/types.h>
//#include <sys/stat.h>
//#include <string.h>
#include <pthread.h>

#include "csp.h"
#include "v4l.h"
#include "vgrab.h"
#include "get_cur_u_time.h"
#include "vframe.h"
#include "vframe_mgr.h"
#include "error_mgr.h"


/* Thread functions */

static void vgrab_error(vgrab_t *vgrab, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "vgrab: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);

  if (!vgrab) return;

  pthread_mutex_lock(&vgrab->mutex);
  vgrab->state |= VGRAB_STATE_ERROR;
  pthread_mutex_unlock(&vgrab->mutex);
  if (vgrab->error_cb)
    vgrab->error_cb(vgrab->error_cb_data);
//  exit(1);
}

static unsigned char*
grab_pic_sync(vgrab_t *vgrab)
{
  unsigned char *buf = NULL;
  int error_count = 0;

  for (error_count = 0; buf == NULL; error_count++) {
    buf = v4l_grab_pic_sync(vgrab->v4l);
    vgrab->grab_time = get_cur_u_time();
    if (buf != NULL) {
      break;
    } else if (error_count > 30) {
      vgrab_error(vgrab, "vgrab_thread: buf NULL count %d\n", error_count);
      break;
    }
  }
  vgrab->grab_buf = buf;
  return buf;
}

static int
grab_pic_capture(vgrab_t *vgrab)
{
  if (v4l_grab_pic_capture(vgrab->v4l) < 0) {
    fprintf(stderr, "vgrab_thread: v4l capture error.\n");
    return VGRAB_FAIL;
  }
  return VGRAB_OK;
}

static int
add_frame_chain(vgrab_t *vgrab, vframe_t *vfrm)
{
  int ret;
  pthread_mutex_lock(&vgrab->mutex);
  ret = vframe_chain_add(&vgrab->vframe_chain, vfrm);
  pthread_mutex_unlock(&vgrab->mutex);
  if (vgrab->vframe_ready_cb)
    vgrab->vframe_ready_cb(vgrab->vframe_ready_cb_data);
  return ret;
}

static int
put_grab_pic(vgrab_t *vgrab)
{
  vframe_t *vfrm;

  vfrm = vframe_mgr_get_vframe(vgrab->vframe_mgr);
  if (vfrm == NULL) {
#ifndef NDEBUG
    fprintf(stderr, "vgrab: frame %d dropped, drop total %d\n", vgrab->frame_num, vgrab->drop_frame_num);
#endif /* NDEBUG */
    vfrm = vframe_mgr_get_ext_vframe(vgrab->vframe_mgr);
    if (vfrm == NULL) {
      vgrab_error(vgrab, "vgrab: vframe_mgr_get_ext_frame failed.\n");
      return VGRAB_FAIL;
    }
    vfrm->flag = VFRAME_FLAG_DROP|VFRAME_FLAG_EXT;
    vfrm->data_length = 0;
    vfrm->csp = CSP_UNKNOWN;
    vgrab->drop_frame_num++;
  } else {
    vfrm->flag = VFRAME_FLAG_NORMAL;
    vgrab->planes[0] = vgrab->grab_buf + vgrab->offset[0];
    vgrab->planes[1] = vgrab->grab_buf + vgrab->offset[1];
    vgrab->planes[2] = vgrab->grab_buf + vgrab->offset[2];
    vgrab->csp_conv_func(vgrab->planes, vfrm->planes, vgrab->stride, vfrm->stride, vgrab->width, vgrab->height);
    vfrm->data_length = 0;
    vfrm->csp = vgrab->csp;
  }
  if (vgrab->vfrm) {
    vgrab->vfrm->duration = vgrab->grab_time - vgrab->vfrm_time;
    add_frame_chain(vgrab, vgrab->vfrm);
  }
  vgrab->vfrm = vfrm;
  vgrab->vfrm_time = vgrab->grab_time;
  vgrab->frame_num++;
  return VGRAB_OK;
}

static int
put_grab_pic_flush(vgrab_t *vgrab)
{
  if (vgrab->vfrm) {
    vgrab->vfrm->duration = vgrab->grab_time - vgrab->vfrm_time;
    add_frame_chain(vgrab, vgrab->vfrm);
  }
  vgrab->vfrm = NULL;
  vgrab->vfrm_time = vgrab->grab_time;
  vgrab->frame_num++;
  return VGRAB_OK;
}

static int
put_ext_vframe(vgrab_t *vgrab, uint32_t flag)
{
  vframe_t *vfrm;

  vfrm = vframe_mgr_get_ext_vframe(vgrab->vframe_mgr);
  if (!vfrm) {
    vgrab_error(vgrab, "vframe_mgr_get_ext_vframe failed.\n");
    return VGRAB_FAIL;
  }
  vfrm->flag = flag|VFRAME_FLAG_EXT;
  vfrm->duration = 0;
  vfrm->data_length = 0;
  add_frame_chain(vgrab, vfrm);
  return VGRAB_OK;
}

static inline int
get_state(vgrab_t *vgrab)
{
  int state;
  pthread_mutex_lock(&vgrab->mutex);
  state = vgrab->state;
  pthread_mutex_unlock(&vgrab->mutex);
  return state;
}

static void
state_change_done(vgrab_t *vgrab)
{
  pthread_mutex_lock(&vgrab->mutex);
  if (vgrab->state & VGRAB_STATE_INTERFACE_WAIT) {
    vgrab->state &= ~VGRAB_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&vgrab->interface_wait_cond);
  }
  pthread_mutex_unlock(&vgrab->mutex);
}

static void
set_state(vgrab_t *vgrab, uint32_t state)
{
  pthread_mutex_lock(&vgrab->mutex);
  vgrab->state &= ~(VGRAB_STATE_START|VGRAB_STATE_PAUSE|VGRAB_STATE_STOP|VGRAB_STATE_QUIT);
  vgrab->state |= state;
  if (vgrab->state & VGRAB_STATE_INTERFACE_WAIT) {
    vgrab->state &= ~VGRAB_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&vgrab->interface_wait_cond);
  }
  pthread_mutex_unlock(&vgrab->mutex);
}

static inline void
set_next_time(vgrab_t *vgrab)
{
  if (vgrab->real_time_flag)
    vgrab->next_time = vgrab->grab_time;
  else
    vgrab->next_time += vgrab->frame_duration;
}

/* thread function */

static void*
vgrab_thread(void *data)
{
  int state;
  int real_time_flag;
  int rec_start_flag;
  vgrab_t *vgrab = (vgrab_t*)data;

  pthread_mutex_lock(&vgrab->mutex);
  real_time_flag = vgrab->real_time_flag;
  if (vgrab->state & VGRAB_STATE_INTERFACE_WAIT)
    pthread_cond_signal(&vgrab->interface_wait_cond);
  vgrab->state = VGRAB_STATE_STOP;
  pthread_mutex_unlock(&vgrab->mutex);

  if (v4l_grab_pic_start (vgrab->v4l) != V4L_OK) {
    vgrab_error(vgrab, "v4l_grab_pic_start failed.\n");
  }

  rec_start_flag = 0;

  while (1) {
    state = get_state(vgrab);

    if (state & VGRAB_STATE_START) {
      if (!rec_start_flag) {
        if (put_ext_vframe(vgrab, VFRAME_FLAG_START) == VGRAB_FAIL) {
	  vgrab_error(vgrab, "vgrab_thread: put_ext_vframe() failed.\n");
	  break;
        } else {
	  rec_start_flag = 1;
          vgrab->frame_num = 0;
          vgrab->drop_frame_num = 0;
	}
      }

      grab_pic_sync(vgrab);
      while (vgrab->grab_time < vgrab->state_time) {
        grab_pic_capture(vgrab);
        grab_pic_sync(vgrab);
      }
      vgrab->state_time = vgrab->grab_time;
      vgrab->next_time = vgrab->grab_time;
      state_change_done(vgrab);
      put_grab_pic(vgrab);
      set_next_time(vgrab);
      grab_pic_capture(vgrab);

      while (state & VGRAB_STATE_START) {
        grab_pic_sync(vgrab);
	if (vgrab->grab_time >= vgrab->next_time) {
	  if (!vgrab->real_time_flag)
	    vgrab->grab_time = vgrab->next_time;
	  put_grab_pic(vgrab);
	  set_next_time(vgrab);
	}
#ifndef NDEBUG
       	else {
	  static int skip = 0;
	  fprintf(stderr, "skip %d\n", skip++);
	}
#endif /* NDEBUG */
        grab_pic_capture(vgrab);
        state = get_state(vgrab);
      }

      if (state & VGRAB_STATE_ERROR) {
	break;
      }

      grab_pic_sync(vgrab);
      while (vgrab->grab_time < vgrab->next_time) {
        grab_pic_capture(vgrab);
        grab_pic_sync(vgrab);
      }
      grab_pic_capture(vgrab);
      if (!vgrab->real_time_flag)
        vgrab->grab_time = vgrab->next_time;
      put_grab_pic_flush(vgrab);
      vgrab->state_time = vgrab->grab_time;
    }
    else if (state & VGRAB_STATE_PAUSE) {
      state_change_done(vgrab);
      while (state & VGRAB_STATE_PAUSE) {
        grab_pic_sync(vgrab);
        grab_pic_capture(vgrab);
        state = get_state(vgrab);
      }
    }
    else if (state & VGRAB_STATE_STOP) {
      if (rec_start_flag) {
        if (put_ext_vframe(vgrab, VFRAME_FLAG_END) == VGRAB_FAIL) {
          vgrab_error(vgrab, "vgrab_thread: put_ext_vframe() failed.\n");
          break;
        }
	rec_start_flag = 0;
      }
      state_change_done(vgrab);

      while (state & VGRAB_STATE_STOP) {
        grab_pic_sync(vgrab);
        grab_pic_capture(vgrab);
        state = get_state(vgrab);
      }
    }
    else if (state & VGRAB_STATE_QUIT) {
      if (rec_start_flag) {
        if (put_ext_vframe(vgrab, VFRAME_FLAG_END) == VGRAB_FAIL) {
          vgrab_error(vgrab, "vgrab_thread: put_ext_vframe() failed.\n");
          break;
        }
	rec_start_flag = 0;
      }
      if (put_ext_vframe(vgrab, VFRAME_FLAG_QUIT) == VGRAB_FAIL) {
	vgrab_error(vgrab, "vgrab_thread: put_ext_vframe() failed.\n");
      }
      break;
    }
    else if (state & VGRAB_STATE_ERROR) {
      if (put_ext_vframe(vgrab, VFRAME_FLAG_ERROR) == VGRAB_FAIL) {
	vgrab_error(vgrab, "vgrab_thread: put_ext_vframe() failed.\n");
      }
      break;
    }
  }

  v4l_grab_pic_stop(vgrab->v4l);

  pthread_mutex_lock(&vgrab->mutex);
  if (vgrab->state & VGRAB_STATE_ERROR)
    fprintf(stderr, "vgrab_thread: STATE_ERROR break\n");
  if (vgrab->state & VGRAB_STATE_INTERFACE_WAIT) {
    vgrab->state &= ~VGRAB_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&vgrab->interface_wait_cond);
  }
  vgrab->state &= (VGRAB_STATE_QUIT|VGRAB_STATE_ERROR);
  vgrab->state |= VGRAB_STATE_THREAD_FINISH;
  pthread_mutex_unlock(&vgrab->mutex);
  //state_change_done(vgrab);

#ifndef NDEBUG
  fprintf(stderr, "vgrab_thread: done\n");
#endif /* NDEBUG */
  return vgrab;
}

/* Interface functions */

int
vgrab_get_width(vgrab_t *vgrab)
{
  int width;
  pthread_mutex_lock(&vgrab->mutex);
  width = vgrab->width;
  pthread_mutex_unlock(&vgrab->mutex);
  return width;
}

int
vgrab_get_height(vgrab_t *vgrab)
{
  int height;
  pthread_mutex_lock(&vgrab->mutex);
  height = vgrab->height;
  pthread_mutex_unlock(&vgrab->mutex);
  return height;
}

double
vgrab_get_fps(vgrab_t *vgrab)
{
  double fps;
  pthread_mutex_lock(&vgrab->mutex);
  fps = vgrab->fps;
  pthread_mutex_unlock(&vgrab->mutex);
  return fps;
}

int
vgrab_set_vframe_ready_cb(vgrab_t *vgrab, vframe_ready_cb_t vframe_ready_cb, void *vframe_ready_cb_data)
{
  pthread_mutex_lock(&vgrab->mutex);
  vgrab->vframe_ready_cb = vframe_ready_cb;
  vgrab->vframe_ready_cb_data = vframe_ready_cb_data;
  pthread_mutex_unlock(&vgrab->mutex);
  return VGRAB_OK;
}

int
vgrab_set_vframe_mgr(vgrab_t *vgrab, vframe_mgr_t *vfrm_mgr)
{
  pthread_mutex_lock(&vgrab->mutex);
  vgrab->vframe_mgr = vfrm_mgr;
  pthread_mutex_unlock(&vgrab->mutex);
  return VGRAB_OK;
}

vframe_t*
vgrab_get_vframe(vgrab_t *vgrab)
{
  vframe_t *vfrm;
  pthread_mutex_lock(&vgrab->mutex);
  vfrm = vframe_chain_remove(&vgrab->vframe_chain);
  pthread_mutex_unlock(&vgrab->mutex);
  return vfrm;
}

int
vgrab_set_error_cb(vgrab_t *vgrab, error_cb_t error_cb, void *error_cb_data)
{
  pthread_mutex_lock(&vgrab->mutex);
  vgrab->error_cb = error_cb;
  vgrab->error_cb_data = error_cb_data;
  pthread_mutex_unlock(&vgrab->mutex);
  return VGRAB_OK;
}

static int
vgrab_state_change(vgrab_t *vgrab, u_time_t *state_time, int state)
{
  int ret = VGRAB_FAIL;

  pthread_mutex_lock(&vgrab->mutex);
  if (vgrab->state & VGRAB_STATE_THREAD_WAIT) {
    vgrab->state &= ~VGRAB_STATE_THREAD_WAIT;
    pthread_cond_signal(&vgrab->thread_wait_cond);
  }
  if (vgrab->state & (state|VGRAB_STATE_QUIT|VGRAB_STATE_ERROR|VGRAB_STATE_THREAD_FINISH)) {
    ret = VGRAB_OK;
  } else {
    vgrab->state &= ~(VGRAB_STATE_START|VGRAB_STATE_PAUSE|VGRAB_STATE_STOP);
    vgrab->state |= state;
    vgrab->state |= VGRAB_STATE_INTERFACE_WAIT;
    vgrab->state_time = *state_time;
    while (vgrab->state & VGRAB_STATE_INTERFACE_WAIT)
      pthread_cond_wait(&vgrab->interface_wait_cond, &vgrab->mutex);
    *state_time = vgrab->state_time;
    ret = VGRAB_OK;
  }
  pthread_mutex_unlock(&vgrab->mutex);
  return ret;
}

int
vgrab_state_start(vgrab_t *vgrab, u_time_t *state_time)
{
  return vgrab_state_change(vgrab, state_time, VGRAB_STATE_START);
}

int
vgrab_state_stop(vgrab_t *vgrab, u_time_t *state_time)
{
  return vgrab_state_change(vgrab, state_time, VGRAB_STATE_STOP);
}

int
vgrab_state_toggle_pause(vgrab_t *vgrab, u_time_t *state_time)
{
  int state;

  state = get_state(vgrab);
  if (state & VGRAB_STATE_START)
    return vgrab_state_change(vgrab, state_time, VGRAB_STATE_PAUSE);
  else if (state & VGRAB_STATE_PAUSE)
    return vgrab_state_change(vgrab, state_time, VGRAB_STATE_START);
  return VGRAB_OK;
}

int
vgrab_state_quit(vgrab_t *vgrab, u_time_t *state_time)
{
  return vgrab_state_change(vgrab, state_time, VGRAB_STATE_QUIT);
}

int
vgrab_state_error(vgrab_t *vgrab, u_time_t *state_time)
{
  pthread_mutex_lock(&vgrab->mutex);
  if (vgrab->state & VGRAB_STATE_THREAD_WAIT) {
    vgrab->state &= ~VGRAB_STATE_THREAD_WAIT;
    pthread_cond_signal(&vgrab->thread_wait_cond);
  }
  if (vgrab->state & (VGRAB_STATE_ERROR|VGRAB_STATE_THREAD_FINISH)) {
    pthread_mutex_unlock(&vgrab->mutex);
    return VGRAB_OK;
  }
  vgrab->state = VGRAB_STATE_ERROR;
  *state_time = *state_time;
  pthread_mutex_unlock(&vgrab->mutex);
  return VGRAB_OK;
}

int
vgrab_error_quit (vgrab_t *vgrab, u_time_t *state_time)
{
  int retcode, state;
  void* retval;
  vframe_t *vfrm;

  vgrab_state_error(vgrab, state_time);

  state = get_state(vgrab);
  if (!(state & VGRAB_STATE_THREAD_FINISH) &&
      !(state & VGRAB_STATE_NOT_READY)) {
    retcode = pthread_join(vgrab->vgrab_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "vgrab_error_quit: thread join error %d\n", retcode);
  }

  if (vgrab->v4l)
    v4l_quit(vgrab->v4l);
  if (vgrab->vfrm)
    vframe_mgr_add_free_frame(vgrab->vframe_mgr, vgrab->vfrm);
  pthread_mutex_lock(&vgrab->mutex);
  while ((vfrm = vframe_chain_remove(&vgrab->vframe_chain)) != NULL)
    vframe_mgr_add_free_frame(vgrab->vframe_mgr, vfrm);
  pthread_mutex_unlock(&vgrab->mutex);
  free(vgrab);
  
  return VGRAB_OK;
}

int
vgrab_quit (vgrab_t *vgrab, u_time_t *quit_time)
{
  int retcode, state;
  void* retval;

  vgrab_state_quit(vgrab, quit_time);

  state = get_state(vgrab);
  if (!(state & VGRAB_STATE_THREAD_FINISH) &&
      !(state & VGRAB_STATE_NOT_READY)) {
    retcode = pthread_join(vgrab->vgrab_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "vgrab_error_quit: thread join error %d\n", retcode);
  }

  if (vgrab->v4l)
    v4l_quit(vgrab->v4l);
  if (vgrab->vfrm)
    vframe_mgr_add_free_frame(vgrab->vframe_mgr, vgrab->vfrm);
  free(vgrab);
  
  return VGRAB_OK;
}

static struct CSP_TABLE {
  int csp;
  int v4l_csp;
  int grab_csp;
} csp_table[] = {
  { CSP_RGB24,  V4L_CSP_RGB24,  CSP_BGR24 },
  { CSP_BGR24,  V4L_CSP_RGB24,  CSP_BGR24 },
  { CSP_ARGB32, V4L_CSP_ARGB32, CSP_BGRA32 },
  { CSP_BGRA32, V4L_CSP_ARGB32, CSP_BGRA32 },
  { CSP_YUY2,   V4L_CSP_YUY2,   CSP_YUY2 },
  { CSP_YUYV,   V4L_CSP_YUYV,   CSP_YUYV },
  { CSP_UYVY,   V4L_CSP_UYVY,   CSP_UYVY },
  { CSP_YUV422, V4L_CSP_YUV422, CSP_YUV422 },
  { CSP_YUV420, V4L_CSP_YUV420, CSP_YUV420 },
  { CSP_YV12,   V4L_CSP_YUV420, CSP_YUV420 },
  { CSP_I420,   V4L_CSP_YUV420, CSP_YUV420 },
};
static int csp_table_num = sizeof(csp_table)/sizeof(struct CSP_TABLE);

static struct CSP_TABLE*
get_csp_table(int csp)
{
  int i;
  for (i = 0; i < csp_table_num; i++)
    if (csp_table[i].csp == csp)
      return &csp_table[i];
  return NULL;
}

vgrab_t*
vgrab_init (int width, int height, double fps, int csp, v4l_param_t *v4l_param)
{
  int retcode;
  int v4l_csp;
  int grab_csp;
  int i, offset;
  CSP_INFO csp_info;
  struct CSP_TABLE *csp_table;
  vgrab_t *vgrab;
  u_time_t state_time = 0;

  vgrab = (vgrab_t*)malloc(sizeof(vgrab_t));
  if (!vgrab) {
    vgrab_error(vgrab, "memory allocate error.\n");
    return NULL;
  }
  vgrab->state = VGRAB_STATE_NOT_READY;
  vgrab->vfrm = NULL;
  vgrab->vframe_mgr = NULL;
  vgrab->vframe_ready_cb = NULL;
  vgrab->vframe_ready_cb_data = NULL;
  vgrab->error_cb = NULL;
  vgrab->error_cb_data = NULL;
  vframe_chain_init(&vgrab->vframe_chain);
  vgrab->v4l = NULL;

  vgrab->width = width;
  vgrab->height = height;
  vgrab->csp = csp;
  csp_table = get_csp_table(csp);
  if (!csp_table) {
    vgrab_error(vgrab, "vgrab_init: vgrab doesn't support \"%s\" capture.\n", csp_to_name(csp));
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  grab_csp = csp_table->grab_csp;
  v4l_csp = csp_table->v4l_csp;

  vgrab->v4l = v4l_init(v4l_param->device_name, v4l_param->source_channel,
      v4l_param->tuner, v4l_param->norm, v4l_param->audio);
  if (!vgrab->v4l) {
    vgrab_error(vgrab, "vgrab_init: v4l_init failed.\n");
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  if (v4l_capture_cap(vgrab->v4l) == V4L_FAIL) {
    vgrab_error(vgrab, "vgrab_init: video device doesn't support capture.\n");
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  if (v4l_set_window_size(vgrab->v4l, width, height) == V4L_FAIL) {
    vgrab_error(vgrab, "vgrab_init: v4l_set_window_size() failed.\n");
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  if (v4l_get_window_size(vgrab->v4l, &width, &height) == V4L_FAIL) {
    vgrab_error(vgrab, "vgrab_init: v4l_get_window_size() failed.\n");
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  if (width != vgrab->width || height != vgrab->height) {
    fprintf(stderr, "vgrab_init: can't set picture size %dx%d, use %dx%d\n",
	vgrab->width, vgrab->height, width, height);
    vgrab->width = width;
    vgrab->height = height;
  }
  v4l_param->norm = v4l_get_norm(vgrab->v4l);

  v4l_set_brightness(vgrab->v4l, v4l_param->brightness);
  v4l_set_hue(vgrab->v4l, v4l_param->hue);
  v4l_set_colour(vgrab->v4l, v4l_param->colour);
  v4l_set_contrast(vgrab->v4l, v4l_param->contrast);
  v4l_set_whiteness(vgrab->v4l, v4l_param->whiteness);

  if (v4l_param->freq > 0)
    v4l_set_freq(vgrab->v4l, v4l_param->freq * 16 / 1000);

  if (v4l_set_csp(vgrab->v4l, v4l_csp) == V4L_FAIL) {
    vgrab_error(vgrab, "vgrab_init: video device doesn't support \"%s\" capture.\n", csp_to_name(csp));
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }

  vgrab->csp_conv_func = csp_conv_get_convert_func(grab_csp, csp);
  if (!vgrab->csp_conv_func) {
    vgrab_error(vgrab, "vgrab_init: cannot convert csp %s to %s\n", csp_to_name(v4l_csp), csp_to_name(csp));
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }
  csp_get_info(grab_csp, &csp_info);
  vgrab->stride[1] = vgrab->stride[2] = 0;
  for (i = 0; i < csp_info.num_planes; i++) {
    vgrab->stride[i] = width / csp_info.horz_period[i];
  }
  vgrab->offset[0] = vgrab->offset[1] = vgrab->offset[2] = 0;
  offset = 0;
  for (i = 0; i < csp_info.num_planes; i++) {
    vgrab->offset[i] = offset;
    offset += width*height / (csp_info.horz_period[i]*csp_info.vert_period[i]);
  }

  if (fps == 0) {
    vgrab->real_time_flag = 1;
    vgrab->frame_duration = 0;
    switch (v4l_param->norm) {
      case V4L_NORM_NTSC:
        fps = 29.97;
        break;
      case V4L_NORM_PAL:
        fps = 25;
        break;
      default:
	fps = 25;
	break;
    }
  } else {
    vgrab->real_time_flag = 0;
    vgrab->frame_duration = (uint32_t)(1000000.0 / fps + 0.5);
  }

  vgrab->fps = fps;

  vgrab->frame_num = 0;
  vgrab->drop_frame_num = 0;
  vgrab->grab_buf = NULL;
  vgrab->vfrm_time = 0;

  pthread_mutex_init (&vgrab->mutex, NULL);
  pthread_cond_init (&vgrab->interface_wait_cond, NULL);
  pthread_cond_init (&vgrab->thread_wait_cond, NULL);

  vgrab->state = VGRAB_STATE_NOT_READY;
  retcode = pthread_create (&vgrab->vgrab_th, NULL, vgrab_thread, vgrab);
  if (retcode != 0) {
    vgrab_error(vgrab, "vgrab_init: thread create failed %d\n", retcode);
    vgrab_error_quit(vgrab, &state_time);
    return NULL;
  }

  pthread_mutex_lock(&vgrab->mutex);
  while (vgrab->state & VGRAB_STATE_NOT_READY) {
    vgrab->state |= VGRAB_STATE_INTERFACE_WAIT;
    pthread_cond_wait(&vgrab->interface_wait_cond, &vgrab->mutex);
  }
  pthread_mutex_unlock(&vgrab->mutex);

  return vgrab;
}

