/* vgrab.h */

#ifndef VGRAB_H
#define VGRAB_H

#include <inttypes.h>
#include <pthread.h>

#include "csp.h"
#include "get_cur_u_time.h"
#include "error_mgr.h"
#include "v4l.h"
#include "vframe.h"
#include "vframe_mgr.h"


#define VGRAB_STATE_NOT_READY       (1UL<<0)
#define VGRAB_STATE_STOP            (1UL<<1)
#define VGRAB_STATE_START           (1UL<<2)
#define VGRAB_STATE_PAUSE           (1UL<<3)
#define VGRAB_STATE_QUIT            (1UL<<4)
#define VGRAB_STATE_ERROR           (1UL<<5)
#define VGRAB_STATE_INTERFACE_WAIT  (1UL<<6)
#define VGRAB_STATE_THREAD_WAIT     (1UL<<7)
#define VGRAB_STATE_THREAD_FINISH   (1UL<<8)

//typedef int (*vframe_ready_cb_t)(void *);

typedef struct {
  pthread_t vgrab_th;
  pthread_mutex_t mutex;
  pthread_cond_t interface_wait_cond;
  pthread_cond_t thread_wait_cond;

  csp_conv_func_t csp_conv_func;
  uint8_t *planes[3];
  int32_t stride[3];
  int32_t offset[3];
  int width;
  int height;
  int csp;
  int frame_num;
  int drop_frame_num;
  v4l_t *v4l;
  uint32_t state;
  int real_time_flag;
  u_time_t state_time;
  u_time_t grab_time;
  u_time_t next_time;
  uint8_t *grab_buf;
  uint32_t frame_duration;
  vframe_t *vfrm;
  u_time_t vfrm_time;
  vframe_mgr_t *vframe_mgr;
  vframe_chain_t vframe_chain;
  vframe_ready_cb_t vframe_ready_cb;
  void *vframe_ready_cb_data;
  double fps;

  error_cb_t error_cb;
  error_cb_t error_cb_data;
} vgrab_t;

#define VGRAB_OK    0
#define VGRAB_FAIL -1

vgrab_t* vgrab_init (int width, int height, double fps, int csp,
    v4l_param_t *v4l_param);
int vgrab_quit (vgrab_t *vgrab, u_time_t *quit_time);
int vgrab_error_quit (vgrab_t *vgrab, u_time_t *quit_time);
int vgrab_state_start(vgrab_t *vgrab, u_time_t *state_time);
int vgrab_state_toggle_pause(vgrab_t *vgrab, u_time_t *state_time);
int vgrab_state_stop(vgrab_t *vgrab, u_time_t *state_time);
int vgrab_state_quit(vgrab_t *vgrab, u_time_t *state_time);
int vgrab_state_error(vgrab_t *vgrab, u_time_t *state_time);

int vgrab_get_width(vgrab_t *vgrab);
int vgrab_get_height(vgrab_t *vgrab);
double vgrab_get_fps(vgrab_t *vgrab);
vframe_t* vgrab_get_vframe(vgrab_t *vgrab);
int vgrab_set_vframe_mgr(vgrab_t *vgrab, vframe_mgr_t *vfrm_mgr);
int vgrab_set_vframe_ready_cb(vgrab_t *vgrab,
    vframe_ready_cb_t vframe_ready_cb, void *vframe_ready_cb_data);
int vgrab_set_error_cb(vgrab_t *vgrab, error_cb_t error_cb,void *error_cb_data);

#endif /* VGRAB_H */

